Package: dsROCrate
Title: 'DataSHIELD' RO-Crate Governance Functions
Version: 0.1.0
Authors@R: c(
    person(given = "Roberto",
           family = "Villegas-Diaz",
           role = c("aut", "cre"),
           email = "r.villegas-diaz@outlook.com",
           comment = c(ORCID = "0000-0001-5036-8661")),
    person(given = "Becca",
           family = "Wilson",
           role = c("aut"),
           comment = c(ORCID = "0000-0003-2294-593X")),
    person(given = "Olly",
           family = "Butters",
           role = c("aut"),
           comment = c(ORCID = "0000-0003-0354-8461")),
    person(given = "Stuart",
           family = "Wheater",
           role = c("aut"),
           comment = c(ORCID = "0009-0003-2419-1964")),
    person("University of Liverpool", role = c("cph")))
Description: Tools for wrapping 'DataSHIELD' analyses into RO-Crate 
    (Research Object Crate) objects. Provides functions to create structured
    metadata for federated data analysis projects, enabling governance 
    tracking of data access, project membership, analysis execution and 
    output validation across distributed data sources.
License: MIT + file LICENSE
Suggests: dsBaseClient, DSI, DSOpal, fs, knitr, MolgenisArmadillo,
        testthat (>= 3.0.0), withr
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-GB
Imports: digest, dplyr, DSMolgenisArmadillo, jsonlite, opalr (>=
        3.6.1), purrr, RcppTOML, rmarkdown, rocrateR (>= 0.1.0),
        tibble, vtree, xptr, yaml
Depends: R (>= 4.1.0)
VignetteBuilder: knitr
URL: https://github.com/FederatedMethods/dsROCrate
BugReports: https://github.com/FederatedMethods/dsROCrate/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-01 12:17:56 UTC; robertovillegas-diaz
Author: Roberto Villegas-Diaz [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5036-8661>),
  Becca Wilson [aut] (ORCID: <https://orcid.org/0000-0003-2294-593X>),
  Olly Butters [aut] (ORCID: <https://orcid.org/0000-0003-0354-8461>),
  Stuart Wheater [aut] (ORCID: <https://orcid.org/0009-0003-2419-1964>),
  University of Liverpool [cph]
Maintainer: Roberto Villegas-Diaz <r.villegas-diaz@outlook.com>
Repository: CRAN
Date/Publication: 2026-06-01 12:40:08 UTC
Built: R 4.5.2; ; 2026-06-01 14:53:06 UTC; unix

Package: dsROCrate
Title: 'DataSHIELD' RO-Crate Governance Functions
Version: 0.1.0
Authors@R: c(
    person(given = "Roberto",
           family = "Villegas-Diaz",
           role = c("aut", "cre"),
           email = "r.villegas-diaz@outlook.com",
           comment = c(ORCID = "0000-0001-5036-8661")),
    person(given = "Becca",
           family = "Wilson",
           role = c("aut"),
           comment = c(ORCID = "0000-0003-2294-593X")),
    person(given = "Olly",
           family = "Butters",
           role = c("aut"),
           comment = c(ORCID = "0000-0003-0354-8461")),
    person(given = "Stuart",
           family = "Wheater",
           role = c("aut"),
           comment = c(ORCID = "0009-0003-2419-1964")),
    person("University of Liverpool", role = c("cph")))
Description: Tools for wrapping 'DataSHIELD' analyses into RO-Crate 
    (Research Object Crate) objects. Provides functions to create structured
    metadata for federated data analysis projects, enabling governance 
    tracking of data access, project membership, analysis execution and 
    output validation across distributed data sources.
License: MIT + file LICENSE
Suggests: dsBaseClient, DSI, DSOpal, fs, knitr, MolgenisArmadillo,
        testthat (>= 3.0.0), withr
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-GB
Imports: digest, dplyr, DSMolgenisArmadillo, jsonlite, opalr (>=
        3.6.1), purrr, RcppTOML, rmarkdown, rocrateR (>= 0.1.0),
        tibble, vtree, xptr, yaml
Depends: R (>= 4.1.0)
VignetteBuilder: knitr
URL: https://github.com/FederatedMethods/dsROCrate
BugReports: https://github.com/FederatedMethods/dsROCrate/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-01 12:17:56 UTC; robertovillegas-diaz
Author: Roberto Villegas-Diaz [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5036-8661>),
  Becca Wilson [aut] (ORCID: <https://orcid.org/0000-0003-2294-593X>),
  Olly Butters [aut] (ORCID: <https://orcid.org/0000-0003-0354-8461>),
  Stuart Wheater [aut] (ORCID: <https://orcid.org/0009-0003-2419-1964>),
  University of Liverpool [cph]
Maintainer: Roberto Villegas-Diaz <r.villegas-diaz@outlook.com>
Repository: CRAN
Date/Publication: 2026-06-01 12:40:08 UTC
Built: R 4.5.2; ; 2026-06-01 13:41:49 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,safe_call)
S3method(as.data.frame,symbol_registry)
S3method(audit,ArmadilloCredentials)
S3method(audit,character)
S3method(audit,cr8tor)
S3method(audit,default)
S3method(audit,list)
S3method(audit,opal)
S3method(audit,rocrate)
S3method(audit_engine,cr8tor)
S3method(audit_engine,default)
S3method(audit_engine,opal)
S3method(backend_logs,opal)
S3method(backend_options,opal)
S3method(backend_packages,opal)
S3method(backend_profile_exists,opal)
S3method(backend_project,opal)
S3method(backend_project_exists,opal)
S3method(backend_project_perms,opal)
S3method(backend_projects,opal)
S3method(backend_resource_perms,opal)
S3method(backend_resources,opal)
S3method(backend_sys_perms,opal)
S3method(backend_table_perms,opal)
S3method(backend_tables,opal)
S3method(backend_user_exists,opal)
S3method(backend_users,opal)
S3method(check_permissions,ArmadilloCredentials)
S3method(check_permissions,default)
S3method(check_permissions,opal)
S3method(extract_safe_data,opal)
S3method(extract_safe_data,rocrate)
S3method(extract_safe_output,opal)
S3method(extract_safe_output,rocrate)
S3method(extract_safe_people,opal)
S3method(extract_safe_people,rocrate)
S3method(extract_safe_project,opal)
S3method(extract_safe_project,rocrate)
S3method(extract_safe_setting,opal)
S3method(extract_safe_setting,rocrate)
S3method(filter_safe_people,opal)
S3method(flatten_safe_data,default)
S3method(flatten_safe_data,rocrate)
S3method(flatten_safe_output,default)
S3method(flatten_safe_output,rocrate)
S3method(flatten_safe_people,default)
S3method(flatten_safe_people,rocrate)
S3method(flatten_safe_project,default)
S3method(flatten_safe_project,rocrate)
S3method(flatten_safe_setting,default)
S3method(flatten_safe_setting,rocrate)
S3method(init,ArmadilloCredentials)
S3method(init,opal)
S3method(init,rocrate)
S3method(is_admin_con,opal)
S3method(is_audit_con,opal)
S3method(parse_user_profiles,ArmadilloCredentials)
S3method(parse_user_profiles,opal)
S3method(print,cr8tor_bundle)
S3method(print,safe_call)
S3method(print,safe_symbol)
S3method(project_exists,ArmadilloCredentials)
S3method(project_exists,opal)
S3method(report,ArmadilloCredentials)
S3method(report,character)
S3method(report,default)
S3method(report,list)
S3method(report,rocrate)
S3method(safe_call,call)
S3method(safe_call,character)
S3method(safe_data,ArmadilloCredentials)
S3method(safe_data,character)
S3method(safe_data,default)
S3method(safe_data,opal)
S3method(safe_data,rocrate)
S3method(safe_output,ArmadilloCredentials)
S3method(safe_output,character)
S3method(safe_output,default)
S3method(safe_output,opal)
S3method(safe_output,rocrate)
S3method(safe_people,ArmadilloCredentials)
S3method(safe_people,character)
S3method(safe_people,default)
S3method(safe_people,opal)
S3method(safe_people,rocrate)
S3method(safe_project,ArmadilloCredentials)
S3method(safe_project,character)
S3method(safe_project,default)
S3method(safe_project,opal)
S3method(safe_project,rocrate)
S3method(safe_setting,ArmadilloCredentials)
S3method(safe_setting,character)
S3method(safe_setting,cr8tor)
S3method(safe_setting,default)
S3method(safe_setting,opal)
S3method(safe_setting,rocrate)
S3method(validate_backend_version,default)
S3method(validate_backend_version,opal)
S3method(validate_con,default)
S3method(validate_con,opal)
export(armadillo_login)
export(audit)
export(check_permissions)
export(init)
export(report)
export(safe_data)
export(safe_output)
export(safe_people)
export(safe_project)
export(safe_setting)
importFrom(rocrateR,get_entity)
importFrom(utils,write.csv)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# dsROCrate: ‘DataSHIELD’ RO-Crate Governance Functions <img src="man/figures/logo.png" alt="logo" align="right" height="200" style="float:right; height:200px;"/>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/dsROCrate)](https://CRAN.R-project.org/package=dsROCrate)
[![R-CMD-check](https://github.com/FederatedMethods/dsROCrate/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/FederatedMethods/dsROCrate/actions/workflows/R-CMD-check.yaml)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

## Overview

`{dsROCrate}` provides tools to capture, structure and audit DataSHIELD
analyses using the RO-Crate standard. It enables reproducible governance
of federated analyses by packaging inputs, outputs, metadata, and audit
information into a consistent, portable format.

## Key Features

- Create RO-Crates for DataSHIELD workflows
- Capture project metadata, data references and analysis outputs
- Record and audit disclosure control processes
- Support reproducibility and governance in federated environments

## Installation

You can install the released version of `{dsROCrate}` from
[CRAN](https://cran.r-project.org/package=dsROCrate) with:

``` r
install.packages("dsROCrate")
```

And the development version from
[GitHub](https://github.com/FederatedMethods/dsROCrate/) with:

``` r
# install.packages("pak")
pak::pak("FederatedMethods/dsROCrate@dev")
```

## Quick Start

``` r
# open connection
con <- opalr::opal.login(
  username = "administrator",
  password = "password",
  url = "https://opal-demo.obiba.org"
)

# generate RO-Crate with Five Safes components
crate <- con |>
  # initialise RO-Crate
  dsROCrate::init(project = "CNSIM", tables = "CNSIM1", user = "dsuser") |>
  # extract Five Safes components
  dsROCrate::safe_people() |>
  dsROCrate::safe_project() |>
  dsROCrate::safe_data() |>
  dsROCrate::safe_setting() |>
  dsROCrate::safe_output()

# generate report
dsROCrate::report(crate, title = "DataSHIELD Five Safes report")
```

Alternatively, the `dsROCrate::audit()` function can be used to generate
an audit with all the Five Safes components:

``` r
proj_audit <- opalr::opal.login(
  username = "administrator",
  password = "password",
  url = "https://opal-demo.obiba.org"
) |>
  dsROCrate::audit(project = "CNSIM", tables = "CNSIM1", user = "dsuser")

# generate report
dsROCrate::report(proj_audit, title = "DataSHIELD Safe People - Audit Report")
```

## Learn More

For further details, see the following vignette:

``` r
vignette("getting-started", package = "dsROCrate")
```

## Identity

You are welcome to use any of the following hex codes when referencing
`{dsROCrate}`:

[<img src="man/figures/logo.png" alt="logo" align="left" height="150" style="float:left; height:150px;"/>](man/figures/logo.png)
[<img src="man/figures/logo_white.png" alt="logo-white" align="left" height="150" style="float:left; height:150px;"/>](man/figures/logo_white.png)
[<img src="man/figures/logo_black.png" alt="logo-black" align="left" height="150" style="float:left; height:150px;"/>](man/figures/logo_black.png)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  cache = FALSE,
  out.width = "100%",
  cache = FALSE,
  screenshot.opts = list(vwidth = 2000, vheight = 600, zoom = 3, selector = "div.html-widget")
)

# save the built-in output hook
hook_output <- knitr::knit_hooks$get("output")
hook_message <- knitr::knit_hooks$get("message")

# define a truncation helper
truncate_lines_tail <- function(x, lines) {
  x <- unlist(strsplit(x, "\n"))
  more <- "..."
  if (length(lines) == 1) {
    if (length(x) > lines) {
      # truncate the output, but add `more`
      # x <- c(head(x, n = lines), more)
      x <- c(more, tail(x, n = lines))
    }
  } else {
    x <- c(more, x[lines], more)
  }
  paste(c(x, ""), collapse = "\n")
}

# set the new hooks
knitr::knit_hooks$set(
  output = function(x, options) {
    lines <- options$out.lines
    if (is.null(lines)) return(hook_output(x, options))
    x <- truncate_lines_tail(x, lines)
    return(hook_output(x, options))
  },
  message = function(x, options) {
    lines <- options$out.lines
    if (is.null(lines)) return(hook_message(x, options))
    x <- truncate_lines_tail(x, lines)
    return(hook_message(x, options))
  }
)

# helper function to get the number of rows of an RO-Crate
rocrate_lines <- function(rocrate) {
  rocrate_txt(rocrate) |>
    length()
}

# helper function to read RO-Crate as text file
rocrate_txt <- function(rocrate) {
  # create tmp file to write a JSON file with the output
  tmp_file <- tempfile(fileext = ".json")
  on.exit(unlink(tmp_file, force = TRUE))
  # write RO-Crate in tmp file
  rocrateR::write_rocrate(rocrate, tmp_file)
  # read RO-Crate as text file
  readLines(tmp_file)
}

## ----setup--------------------------------------------------------------------
library(dsROCrate)

# show all the lines in the RO-Crate
oopt <- options(max_lines = Inf)

## ----eval = FALSE-------------------------------------------------------------
# vignette("deploy-local-datashield-server-with-opal", package = "dsROCrate")

## -----------------------------------------------------------------------------
# define global variables
## Opal server access
USERNAME <- "administrator"
USERPASS <- "password"
SERVER <- "https://opal-demo.obiba.org"
## Credentials for `dsuser`
### NOTE: this is only used to simulate an analysis and generate logs
DSUSERPASS <- "P@ssw0rd"

## -----------------------------------------------------------------------------
## Five safes variables
PEOPLE <- "dsuser"
PROJECT <- "CNSIM"
TABLES <- c("CNSIM1")

## -----------------------------------------------------------------------------
# login to local server with `USERNAME` and `USERPASS`.
o <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
)

print(o)

## ----eval = FALSE, echo = TRUE------------------------------------------------
# # install.packages("pak")
# pak::pak("rocrateR")
# 
# # for development version use
# pak::pak("ResearchObject/ro-crate-r@dev")

## -----------------------------------------------------------------------------
basic_rocrate <- rocrateR::rocrate_5s()

## -----------------------------------------------------------------------------
print(basic_rocrate)

## ----out.lines=-(rocrate_lines(basic_rocrate) - 2)----------------------------
basic_rocrate <- o |>
  dsROCrate::safe_data(rocrate = basic_rocrate,
                       project = PROJECT,
                       tables = TABLES)

print(basic_rocrate) # note that the output will be truncated

## ----out.lines=-(rocrate_lines(basic_rocrate) - 2)----------------------------
basic_rocrate <- o |>
  dsROCrate::safe_project(rocrate = basic_rocrate,
                          project = PROJECT)

print(basic_rocrate) # note that the output will be truncated

## ----safe_people, out.lines=-(rocrate_lines(basic_rocrate) + 1)---------------
basic_rocrate <- o |>
  dsROCrate::safe_people(rocrate = basic_rocrate, user = PEOPLE)

print(basic_rocrate) # note that the output will be truncated

## ----eval = FALSE-------------------------------------------------------------
# # close previous connection
# opalr::opal.logout(o)
# 
# # open new connection as administrator
# o <- opalr::opal.login(
#   username = "administrator",
#   password = "password",
#   url = SERVER
# )

## ----out.lines=-(rocrate_lines(basic_rocrate) - 2)----------------------------
basic_rocrate <- o |>
  dsROCrate::safe_setting(rocrate = basic_rocrate)

print(basic_rocrate) # note that the output will be truncated

## ----eval = FALSE-------------------------------------------------------------
# # close previous connection
# opalr::opal.logout(o)
# 
# # open new connection as administrator
# o <- opalr::opal.login(
#   username = "administrator",
#   password = "password",
#   url = SERVER
# )

## ----eval = FALSE-------------------------------------------------------------
# pak::pak("DSI")
# pak::pak("DSOpal")
# pak::pak("dsBaseClient")

## ----eval = TRUE--------------------------------------------------------------
# run some test commands with dsBaseClient
## needed to defined the OpalDriver class in the current environment
DSOpal::Opal()
## create new login object, note that here we use the `dsuser`
builder <- DSI::newDSLoginBuilder()
builder$append(server = "study1",
               url = SERVER,
               user = "dsuser",
               password = DSUSERPASS,
               driver = "OpalDriver")
logindata <- builder$build()
conns <- DSI::datashield.login(logins = logindata)

## ----eval = FALSE-------------------------------------------------------------
# ## assign data
# DSI::datashield.assign.table(conns["study1"],
#                              symbol = "dsROCrate_test",
#                              table = paste0(PROJECT, ".", TABLES[1]),
#                              errors.print = TRUE)
# 
# dsBaseClient::ds.ls(datasources = conns["study1"])
# #> $study1
# #> $study1$environment.searched
# #> [1] "R_GlobalEnv"
# #>
# #> $study1$objects.found
# #> [1] "dsROCrate_test"
# dsBaseClient::ds.summary("dsROCrate_test")
# #> $study1
# #> $study1$class
# #> [1] "data.frame"
# #>
# #> $study1$`number of rows`
# #> [1] 2163
# #>
# #> $study1$`number of columns`
# #> [1] 11
# #>
# #> $study1$`variables held`
# #>  [1] "LAB_TSC"            "LAB_TRIG"           "LAB_HDL"
# #>  [4] "LAB_GLUC_ADJUSTED"  "PM_BMI_CONTINUOUS"  "DIS_CVA"
# #>  [7] "MEDI_LPD"           "DIS_DIAB"           "DIS_AMI"
# #> [10] "GENDER"             "PM_BMI_CATEGORICAL"

## ----echo = FALSE, warning = FALSE, message = FALSE, results='hide'-----------
# check if there are any logs available, if not simulate some operations
dsuser_logs_tbl <- opalr::opal.login(
    username = "administrator",
    password = "password",
    url = SERVER
) |> 
  opalr::dsadmin.log() |>
  dplyr::bind_rows(tibble::tibble(username = NA, ds_action = NA)) |>
  dplyr::filter(!is.na(username), username == "dsuser") |>
  dplyr::filter(ds_action %in% c("ASSIGN", "AGGREGATE", "OPEN"))

if (nrow(dsuser_logs_tbl) < 1) {
  ## assign data
  DSI::datashield.assign.table(conns["study1"], 
                               symbol = "dsROCrate_test",
                               table = paste0(PROJECT, ".", TABLES[1]),
                               errors.print = TRUE)
  dsBaseClient::ds.ls(datasources = conns["study1"])
  dsBaseClient::ds.summary("dsROCrate_test")
}

## ----safe_outputs_internal, echo=FALSE----------------------------------------
lines_before_safe_outputs <- rocrate_lines(basic_rocrate)

## ----safe_outputs-------------------------------------------------------------
basic_rocrate <- o |>
  dsROCrate::safe_output(rocrate = basic_rocrate,
                         logs_from = Sys.time() - 8.64E4, # capture the last 24 hours
                         logs_to = Sys.time())

## ----out.lines=-(lines_before_safe_outputs + 6)-------------------------------
print(basic_rocrate) # note that the output will be truncated

## -----------------------------------------------------------------------------
opalr::opal.logout(o)

## -----------------------------------------------------------------------------
# create temp directory (only for demo purposes)
tmp_path_bag <- file.path(tempdir(), "dsROCrate-getting-started")
dir.create(tmp_path_bag, showWarnings = FALSE)

# create RO-Crate bag
path_to_rocrate_bag <- basic_rocrate |>
  rocrateR::bag_rocrate(path = tmp_path_bag, overwrite = TRUE)

## -----------------------------------------------------------------------------
# extract files in temporary directory
path_to_rocrate_bag |>
  # extract contents inside /tmp_path_bag/ROC
  rocrateR::unbag_rocrate(output = file.path(tmp_path_bag, "ROC"), quiet = TRUE) |>
  # create tree with the files
  fs::dir_tree()

## -----------------------------------------------------------------------------
unlink(tmp_path_bag, recursive = TRUE, force = TRUE)

## ----echo = FALSE-------------------------------------------------------------
options(oopt)
oopt <- options(max_lines = 40)

## ----warning=FALSE------------------------------------------------------------
safe_people_crate_v1 <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
) |>
  dsROCrate::audit(user = "dsuser", project = "CNSIM")

print(safe_people_crate_v1)

## ----safe_people_crate_audit_v1-----------------------------------------------
safe_people_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_people_crate_contents <- safe_people_crate_v1 |>
  dsROCrate::report(filepath = safe_people_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_people_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_people_crate_contents$overview_data |>
  knitr::kable()

## ----eval = FALSE-------------------------------------------------------------
# safe_people_crate_v1 |>
#   dsROCrate::report(filepath = safe_people_crate_v1_rmd,
#                             title = "DataSHIELD Safe People - Audit Report",
#                             render = TRUE,
#                             overwrite = TRUE)

## ----warning=FALSE------------------------------------------------------------
safe_project_crate_v1 <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
) |>
  dsROCrate::audit(project = "CNSIM")

print(safe_project_crate_v1)

## ----safe_project_crate_audit_v1----------------------------------------------
safe_project_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_project_crate_contents <- safe_project_crate_v1 |>
  dsROCrate::report(filepath = safe_project_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_project_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_project_crate_contents$overview_data |>
  knitr::kable()

## ----eval = FALSE-------------------------------------------------------------
# safe_project_crate_v1 |>
#   dsROCrate::report(filepath = safe_project_crate_v1_rmd,
#                             title = "DataSHIELD Safe Project - Audit Report",
#                             render = TRUE,
#                             overwrite = TRUE)

## ----warning=FALSE------------------------------------------------------------
study_crate_v1 <- 
  list(
    "opal_test" = opalr::opal.login(
      username = USERNAME,
      password = USERPASS,
      url = "https://opal-demo.obiba.org"
    ),
    "opal_demo" = opalr::opal.login(
      username = USERNAME,
      password = USERPASS,
      url = "https://opal-demo.obiba.org"
    )
  ) |>
  dsROCrate::audit(project = "CNSIM")

print(study_crate_v1)

## ----study_crate_audit_v1-----------------------------------------------------
study_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_project_crate_contents <- study_crate_v1 |>
  dsROCrate::report(filepath = study_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_project_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_project_crate_contents$overview_data |>
  knitr::kable()

## ----eval = FALSE-------------------------------------------------------------
# study_crate_v1 |>
#   dsROCrate::report(filepath = study_crate_v1_rmd,
#                             title = "DataSHIELD Study audit",
#                             render = TRUE,
#                             overwrite = TRUE)

## ----echo = FALSE, message = FALSE, error = FALSE-----------------------------
unlink(safe_people_crate_v1_rmd, TRUE, TRUE)
unlink(safe_project_crate_v1_rmd, TRUE, TRUE)
unlink(study_crate_v1_rmd, TRUE, TRUE)
# reverse options
options(oopt)

#' Login to a MOLGENIS' Armadillo server
#'
#' Login to a MOLGENIS' Armadillo server. Wrapper for the function
#' [DSMolgenisArmadillo::armadillo.get_credentials()].
#'
#' @inheritParams DSMolgenisArmadillo::armadillo.get_credentials
#'
#' @returns MOLGENIS' Armadillo connection object.
#' @export
armadillo_login <- function(server) {
  conn <- DSMolgenisArmadillo::armadillo.get_credentials(server)
  ns <- loadNamespace("MolgenisArmadillo")
  pkgenv <- get(".pkgglobalenv", envir = ns)
  assign("armadillo_url", server, envir = pkgenv)
  assign("auth_token", getElement(conn, "access_token"), envir = pkgenv)
  return(invisible(conn))
}

# -------------------------------------------------------------------------
# basic behaviour (unit tests with mocking)
# -------------------------------------------------------------------------
test_that("armadillo_login returns invisible connection object", {
  skip_if_not_installed("MolgenisArmadillo")

  mock_conn <- list(access_token = "fake_token_123")

  with_mocked_bindings(
    armadillo.get_credentials = function(server) {
      expect_equal(server, "https://armadillo.example.org")
      mock_conn
    },
    {
      result <- armadillo_login("https://armadillo.example.org")
    },
    .package = "DSMolgenisArmadillo"
  )

  # should return the connection invisibly
  expect_equal(result, mock_conn)
})

test_that("armadillo_login calls armadillo.get_credentials with correct server", {
  skip_if_not_installed("MolgenisArmadillo")

  called_server <- NULL

  with_mocked_bindings(
    armadillo.get_credentials = function(server) {
      called_server <<- server
      list(access_token = "abc")
    },
    {
      armadillo_login("https://my-armadillo.org")
    },
    .package = "DSMolgenisArmadillo"
  )

  expect_equal(called_server, "https://my-armadillo.org")
})

# -------------------------------------------------------------------------
# Side-effect tests: namespace global assignments
# -------------------------------------------------------------------------

test_that("armadillo_login assigns armadillo_url in MolgenisArmadillo namespace", {
  skip_if_not_installed("MolgenisArmadillo")

  pkgenv <- get_pkgenv()
  original_url <- if (exists("armadillo_url", envir = pkgenv)) {
    get("armadillo_url", envir = pkgenv)
  } else {
    NULL
  }

  on.exit(
    {
      # restore original state to avoid test pollution
      if (is.null(original_url)) {
        if (exists("armadillo_url", envir = pkgenv)) {
          rm("armadillo_url", envir = pkgenv)
        }
      } else {
        assign("armadillo_url", original_url, envir = pkgenv)
      }
    },
    add = TRUE
  )

  with_mocked_bindings(
    armadillo.get_credentials = function(server) {
      list(access_token = "token_xyz")
    },
    {
      armadillo_login("https://armadillo.test")
    },
    .package = "DSMolgenisArmadillo"
  )

  expect_true(exists("armadillo_url", envir = pkgenv))
  expect_equal(get("armadillo_url", envir = pkgenv), "https://armadillo.test")
})

test_that("armadillo_login assigns auth_token from connection access_token", {
  skip_if_not_installed("MolgenisArmadillo")

  pkgenv <- get_pkgenv()
  original_token <- if (exists("auth_token", envir = pkgenv)) {
    get("auth_token", envir = pkgenv)
  } else {
    NULL
  }

  on.exit(
    {
      # restore original state
      if (is.null(original_token)) {
        if (exists("auth_token", envir = pkgenv)) {
          rm("auth_token", envir = pkgenv)
        }
      } else {
        assign("auth_token", original_token, envir = pkgenv)
      }
    },
    add = TRUE
  )

  with_mocked_bindings(
    armadillo.get_credentials = function(server) {
      list(access_token = "secure_token_456")
    },
    {
      armadillo_login("https://armadillo.secure")
    },
    .package = "DSMolgenisArmadillo"
  )

  expect_true(exists("auth_token", envir = pkgenv))
  expect_equal(get("auth_token", envir = pkgenv), "secure_token_456")
})

# -------------------------------------------------------------------------
# Edge cases
# -------------------------------------------------------------------------
test_that("armadillo_login handles missing access_token gracefully", {
  skip_if_not_installed("MolgenisArmadillo")

  pkgenv <- get_pkgenv()

  with_mocked_bindings(
    armadillo.get_credentials = function(server) {
      list() # no access_token
    },
    {
      expect_error(
        armadillo_login("https://armadillo.no-token"),
        regexp = NA
      )
    },
    .package = "DSMolgenisArmadillo"
  )
})

test_that("armadillo_login propagates credential retrieval errors", {
  skip_if_not_installed("MolgenisArmadillo")

  with_mocked_bindings(
    armadillo.get_credentials = function(server) {
      stop("Authentication failed")
    },
    {
      expect_error(
        armadillo_login("https://armadillo.fail"),
        "Authentication failed"
      )
    },
    .package = "DSMolgenisArmadillo"
  )
})

# -------------------------------------------------------------------------
# Integration-style test (optional, only if real server configured)
# -------------------------------------------------------------------------
test_that("armadillo_login works with real Armadillo server (optional)", {
  server <- Sys.getenv("ARMADILLO_SERVER")
  skip_if(server == "", "No ARMADILLO_SERVER env var set")

  expect_no_error({
    conn <- armadillo_login(server)
  })

  expect_true(is.list(conn) || is.environment(conn))
})

#' Load `content` from external files
#'
#' @param rocrate Object with the [rocrate][rocrateR::rocrate()] class.
#' @param roc_path String with path to the root of the RO-Crate.
#'
#' @returns Update `rocrate` object.
#' @keywords internal
load_content <- function(rocrate, roc_path) {
  # get 'File' entities with missing `content` (if any)
  file_ents <- rocrate |>
    .get_entity(type = "File") |>
    Filter(f = function(x) is.null(getElement(x, "content")))
  ## attempt loading contents, if any entities were found
  for (ent in file_ents) {
    # attempt loading the contents from the filenames given by `@id`
    content <- tryCatch(
      {
        if (getElement(ent, "encodingFormat") == "text/csv") {
          utils::read.csv(file.path(roc_path, getElement(ent, "@id")))
        } else {
          readLines(file.path(roc_path, getElement(ent, "@id")))
        }
      },
      error = function(e) {
        NULL
      },
      warning = function(e) {
        NULL
      }
    )

    # update entity within the RO-Crate
    if (!is.null(content)) {
      rocrate <- rocrate |>
        rocrateR::add_entity_value(
          id = getElement(ent, "@id"),
          key = "content",
          value = list(content)
        )
    }
  }

  return(rocrate)
}

test_that("CSV file content is loaded correctly", {
  skip_if_not_installed("rocrateR")

  # create temporary directory
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))

  # create a CSV file
  csv_path <- file.path(tmp_dir, "data.csv")
  write.csv(data.frame(a = 1:3, b = letters[1:3]), csv_path, row.names = FALSE)

  # create minimal RO-Crate
  roc <- rocrateR::rocrate_5s()

  # Add File entity without content
  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "data.csv",
      type = "File",
      encodingFormat = "text/csv"
    )
  )

  # Run function
  roc2 <- load_content(roc, tmp_dir)

  # Retrieve entity
  ent <- rocrateR::get_entity(roc2, id = "data.csv")[[1]]

  expect_false(is.null(ent$content))
  expect_s3_class(ent$content[[1]], "data.frame")
  expect_equal(nrow(ent$content[[1]]), 3)
})

test_that("Text file content is loaded correctly", {
  skip_if_not_installed("rocrateR")

  # create temporary directory
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))

  # create text file
  txt_path <- file.path(tmp_dir, "notes.txt")
  writeLines(c("line1", "line2"), txt_path)

  # create minimal RO-Crate
  roc <- rocrateR::rocrate_5s()

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "notes.txt",
      type = "File",
      encodingFormat = "text/plain"
    )
  )

  roc2 <- load_content(roc, tmp_dir)

  ent <- rocrateR::get_entity(roc2, id = "notes.txt")[[1]]

  expect_false(is.null(ent$content))
  expect_equal(ent$content[[1]], c("line1", "line2"))
})

test_that("Entities with existing content are not modified", {
  skip_if_not_installed("rocrateR")

  # create temporary directory
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))

  # create minimal RO-Crate
  roc <- rocrateR::rocrate_5s()

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "existing.txt",
      type = "File",
      encodingFormat = "text/plain",
      content = list("already here")
    )
  )

  roc2 <- load_content(roc, tmp_dir)

  ent <- rocrateR::get_entity(roc2, id = "existing.txt")[[1]]

  expect_equal(ent$content[[1]], "already here")
})

test_that("Missing files do not break function and are skipped", {
  skip_if_not_installed("rocrateR")

  # create temporary directory
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))

  # create minimal RO-Crate
  roc <- rocrateR::rocrate_5s()

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "missing.txt",
      type = "File",
      encodingFormat = "text/plain"
    )
  )

  expect_no_error({
    roc2 <- load_content(roc, tmp_dir)
  })

  ent <- rocrateR::get_entity(roc2, id = "missing.txt")[[1]]

  expect_true(is.null(ent$content))
})

test_that("load_content works end-to-end with real RO-Crate structure", {
  skip_if_not_installed("rocrateR")

  # create temporary directory
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))

  # create files inside RO-Crate
  writeLines("hello world", file.path(tmp_dir, "file1.txt"))
  write.csv(
    data.frame(x = 42),
    file.path(tmp_dir, "file2.csv"),
    row.names = FALSE
  )

  # create minimal RO-Crate
  roc <- rocrateR::rocrate_5s()

  # add both files as entities
  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "file1.txt",
      type = "File",
      encodingFormat = "text/plain"
    )
  )

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "file2.csv",
      type = "File",
      encodingFormat = "text/csv"
    )
  )

  # run function
  roc2 <- load_content(roc, tmp_dir)

  ent1 <- rocrateR::get_entity(roc2, id = "file1.txt")[[1]]
  ent2 <- rocrateR::get_entity(roc2, id = "file2.csv")[[1]]

  expect_equal(ent1$content[[1]], "hello world")
  expect_s3_class(ent2$content[[1]], "data.frame")
  expect_equal(ent2$content[[1]]$x, 42)
})

#' Create an audit RO-Crate
#'
#' Create an audit RO-Crate following the 5 Safes Principles.
#'
#' This function handles various audit types, which will be dispatched based on
#' the input object. If the input object is
#'
#' \itemize{
#'     \item a _connection_ to a DataSHIELD server (e.g., OBiBa's Opal):
#'.    generates an RO-Crate object with deployment details, including outputs.
#'     \item a _path_ pointing to
#'     \itemize{
#'         \item **a `cr8tor` archive / governance bundle**: generates an
#'         RO-Crate object with pre-deployment governance details.
#'         \item **an RO-Crate object**: generates an RO-Crate object with
#'         clearly defined 5 Safes elements.
#'     }
#'     \item an _RO-Crate_ object: generates an RO-Crate object with clearly
#'     defined 5 Safes elements.
#' }
#' @param x Object to be audited. This can be
#'     \itemize{
#'         \item a _connection_ to a DataSHIELD server (e.g., OBiBa's Opal).
#'         \item a _path_ pointing to an RO-Crate OR a `cr8tor` archive /
#'.        governance bundle.
#'         \item an _RO-Crate_ object.
#'     }
#'     Alternatively, a list of any of the above.
#' @param ... Additional arguments.
#' @param intent Additional object with governance bundle/specification of the
#'     intent of a project. It takes the same types as `x`.
#' @inheritParams audit_engine
#'
#' @returns RO-Crate with audit details.
#' @export
audit <- function(x, ...) {
  UseMethod("audit")
}

#' @rdname audit
#' @export
audit.ArmadilloCredentials <- function(x, ..., intent = NULL) {
  stop(
    "The audit for Armadillo backend is not currently implemented!",
    call. = FALSE
  )
}

#' @rdname audit
#' @export
audit.character <- function(x, ..., intent = NULL) {
  # verify if the given path exists, if not, return an error message
  if (!file.exists(x)) {
    stop("The given file does not exist!", call. = FALSE)
  }

  # attempt loading a `cr8tor` bundle
  x_obj <- tryCatch(
    load_cr8tor_bundle(x, ...),
    error = function(e) NULL
  )
  # alternatively, attempt loading an RO-Crate
  if (is.null(x_obj)) {
    x_obj <- tryCatch(
      rocrateR::load_rocrate(x, ...),
      error = function(e) NULL
    )
  }

  if (is.null(x_obj)) {
    stop(
      "The given path does not point to a valid `cr8tor` archive nor an `rocrate",
      call. = FALSE
    )
  }

  # attempt auditing intent
  intent_lst <- audit_intent(intent, ...)

  # call next method
  main_audit <- audit(x_obj, intent_lst$main_audit_args)

  # return list
  if (is.null(intent_lst$intent_audit)) {
    return(main_audit)
  }
  list(intent = intent_lst$intent_audit, deployment = main_audit)
}

#' @rdname audit
#' @export
audit.cr8tor <- function(x, ..., intent = NULL) {
  # attempt auditing intent
  intent_lst <- audit_intent(intent, ..., excluded_args = "path")

  # call next method
  main_audit <- audit_engine(x, intent_lst$main_audit_args)

  # return list
  if (is.null(intent_lst$intent_audit)) {
    return(main_audit)
  }
  list(intent = intent_lst$intent_audit, deployment = main_audit)
}

#' @rdname audit
#' @export
audit.list <- function(x, ..., intent = NULL) {
  purrr::map(x, audit, ..., intent = intent)
}

#' @rdname audit
#' @export
audit.opal <- function(
  x,
  ...,
  intent = NULL,
  project = NULL,
  user = NULL,
  logs_from = -Inf,
  logs_to = Inf,
  path = NULL
) {
  # attempt auditing intent
  intent_lst <- audit_intent(intent, ...)

  # call next method
  main_audit <- audit_engine(
    x,
    project = c(intent_lst$main_audit_args$project, project),
    user = c(intent_lst$main_audit_args$user, user),
    logs_from = logs_from,
    logs_to = logs_to,
    path = path,
    intent_lst$main_audit_args
  )

  # return list
  if (is.null(intent_lst$intent_audit)) {
    return(main_audit)
  }
  list(intent = intent_lst$intent_audit, deployment = main_audit)
}

#' @rdname audit
#' @export
audit.rocrate <- function(x, ..., intent = NULL) {
  # return input rocrate object
  x
}

#' @export
print.cr8tor_bundle <- function(x, ...) {
  msg <- "<cr8tor governance bundle>"

  is_valid_roc <- function(x) {
    inherits(x$rocrate, "rocrate") &&
      rocrateR::is_rocrate(x$rocrate, error = FALSE)
  }

  if (is_valid_roc(x)) {
    msg <- c(msg, " \U2714 Valid RO-Crate")
  } else {
    msg <- c(msg, " \U2716 Invalid RO-Crate")
  }

  # project metadata (safe extraction)
  proj <- tryCatch(
    x$resources[["governance/cr8-governance.yaml"]]$project,
    error = function(e) NULL
  )

  if (!is.null(proj)) {
    msg <- c(
      msg,
      paste0(" Project: ", proj$name, " (", proj$reference, ")")
    )
  }

  # counts
  n_resources <- length(x$resources)
  n_entities <- length(x$rocrate$`@graph` %||% list())
  n_actions <- length(proj$actions %||% list())
  n_users <- length(
    x$resources[["governance/cr8-governance.yaml"]]$users %||% list()
  )

  msg <- c(
    msg,
    " Contents:",
    paste0("  - Entities:  ", n_entities),
    paste0("  - Resources: ", n_resources),
    paste0("  - Actions:   ", n_actions),
    paste0("  - Users:     ", n_users)
  )

  message(paste0(msg, collapse = "\n"))

  invisible(x)
}

#' Validate POSIXct string
#'
#' @param x POSIXct string.
#' @param tz String with time zone,
#'
#' @returns Boolean value to indicate if the given string is a valid POSIXct
#' string.
#' @keywords internal
is_valid_posixct <- function(x, tz = "UTC") {
  tryCatch(
    {
      !is.na(as.POSIXct(
        x,
        tz = tz,
        tryFormats = c(
          "%Y-%m-%d %H:%M:%S",
          "%Y/%m/%d %H:%M:%S"
        )
      ))
    },
    error = function(e) {
      return(FALSE)
    }
  )
}

#' Unfill vector
#'
#' @param x Input vector.
#' @param val Default value.
#'
#' @returns Unfilled vector.
#' @keywords internal
#' @source https://github.com/tidyverse/tidyr/issues/250#issuecomment-344984802
unfill_vec <- function(x, val = "") {
  same <- x == dplyr::lag(x)
  ifelse(!is.na(same) & same, val, x)
}

#' Fill vector
#'
#' @param x Input vector.
#' @param val Default value.
#'
#' @returns Filled vector.
#' @keywords internal
refill_vec <- function(x, val = "") {
  for (i in seq_along(x)) {
    if (i > 1 && (is.na(x[i]) || x[i] == val)) {
      x[i] <- x[i - 1]
    }
  }
  x
}

test_that("unfill_vec replaces repeated values with default val", {
  x <- c("a", "a", "b", "b", "b", "c")

  expect_equal(
    unfill_vec(x),
    c("a", "", "b", "", "", "c")
  )
})

test_that("unfill_vec respects custom val", {
  x <- c("x", "x", "y", "y")

  expect_equal(
    unfill_vec(x, val = NA_character_),
    c("x", NA, "y", NA)
  )
})

test_that("unfill_vec handles NA values correctly", {
  x <- c("a", NA, NA, "b", "b")

  expect_equal(
    unfill_vec(x),
    c("a", NA, NA, "b", "")
  )
})

test_that("unfill_vec works with length-1 vectors", {
  expect_equal(
    unfill_vec("a"),
    "a"
  )
})


test_that("refill_vec fills down missing or default values", {
  x <- c("a", "", "", "b", "", "c")

  expect_equal(
    refill_vec(x),
    c("a", "a", "a", "b", "b", "c")
  )
})

test_that("refill_vec fills down NA values", {
  x <- c("a", NA, NA, "b", NA)

  expect_equal(
    refill_vec(x),
    c("a", "a", "a", "b", "b")
  )
})

test_that("refill_vec respects custom val", {
  x <- c("a", "-", "-", "b", "-")

  expect_equal(
    refill_vec(x, val = "-"),
    c("a", "a", "a", "b", "b")
  )
})

test_that("refill_vec leaves first element unchanged", {
  x <- c("", "a", "")

  expect_equal(
    refill_vec(x),
    c("", "a", "a")
  )
})

test_that("refill_vec works with length-1 vectors", {
  expect_equal(
    refill_vec("a"),
    "a"
  )

  expect_equal(
    refill_vec(NA_character_),
    NA_character_
  )
})

#' Safe People details
#'
#' Safe People details for the RO-Crate.
#'
#' Researchers must be accredited and trained before accessing the data that
#' has been prepared for them.
#'
#' The access service provider may require the researcher to sign a statement
#' that they know and understand the regulations of the TRE.
#'
#' @param ... Other optional arguments. See the full documentation,
#'     [`?dsROCrate::safe_people`][safe_people()].
#' @param user_id_suffix String with ID suffix for the tables/datasets
#'     entities in the RO-Crate (default: `"#dataset:"`).
#' @param set_author Boolean flag to indicate if the current user should be
#'     set as the author of the RO-Crate.
#' @param set_project Boolean flag to indicate if any `Project` entities found
#'     in `x` should be linked to the Safe People entity.
#' @inheritParams init
#'
#' @returns Updated RO-Crate object with Safe People information.
#' @export
#'
#' @source
#' \itemize{
#'  \item Research Data Scotland, 2025. "What is the Five Safes framework?".
#'  <https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/>
#' }
safe_people <- function(x, ...) {
  UseMethod("safe_people")
}

#' @export
safe_people.default <- function(x, ...) {
  stop(
    "Unknown class, please try either a file path or",
    " an object with `rocrate` class!"
  )
}

#' @rdname safe_people
#' @export
safe_people.character <- function(
  x,
  ...,
  user = attr(x, "user"),
  user_id_suffix = "#person:",
  set_author = TRUE,
  set_project = TRUE,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables")
) {
  # attempt loading the RO-Crate
  rocrate <- rocrateR::load_rocrate(x)

  # call method with given `rocrate` object:
  safe_people(
    rocrate,
    connection = connection,
    user = user,
    user_id_suffix = user_id_suffix,
    set_author = set_author,
    set_project = set_project,
    path = path,
    profile = profile,
    project = project,
    resources = resources,
    tables = tables
  )
}

#' @rdname safe_people
#' @export
safe_people.opal <- function(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  user = NULL,
  user_id_suffix = "#person:",
  set_author = TRUE,
  set_project = TRUE,
  path = NULL,
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL
) {
  # x is a valid opal connection object
  validate_opal_con(x)

  # attempt to retrieve project entity
  project_id <- id_hash("#project:", project)
  safe_project_entity <- rocrate |>
    .get_entity(id = project_id, type = "Project")

  # initialise empty user entity
  user_entity <- NULL

  # check if `user` was given
  if (!is.null(user)) {
    # check if user is a list, if so, then use all the elements as part of
    # the user_entity
    if ("list" %in% class(user)) {
      user_entity <- rocrateR::entity(
        id = c(getElement(user, "@id"), getElement(user, "id")),
        type = "Person",
        name = c(getElement(user, "name"), getElement(user, "username")),
        givenName = c(
          getElement(user, "givenName"),
          getElement(user, "firstname")
        ),
        familyName = c(
          getElement(user, "familyName"),
          getElement(user, "surname")
        ),
        affiliation = list(`@id` = c(getElement(user, "affiliation")))
      )
    } else {
      user_entity <- rocrateR::entity(
        id = paste0(user_id_suffix, digest::digest(user)),
        type = "Person",
        name = user
      )
    }
  } else {
    # create basic user entity
    user_entity <- rocrateR::entity(
      id = paste0(user_id_suffix, digest::digest(x$username)),
      type = "Person",
      name = x$username
    )
    user <- x$username
  }

  # extract user information from the connection object
  user_profile_tbl <- parse_user_profiles(x, user = user)

  # check if the `user_profile_tbl` has a `userInfo` column, if so, then
  # attach the fields in this one to the user entity
  if ("userInfo" %in% colnames(user_profile_tbl)) {
    user_info_tbl <- getElement(user_profile_tbl, "userInfo") |>
      dplyr::bind_rows()
    user_info_cols <- colnames(user_info_tbl)
    for (i in seq_along(user_info_cols)) {
      # avoid overwriting existing fields
      if (!(user_info_cols[i] %in% names(user_entity))) {
        user_entity[user_info_cols[i]] <- user_info_tbl[, i]
      }
    }
  }

  # add membership information, if Safe Project was found and set_project = TRUE
  if (length(safe_project_entity) && set_project) {
    user_entity$memberOf <- safe_project_entity |>
      lapply(\(x) list(`@id` = x[["@id"]]))
  }

  # add user to the RO-Crate
  rocrate <- rocrate |>
    rocrateR::add_entity(user_entity, overwrite = TRUE)

  # link new user entity @id to the root (./) author property
  if (set_author) {
    rocrate <- rocrate |>
      rocrateR::add_entity_value(
        id = "./",
        key = "author",
        value = list(`@id` = getElement(user_entity, "@id")),
        overwrite = TRUE
      )
  }

  # attach input arguments as attributes
  attr(rocrate, "connection") <- x
  attr(rocrate, "path") <- path
  attr(rocrate, "profile") <- profile
  attr(rocrate, "project") <- project
  attr(rocrate, "resources") <- resources
  attr(rocrate, "tables") <- tables
  attr(rocrate, "user") <- user

  # return RO-Crate with the new entity
  return(rocrate)
}

#' @rdname safe_people
#' @export
safe_people.rocrate <- function(
  x,
  ...,
  user = attr(x, "user"),
  user_id_suffix = "#person:",
  set_author = TRUE,
  set_project = TRUE,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables")
) {
  # check if the connection was given
  if (is.null(connection)) {
    stop("A `connection` object is required!", call. = FALSE)
  }

  # call method with given `connection` object:
  safe_people(
    connection,
    rocrate = x,
    user = user,
    user_id_suffix = user_id_suffix,
    set_author = set_author,
    set_project = set_project,
    path = path,
    profile = profile,
    project = project,
    resources = resources,
    tables = tables
  )
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
# - - - - - - - - - - - - - - - - - ERRORS - - - - - - - - - - - - - - - - - #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
test_that("safe_people fails when attempt calling function with invalid class", {
  expect_error(
    dsROCrate::safe_people(structure(list(), class = "InvalidClass"))
  )
})

test_that("safe_people fails when attempt calling with invalid connection", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  expect_error(
    basic_rocrate |>
      dsROCrate::safe_people(user = attr(opal_con, "PEOPLE"), connection = NULL)
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
# - - - - - - - - - - - - - - - - - VALID  - - - - - - - - - - - - - - - - - #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
test_that("safe_people works", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  # add project entity to the RO-Crate
  basic_rocrate_2 <- basic_rocrate |>
    rocrateR::add_entity(rocrateR::entity("Fake Project", type = "Project"))

  # add default Safe People details (user logged in)
  basic_rocrate_3 <- basic_rocrate_2 |>
    dsROCrate::safe_people(connection = opal_con)
  ## retrieve safe_people entity
  basic_rocrate_2_sp <- basic_rocrate_3 |>
    rocrateR::get_entity(type = "Person")
  expect_equal(length(basic_rocrate_2_sp), 1)

  # add custom Safe People
  basic_rocrate_4 <- basic_rocrate_2 |>
    dsROCrate::safe_people(
      connection = opal_con,
      user = list(id = "test_user", name = "Test User")
    )
  ## retrieve safe_people entity
  basic_rocrate_4_sp <- basic_rocrate_4 |>
    rocrateR::get_entity(type = "Person")
  expect_equal(length(basic_rocrate_4_sp), 1)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
# - - - - - - - - - - - - - - - - - ERRORS - - - - - - - - - - - - - - - - - #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
test_that("safe_output fails when calling with invalid object", {
  expect_error(
    dsROCrate::safe_output(structure(list(), class = "InvalidClass"))
  )
})

test_that("safe_output fails when attempt calling with invalid connection", {
  # setup
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  expect_error(
    basic_rocrate |>
      dsROCrate::safe_output(connection = NULL)
  )
})

test_that("safe_output warns when calling with RO-Crate without Safe People details", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  expect_warning(
    basic_rocrate |>
      dsROCrate::safe_output(connection = opal_con)
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("safe_output warns when attempt setting multiple users as the RO-Crate authors", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  expect_error(
    basic_rocrate |>
      dsROCrate::safe_project(
        connection = opal_con,
        project = attr(opal_con, "PROJECT")
      ) |>
      dsROCrate::safe_people(user = 1) |>
      dsROCrate::safe_people(user = 2) |>
      rocrateR::add_entity_value(
        id = "./",
        key = "author",
        value = lapply(
          c(
            "#person:6717f2823d3202449301145073ab8719",
            "#person:db8e490a925a60e62212cefc7674ca02"
          ),
          \(x) list(`@id` = x)
        )
      ) |>
      dsROCrate::safe_output(connection = opal_con, user = NULL)
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
# - - - - - - - - - - - - - - - - - VALID  - - - - - - - - - - - - - - - - - #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
test_that("safe_output works", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  # add Safe People details
  ## add project entity to the RO-Crate
  basic_rocrate_2 <- basic_rocrate |>
    rocrateR::add_entity(rocrateR::entity("Fake Project", type = "Project"))

  ## add Safe People details for
  basic_rocrate_3 <- basic_rocrate_2 |>
    dsROCrate::safe_people(
      connection = opal_con,
      user = attr(opal_con, "PEOPLE")
    )

  # call the function with an RO-Crate with multiple users
  basic_rocrate_4 <- basic_rocrate_3 |>
    # add another username
    dsROCrate::safe_people(
      connection = opal_con,
      user = list(id = "extra_username", name = "Extra username"),
      set_author = FALSE
    ) |>
    # set multiple users as the author
    rocrateR::add_entity_value(
      id = "./",
      key = "author",
      value = list(
        list(`@id` = "extra_username"),
        list(`@id` = paste0("#person:", digest::digest("dsuser")))
      ),
      overwrite = TRUE
    )

  # test for warning about multiple authors in the root entity
  expect_warning(
    basic_rocrate_4 |>
      dsROCrate::safe_output(connection = opal_con)
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)

  # simulate DataSHIELD operations
  ## run some test commands with dsBaseClient
  ### needed to defined the OpalDriver class in the current environment
  DSOpal::Opal()
  ### create new login object, note that here we use the `dsuser`
  builder <- DSI::newDSLoginBuilder()
  builder$append(
    server = "study1",
    url = attr(opal_con, "SERVER"),
    user = "dsuser",
    password = attr(opal_con, "DSUSERPASS"),
    driver = "OpalDriver"
  )
  logindata <- builder$build()
  conns <- DSI::datashield.login(logins = logindata)

  ### assign data
  DSI::datashield.assign.table(
    conns["study1"],
    symbol = "dsROCrate_test",
    table = paste0(attr(opal_con, "PROJECT"), ".", attr(opal_con, "TABLES")[1]),
    errors.print = TRUE
  )

  dsBaseClient::ds.ls(datasources = conns["study1"])

  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  expect_warning(
    basic_rocrate_5 <- basic_rocrate_3 |>
      dsROCrate::safe_output(
        connection = opal_con,
        logs_from = Sys.time() - 60, # capture the last min
        logs_to = Sys.time(),
        user = "dsuser"
      )
  )

  # run function with logs for an invalid period (no logs)
  expect_warning(
    basic_rocrate_6 <- basic_rocrate_3 |>
      dsROCrate::safe_output(
        connection = opal_con,
        logs_from = Sys.time() - 60^2 * 24, # 1 day ago
        logs_to = Sys.time() - 60^2 * 23,
        user = "dsuser"
      )
  )

  # provide invalid path to write the logs
  expect_warning(
    basic_rocrate_7 <- basic_rocrate_3 |>
      dsROCrate::safe_output(
        connection = opal_con,
        logs_from = Sys.time() - 60^2 * 24, # 1 day ago
        logs_to = Sys.time(),
        user = "dsuser",
        path = "/invalid/path/rocrateR"
      )
  )

  # write logs in temporary file
  tempdir_name <- tempdir()
  on.exit(unlink(tempdir_name, force = TRUE, recursive = TRUE))
  expect_true(dir.exists(tempdir_name))
  ## ignore warnings about missing logs
  suppressWarnings(
    basic_rocrate_8 <- basic_rocrate_3 |>
      dsROCrate::safe_output(
        connection = opal_con,
        logs_from = Sys.time() - 60, # capture the last min
        logs_to = Sys.time(),
        user = "dsuser",
        path = tempdir_name
      )
  )
  unlink(tempdir_name, force = TRUE, recursive = TRUE)
  expect_false(dir.exists(tempdir_name))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

#' Safe Setting details
#'
#' Safe Setting details for the RO-Crate.
#'
#' The organisational and technical settings used to access data are designed
#' to minimise the risk of accidental disclosure of data.
#'
#' These settings also prevent the deliberate disclosure of data to others.
#'
#' Physical settings for data access can include locations like
#' [SafePods](https://safepodnetwork.ac.uk/) – secured rooms that use controlled
#' door access, CCTV and secure technology to ensure that sensitive data cannot
#' be mishandled or removed from the Safe Setting. Researchers can analyse the
#' data in these secure rooms, but do not have access to the internet, external
#' devices (such as printers), or any other way of removing protected data from
#' the space.
#'
#' Digital Safe Settings provide secure access to data from a remote location.
#' To be approved for remote data access, researchers will need to prove that
#' their organisation meets physical and IT security standards.
#'
#' @param ... Other optional arguments. See the full documentation,
#'     [`?dsROCrate::safe_setting`][safe_setting()].
#' @inheritParams init
#'
#' @returns Updated RO-Crate object with Safe Settings information.
#' @export
#'
#' @source
#' \itemize{
#'  \item Research Data Scotland, 2025. "What is the Five Safes framework?".
#'  <https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/>
#' }
safe_setting <- function(x, ...) {
  UseMethod("safe_setting")
}

#' @export
safe_setting.default <- function(x, ...) {
  stop(
    "Unknown class, please try either a file path or",
    " an object with `rocrate` class!"
  )
}

#' @rdname safe_setting
#' @export
safe_setting.character <- function(
  x,
  ...,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  tables = attr(x, "tables"),
  resources = attr(x, "resources"),
  user = attr(x, "user")
) {
  # attempt loading the RO-Crate
  rocrate <- rocrateR::load_rocrate(x)

  # call method with given `rocrate` object:
  safe_setting(
    rocrate,
    connection = connection,
    path = path,
    profile = profile,
    project = project,
    resources = resources,
    tables = tables,
    user = user
  )
}

#' @export
#' @rdname safe_setting
safe_setting.cr8tor <- function(x, ..., rocrate = rocrateR::rocrate_5s()) {
  # x: parsed CR8TOR resources list

  # technical infrastructure controls ----
  tech_controls <- list(
    rocrateR::entity(
      id = "#control:containerised-runtime",
      type = "CreativeWork",
      name = "Containerised Execution Environment"
    ),
    rocrateR::entity(
      id = "#control:network-isolation",
      type = "CreativeWork",
      name = "Network Isolation Policies"
    ),
    rocrateR::entity(
      id = "#control:role-based-access",
      type = "CreativeWork",
      name = "Role-Based Access Control"
    )
  )

  # physical controls -----
  physical_controls <- list(
    rocrateR::entity(
      id = "#control:secure-hosting",
      type = "CreativeWork",
      name = "Secure Data Centre Hosting"
    )
  )

  # organisational controls ----
  org_controls <- list(
    rocrateR::entity(
      id = "#control:project-governance",
      type = "CreativeWork",
      name = "Project-Level Governance Model"
    )
  )

  safe_root <- rocrateR::entity(
    id = id_hash("#safesetting:", "cr8tor"),
    type = "CreativeWork",
    name = "Safe Setting Controls (CR8TOR Deployment)",
    hasPart = purrr::map(
      c(tech_controls, physical_controls, org_controls) |> purrr::list_c(),
      ~ list("@id" = .x$`@id`)
    )
  )

  all_entities <- purrr::list_c(
    tech_controls,
    physical_controls,
    org_controls,
    list(safe_root)
  )

  purrr::reduce(all_entities, rocrateR::add_entity, .init = rocrate)
}

#' @rdname safe_setting
#' @export
safe_setting.opal <- function(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  path = NULL,
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL,
  user = NULL
) {
  # local binding
  Package <- NULL

  # validate connection ----
  # x is a valid opal connection object
  validate_opal_con(x)
  # validate that connection user has administrative rights
  is_opal_admin_con(x)

  # validate profile ----
  if (!opalr::dsadmin.profile_exists(x, profile)) {
    stop(
      sprintf("The given profile name, `%s`, does not exist!", profile),
      call. = FALSE
    )
  }

  # statistical disclosure controls ----
  # extract disclosure settings and create `PropertyValue` entities
  disc_setting_entities <- opalr::dsadmin.get_options(x, profile = profile) |>
    tibble::as_tibble() |>
    purrr::pmap(function(name, value, ...) {
      rocrateR::entity(
        id = id_hash("#disc:", paste0(name, value)),
        type = "PropertyValue",
        name = name,
        value = as.character(value)
      )
    })

  disclosure_env <- rocrateR::entity(
    id = id_hash("#env:disclosure_settings:", profile),
    type = "CreativeWork",
    name = "Disclosure Control Environment",
    description = sprintf(
      paste(
        "Disclosure control settings extract from the OBiBa Opal server",
        "connection provided, using the profile: '%s'."
      ),
      profile
    ),
    hasPart = purrr::map(disc_setting_entities, ~ list("@id" = .x$`@id`))
  )

  # computational environment ----
  # extract information about R packages installed in the environment
  pkg_tbl <- opalr::dsadmin.package_descriptions(x) |>
    tibble::as_tibble()
  pkg_entities <- pkg_tbl |>
    purrr::pmap(function(Package, Version, Description, Author, ...) {
      # create new entity
      rocrateR::entity(
        id = id_hash("#software:", paste0(Package, Version)),
        type = "SoftwareApplication",
        name = Package,
        version = Version,
        description = Description |>
          gsub(pattern = "[[:space:]]+", replacement = " ") |>
          trimws()
      )
    })

  # Opal server version
  opal_version <- tryCatch(
    {
      x$version
    },
    error = function(e) NA_character_
  )

  opal_entity <- rocrateR::entity(
    id = id_hash("#software:", paste0("opal", opal_version)),
    type = "SoftwareApplication",
    name = "Opal",
    version = opal_version,
    description = paste(
      "Opal is OBiBa's (https://www.obiba.org/) core database application for",
      "epidemiological studies. Participant data, collected by questionnaires,",
      "medical instruments, sensors, administrative databases etc. can be",
      "integrated and stored in a central data repository under a",
      "uniform model."
    )
  )

  # extract version of dsBase / DataSHIELD server
  dsBase_version <- tryCatch(
    pkg_tbl |>
      dplyr::filter(Package == "dsBase") |>
      (\(x) x$Version)(),
    error = function(e) NA_character_
  )

  software_env <- rocrateR::entity(
    id = id_hash("#env:software_stack:", paste0(opal_version, dsBase_version)),
    type = "CreativeWork",
    name = "Approved Analytical Software Environment",
    description = paste(
      "Software packages installed in the controlled Opal/DataSHIELD",
      "environment used for federated analysis."
    ),
    hasPart = purrr::map(
      c(pkg_entities, list(opal_entity)),
      ~ list("@id" = .x$`@id`)
    )
  )

  # technical controls ----
  tech_controls <- list(
    rocrateR::entity(
      id = "#control:output-checking",
      type = "CreativeWork",
      name = "Statistical Disclosure Output Checking",
      description = paste(
        "Automated disclosure control prevents release of",
        "small-cell counts and disclosive statistics."
      )
    ),
    rocrateR::entity(
      id = "#control:server-side-analysis",
      type = "CreativeWork",
      name = "Server-Side Analysis Enforcement",
      description = paste(
        "Raw data never leaves the secure server;",
        "analysis occurs via vetted aggregate functions."
      )
    ),
    rocrateR::entity(
      id = "#control:session-logging",
      type = "CreativeWork",
      name = "Comprehensive Session Logging",
      description = "All analytical actions are logged and auditable."
    )
  )

  # physical controls ----
  physical_controls <- list(
    rocrateR::entity(
      id = "#control:secure-facility",
      type = "CreativeWork",
      name = "Secure Data Facility",
      description = "Access restricted to approved secure premises."
    )
  )

  # organisational controls ----
  org_controls <- list(
    rocrateR::entity(
      id = "#control:access-governance",
      type = "CreativeWork",
      name = "Access Governance Process",
      description = "Data access committee review and approval required."
    )
  )

  # root safe setting entity ----
  safe_setting_root <- rocrateR::entity(
    id = id_hash("#safesetting:", "opal"),
    type = "CreativeWork",
    name = "Safe Setting Controls (Opal)",
    description = paste(
      "Technical, physical and organisational safeguards applied to minimise",
      "disclosure risk."
    ),
    hasPart = c(
      list(disclosure_env, software_env),
      tech_controls,
      physical_controls,
      org_controls
    ) |>
      purrr::map(\(x) list("@id" = x$`@id`))
  )

  # add entities to RO-Crate
  all_entities <- list(
    disc_setting_entities,
    disclosure_env,
    pkg_entities,
    opal_entity,
    software_env,
    tech_controls,
    physical_controls,
    org_controls,
    safe_setting_root
  )

  rocrate <- purrr::reduce(all_entities, rocrateR::add_entity, .init = rocrate)

  # link Safe Settings with safe Projects
  rocrate <- link_safe_settings_to_projects(rocrate)

  # attach input arguments as attributes
  attr(rocrate, "connection") <- x
  attr(rocrate, "path") <- path
  attr(rocrate, "profile") <- profile
  attr(rocrate, "project") <- project
  attr(rocrate, "resources") <- resources
  attr(rocrate, "tables") <- tables
  attr(rocrate, "user") <- user

  # return RO-Crate with the new entity
  return(rocrate)
}

#' @rdname safe_setting
#' @export
safe_setting.rocrate <- function(
  x,
  ...,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  user = attr(x, "user")
) {
  # check if the connection was given
  if (is.null(connection)) {
    stop("A `connection` object is required!", call. = FALSE)
  }

  # call method with given `connection` object:
  safe_setting(
    connection,
    rocrate = x,
    path = path,
    profile = profile,
    project = project,
    resources = resources,
    tables = tables,
    user = user
  )
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
# - - - - - - - - - - - - - - - - - ERRORS - - - - - - - - - - - - - - - - - #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
test_that("safe_setting fails when calling with invalid object", {
  # attempt calling function with invalid class
  expect_error(
    dsROCrate::safe_setting(structure(list(), class = "InvalidClass"))
  )
})

test_that("safe_setting fails when calling with invalid connection", {
  # setup
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  # attempt calling with invalid connection
  expect_error(
    basic_rocrate |>
      dsROCrate::safe_setting(connection = NULL)
  )
})

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
# - - - - - - - - - - - - - - - - - VALID  - - - - - - - - - - - - - - - - - #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
test_that("safe_setting works", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  # get settings for current server
  basic_rocrate_2 <- basic_rocrate |>
    dsROCrate::safe_setting(connection = opal_con, profile = "default")
  ## extract list of all settings
  basic_rocrate_2_ss <- basic_rocrate_2 |>
    rocrateR::get_entity(type = "PropertyValue")
  ## extract list of software tools
  basic_rocrate_2_tools <- basic_rocrate_2 |>
    rocrateR::get_entity(type = "SoftwareApplication")

  ## check that at least one setting was found in the RO-Crate
  expect_true(length(basic_rocrate_2_ss) >= 1)

  ## check that at least one tool was found in the RO-Crate
  expect_true(length(basic_rocrate_2_tools) >= 1)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

#' Extract Safe Output entity(ies)
#'
#' @inheritParams safe_data
#' @param ... Other optional arguments. See the full documentation
#'
#' @returns RO-Crate with Safe Output entity(ies).
#' @rdname extract_safe_output
#' @keywords internal
extract_safe_output <- function(x, ...) {
  UseMethod("extract_safe_output")
}

#' @param rocrate (Optional) RO-Crate object to update with Safe Output details.
#' @inheritParams safe_output
#' @rdname extract_safe_output
#' @export
extract_safe_output.opal <- function(
  x,
  ...,
  path = NULL,
  user = NULL,
  logs_to = Sys.time(),
  logs_from = logs_to - 24 * 60^2,
  rocrate = rocrateR::rocrate_5s()
) {
  # extract all the Safe Outputs for the current Opal connection
  rocrate <- safe_output(
    x,
    path = path,
    user = user,
    logs_to = logs_to,
    logs_from = logs_from,
    rocrate = rocrate
  )

  # return RO-Crate with Safe Output details
  return(rocrate)
}

#' @param id (Optional) Vector with `@id` strings for Safe Output entity(ies)
#'     to be extracted from the given RO-Crate, `x`.
#' @inheritParams safe_output
#' @rdname extract_safe_output
#' @export
extract_safe_output.rocrate <- function(
  x,
  ...,
  id = NULL,
  user = NULL,
  rocrate = rocrateR::rocrate_5s()
) {
  # validate RO-Crate
  rocrateR::is_rocrate(x)

  # extract File entities
  entities_lst <- .get_entity(x, type = "File")

  # if `id` was provided, then filter out only those entities
  if (!is.null(id)) {
    idx <- entities_lst |>
      sapply(\(x) getElement(x, "@id") %in% id)
    entities_lst <- entities_lst[idx]
  }

  # if `user` was provided, ensure that the entities' @id, contains this user
  if (!is.null(user)) {
    idx <- entities_lst |>
      sapply(\(x) grepl(user, getElement(x, "@id")))
    entities_lst <- entities_lst[idx]
  }

  # check if any entities were found
  if (length(entities_lst) == 0) {
    stop("No matching entities were found!", call. = FALSE)
  } else {
    message(
      length(entities_lst),
      " 'File' entit",
      ifelse(length(entities_lst) == 1, "y was", "ies were"),
      " found!"
    )
  }

  # add entities to the RO-Crate
  suppressWarnings({
    rocrate <- rocrate |>
      rocrateR::add_entity(entities_lst, verbose = FALSE)
  })

  # return RO-Crate with the Safe Output details
  return(rocrate)
}

#' Flatten object with Safe Output details
#'
#' @param x Object (e.g., RO-Crate) with Safe Output details. This can be
#'     generated with the [extract_safe_output()] function.
#' @param id Vector of strings with the `@id`s for the outputs to be extracted.
#'     If not provided, extract all entities with `@type = 'File'`.
#'
#' @returns Data frame with object mappings and functions from safe outputs.
#' @rdname flatten_safe_output
#' @keywords internal
flatten_safe_output <- function(x, ...) {
  UseMethod("flatten_safe_output")
}

#' @rdname flatten_safe_output
#' @export
flatten_safe_output.default <- function(x, ...) {
  return(tibble::tibble())
}

#' @rdname flatten_safe_output
#' @export
flatten_safe_output.rocrate <- function(x, ..., id = NULL) {
  tryCatch(
    {
      # extract entities
      entities_tbl <- .get_entity(x, type = "File") |>
        # extract @id, name, description, encodingFormat  and content for each entity
        lapply(function(ent) {
          tibble::tibble(
            id = getElement(ent, "@id"),
            name = getElement(ent, "name"),
            description = getElement(ent, "description"),
            encodingFormat = getElement(ent, "encodingFormat"),
            content = getElement(ent, "content")
          )
        }) |>
        # combine all rows
        purrr::list_flatten() |>
        purrr::map(tibble::as_tibble) |>
        purrr::list_rbind()

      # if `id` is provided, then only keep those entities
      if (!is.null(id)) {
        entities_tbl <- entities_tbl |>
          dplyr::filter(id %in% !!id)
      }

      # return dataset entities
      return(entities_tbl)
    },
    error = function(e) {
      tibble::tibble()
    }
  )
}

# -------------------------------------------------------------------------
# S3 generic dispatch
# -------------------------------------------------------------------------
test_that("extract_safe_output dispatch works for rocrate", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  out <- extract_safe_output(rocrate)

  expect_s3_class(out, "rocrate")
})

test_that("flatten_safe_output default returns empty tibble", {
  out <- flatten_safe_output(list())

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0)
})

# -------------------------------------------------------------------------
# extract_safe_output.rocrate()
# -------------------------------------------------------------------------
test_that("extract_safe_output.rocrate errors when no File entities", {
  # create basic RO-Crate
  rocrate <- rocrateR::rocrate_5s()

  # ignore warning about not finding File entities
  suppressWarnings(
    expect_error(
      extract_safe_output(rocrate),
      "No matching entities were found!"
    )
  )
})

test_that("extract_safe_output.rocrate extracts all File entities", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  out <- extract_safe_output(rocrate)

  expect_s3_class(out, "rocrate")

  files <- .get_entity(out, type = "File")
  expect_equal(length(files), 2)
})

test_that("extract_safe_output.rocrate filters by id correctly", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  out <- extract_safe_output(rocrate, id = "outputs/user1/report.csv")

  files <- .get_entity(out, type = "File")
  ids <- vapply(files, function(x) getElement(x, "@id"), character(1))

  expect_true("outputs/user1/report.csv" %in% ids)
})

test_that("extract_safe_output.rocrate filters by user string in @id", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  out <- extract_safe_output(rocrate, user = "user1")

  files <- .get_entity(out, type = "File")
  ids <- vapply(files, function(x) getElement(x, "@id"), character(1))

  expect_true(all(grepl("user1", ids)))
  expect_false(any(grepl("user2", ids)))
})

test_that("extract_safe_output.rocrate filters by both id and user", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  out <- extract_safe_output(
    rocrate,
    id = "outputs/user1/report.csv",
    user = "user1"
  )

  files <- .get_entity(out, type = "File")
  expect_equal(length(files), 1)
  expect_equal(getElement(files[[1]], "@id"), "outputs/user1/report.csv")
})

test_that("extract_safe_output.rocrate errors when filters remove all entities", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  expect_error(
    extract_safe_output(rocrate, user = "nonexistent_user"),
    "No matching entities were found!"
  )
})

test_that("extract_safe_output.rocrate emits informative message", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  expect_message(
    extract_safe_output(rocrate, user = "user1"),
    "'File' entit"
  )
})

test_that("extract_safe_output.rocrate validates rocrate input", {
  expect_error(
    extract_safe_output(structure(list(), class = "rocrate"))
  )
})

# -------------------------------------------------------------------------
# flatten_safe_output.rocrate()
# -------------------------------------------------------------------------
test_that("flatten_safe_output.rocrate returns tibble with correct columns", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  out <- flatten_safe_output(rocrate)

  expect_s3_class(out, "tbl_df")
  expect_true(all(
    c("id", "name", "description", "encodingFormat", "content") %in% names(out)
  ))
})

test_that("flatten_safe_output.rocrate extracts correct values", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  out <- flatten_safe_output(rocrate)

  expect_true("report.csv" %in% out$name)
  expect_true("summary.txt" %in% out$name)
  expect_true("text/csv" %in% out$encodingFormat)
})

test_that("flatten_safe_output.rocrate filters by id", {
  # create an RO-Crate with File entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user1/report.csv",
        type = "File",
        name = "report.csv",
        description = "User 1 report",
        encodingFormat = "text/csv",
        content = "a,b\n1,2"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "outputs/user2/summary.txt",
        type = "File",
        name = "summary.txt",
        description = "User 2 summary",
        encodingFormat = "text/plain",
        content = "summary content"
      )
    )

  out <- flatten_safe_output(rocrate, id = "outputs/user2/summary.txt")

  expect_equal(nrow(out), 1)
  expect_equal(out$name, "summary.txt")
})

test_that("flatten_safe_output.rocrate returns empty tibble on error", {
  # broken object to trigger tryCatch
  bad_obj <- structure(list(), class = "rocrate")

  out <- flatten_safe_output(bad_obj)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0)
})

# -------------------------------------------------------------------------
# extract_safe_output.opal() (Integration style with safe mocking)
# -------------------------------------------------------------------------
test_that("extract_safe_output.opal calls safe_output and returns rocrate", {
  # dummy opal-like object
  opal_con <- structure(list(), class = "opal")

  # Mock safe_output to avoid real server dependency
  mock_safe_output <- function(x, ..., rocrate) {
    file <- rocrateR::entity(
      id = "outputs/mock/user/file.txt",
      type = "File",
      name = "file.txt"
    )
    rocrateR::add_entity(rocrate, file)
  }

  with_mocked_bindings(
    safe_output = mock_safe_output,
    {
      out <- extract_safe_output.opal(opal_con)
    }
  )

  expect_s3_class(out, "rocrate")

  files <- .get_entity(out, type = "File")
  expect_true(length(files) >= 1)
})

# -------------------------------------------------------------------------
# Optional TRUE Opal demo integration (dsROCrate audit realism)
# -------------------------------------------------------------------------

test_that("extract_safe_output.opal works with Opal demo server (if available)", {
  con <- try(opal_demo_con(), silent = TRUE)
  skip_if(inherits(con, "try-error"), "Opal demo server not available")

  # ignore warning about missing `path` to write outputs
  suppressWarnings(
    rocrate <- extract_safe_output(con, user = attr(con, "PEOPLE"))
  )

  expect_s3_class(rocrate, "rocrate")
})

#' Safe Output details
#'
#' Safe Output details for the RO-Crate.
#'
#' All research outputs are checked to ensure individuals cannot be identified
#' even in the public domain.
#'
#' This means the other four Safes no longer apply.
#'
#' Before outputs are released from the TRE, they are checked to make sure it
#' is reasonably unlikely that any individuals can be identified on publication.
#' Compliance with the Digital Economy Act also requires that the TRE applies
#' methods and standards for output checking that are accredited by the UK
#' Statistics Authority.
#'
#' @param ... Other optional arguments. See the full documentation,
#'     [`?dsROCrate::safe_output`][safe_output()].
#' @param logs_to Upper limit timestamp to filter out the outputs generated
#'     (default: `Sys.time()`, current system time).
#' @param logs_from Lower limit timestamp to filter out the outputs generated
#'     (default: `Sys.time() - 24 * 60 ^ 2`, last 24 hours).
#' @inheritParams init
#'
#' @returns Updated RO-Crate object with Safe Outputs information.
#' @export
#'
#' @source
#' \itemize{
#'  \item Research Data Scotland, 2025. "What is the Five Safes framework?".
#'  <https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/>
#' }
safe_output <- function(x, ...) {
  UseMethod("safe_output")
}

#' @export
safe_output.default <- function(x, ...) {
  stop(
    "Unknown class, please try either a file path or",
    " an object with `rocrate` class!"
  )
}

#' @rdname safe_output
#' @export
safe_output.character <- function(
  x,
  ...,
  path = attr(x, "path"),
  user = attr(x, "user"),
  logs_to = Sys.time(),
  logs_from = logs_to - 24 * 60^2,
  connection = attr(x, "connection"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables")
) {
  # attempt loading the RO-Crate
  rocrate <- rocrateR::load_rocrate(x)

  # call method with given `rocrate` object:
  safe_output(
    rocrate,
    connection = connection,
    path = path,
    user = user,
    logs_to = logs_to,
    logs_from = logs_from,
    profile = profile,
    project = project,
    resources = resources,
    tables = tables
  )
}

#' @importFrom utils write.csv
#' @rdname safe_output
#' @export
safe_output.opal <- function(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  path = NULL,
  user = NULL,
  logs_to = Sys.time(),
  logs_from = logs_to - 24 * 60^2,
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL
) {
  # local bindings
  `@timestamp` <- backend <- logger_name <- safe_people_id <- username <- NULL
  ds_action <- ds_eval <- ds_id <- ds_function <- ds_symbol <- ds_table <- NULL

  # create formatted versions of input dates
  logs_from_is_valid <- FALSE
  if (!is.infinite(logs_from)) {
    logs_from_is_valid <- is_valid_posixct(logs_from)
  } else {
    # assumes that a value of `Inf` is a valid date
    logs_from_is_valid <- TRUE
  }

  logs_to_is_valid <- FALSE
  if (!is.infinite(logs_to)) {
    logs_to_is_valid <- is_valid_posixct(logs_to)
  } else {
    # assumes that a value of `Inf` is a valid date
    logs_to_is_valid <- TRUE
  }

  # display error message if either date is invalid
  if (!all(logs_from_is_valid, logs_to_is_valid)) {
    stop(
      paste0(
        "Invalid date string identified for: ",
        ifelse(logs_from_is_valid, "", "\n  - logs_from"),
        ifelse(logs_to_is_valid, "", "\n  - logs_to")
      ),
      call. = FALSE
    )
  }

  logs_from_formatted <- ifelse(
    is.infinite(logs_from),
    "ALL",
    format(as.POSIXct(logs_from), '%Y-%m-%d %H:%M:%S')
  )
  logs_to_formatted <- ifelse(
    is.infinite(logs_to),
    "ALL",
    format(as.POSIXct(logs_to), '%Y-%m-%d %H:%M:%S')
  )

  # x is a valid opal connection object
  validate_opal_con(x)

  # validate that connection user has administrative rights
  is_opal_admin_con(x)

  # verify if `user` is NULL, if so, retrieve information from the RO-crate
  if (is.null(user)) {
    # get `author` section from the root (./) entity
    rocrate_author <- rocrate |>
      .get_entity(id = "./") |>
      lapply(getElement, name = "author") |>
      sapply(\(x) x)

    # extract @id attribute(s)
    if (length(rocrate_author) > 1) {
      safe_people_id <- rocrate_author |>
        sapply(getElement, name = "@id") |>
        sapply(unlist)
    } else {
      safe_people_id <- rocrate_author["@id"][[1]]
    }

    # check if Safe People section wasn't found
    if (is.null(safe_people_id)) {
      warning(
        "Safe People section not found (i.e., no author for root entity) ",
        "in the given RO-Crate. \nEither run `dsROCrate::safe_people()` ",
        "or set `user` when calling `dsROCrate::safe_output()`!",
        call. = FALSE
      )

      # return the input RO-Crate
      return(rocrate)
    }

    # retrieve Safe People entity for the current user
    safe_people_information <- safe_people_id |>
      sapply(\(x) .get_entity(rocrate, id = x))

    # update user
    user <- safe_people_information |>
      sapply(getElement, name = "name") |>
      unique()
  }

  # check if the given value for `user` is a list
  if ("list" %in% class(user)) {
    user <- user |>
      purrr::pluck("name") |>
      unique()
  }

  # check if for any reason multiple users were found
  if (length(user) != 1) {
    warning(
      "Error when retrieving the Safe People section in the given ",
      "RO-Crate. ",
      length(user),
      " entries in the 'author' ",
      "section of root (./) entity were found!",
      call. = FALSE
    )

    # return the input RO-Crate
    return(rocrate)
  }

  # parse logs
  userlogs_tbl <- opalr::dsadmin.log(x) |>
    tibble::as_tibble() |>
    dplyr::mutate(
      `@timestamp` = as.POSIXct(`@timestamp`, format = "%Y-%m-%dT%H:%M:%S")
    ) |>
    # filter logs
    dplyr::filter(`@timestamp` >= logs_from, `@timestamp` <= logs_to) |>
    dplyr::filter(logger_name == "datashield.user") |>
    dplyr::filter(username %in% user)
  userlogs <- NULL
  if (nrow(userlogs_tbl) > 0) {
    userlogs <- userlogs_tbl |>
      with(
        sprintf(
          "[%s][%s]%-12s%s",
          level,
          format(`@timestamp`, "%Y-%m-%dT%H:%M:%S"),
          paste0("[", ds_action, "]"),
          message
        )
      )
  }

  # check if any logs were found in the given time frame
  if (length(userlogs) == 0) {
    warning(
      "No logs were found for the following configuration:",
      "\n User: ",
      user,
      "\n Period: ",
      logs_from_formatted,
      " -- ",
      logs_to_formatted,
      call. = FALSE
    )

    # return the input RO-Crate
    return(rocrate)
  }

  log_filename <- paste0(
    format(Sys.time(), "%Y%m%dT%H%M%S"),
    "-dslogs-",
    user,
    ".log"
  )

  # create new data entity for log file
  log_entity <- rocrateR::entity(
    id = basename(log_filename),
    type = "File",
    dateModified = Sys.time(),
    name = basename(log_filename),
    description = paste0(
      "This file contains the raw logs for the user: `",
      user,
      "` , between: ",
      logs_from_formatted,
      " and ",
      logs_to_formatted
    ),
    encodingFormat = "text/plain"
  )

  # extract list of functions executed
  ## evaluated functions and tables/symbols mapped
  userlogs_tbl_maps_evals <- userlogs_tbl |>
    dplyr::filter(ds_action %in% c("ASSIGN", "AGGREGATE", "OPEN")) |>
    # create derived `ds_eval` when `ds_action` = 'ASSIGN'
    dplyr::mutate(
      ds_eval = dplyr::coalesce(
        ds_eval,
        paste0(ds_symbol, " <- opal[", ds_table, "]")
      ),
      # attach session ID, `ds_id`, if `ds_eval` if `ds_action == 'OPEN'`
      ds_eval = ifelse(
        ds_action == "OPEN",
        paste0("Open session: ", ds_id),
        ds_eval
      )
    ) |>
    dplyr::distinct(
      ds_id,
      username,
      ds_action,
      ds_eval,
      ds_table,
      `@timestamp`
    ) |>
    # refill values for ds_table, based on ds_id
    dplyr::group_by(ds_id) |>
    dplyr::mutate(
      ds_table = refill_vec(ds_table)
    ) |>
    dplyr::ungroup() |>
    dplyr::mutate(
      # format timestamp
      `@timestamp` = format(`@timestamp`, '%Y-%m-%dT%H:%M:%S'),
      # extract function name from ds_eval
      ds_function = ds_eval |>
        gsub(pattern = "(?=\\().*$", replacement = "", perl = TRUE),
      # extract symbol/object from ds_eval
      ds_symbol = ds_eval |>
        gsub(pattern = "^.*(?<=\\()", replacement = "", perl = TRUE) |>
        gsub(pattern = "(?=\\)).*$", replacement = "", perl = TRUE) |>
        gsub(pattern = '"|\'', replacement = "", perl = TRUE) |>
        gsub(pattern = "(?=\\$).*", replacement = "", perl = TRUE),
      # autofill `ds_function` when `ds_action` = 'ASSIGN'
      ds_function = ifelse(ds_symbol == ds_eval, "base::assign", ds_function),
      # set `ds_function = 'DSI::datashield.login'` if `ds_action == 'OPEN'`
      ds_function = ifelse(
        ds_action == "OPEN",
        "DSI::datashield.login",
        ds_function
      ),
      ds_symbol = ifelse(ds_symbol == ds_eval, NA, ds_symbol),
      # add column with backend
      backend = "OBiBa's Opal",
      .before = 1
    ) |>
    dplyr::select(
      timestamp = `@timestamp`,
      action = ds_action,
      user = username,
      r_cmd = ds_eval,
      fx = ds_function,
      symbol = ds_symbol,
      table = ds_table,
      session = ds_id,
      backend
    )

  log_maps_filename <- paste0(
    format(Sys.time(), "%Y%m%dT%H%M%S"),
    "-dslogs-",
    user,
    "_mappings.csv"
  )

  # create new data entity for log file
  log_maps_entity <- rocrateR::entity(
    id = basename(log_maps_filename),
    type = "File",
    dateModified = Sys.time(),
    name = basename(log_maps_filename),
    description = "This file contains mappings and evaluated functions",
    encodingFormat = "text/csv"
  )

  # check if a `path` was not provided, then display warning and store contents
  # inside the RO-Crate entity
  if (is.null(path)) {
    warning(
      "A `path` wasn't provided! The logs will be included in the ",
      "RO-Crate object, under the `content` tag!",
      call. = FALSE
    )
    log_entity$content <- list(userlogs)
    log_maps_entity$content <- list(userlogs_tbl_maps_evals)
  } else {
    # validate if the given path is valid
    if (!dir.exists(path)) {
      warning(
        "The given `path` is not valid! The logs will be included in the ",
        "RO-Crate object, under the `content` tag!",
        call. = FALSE
      )
      log_entity$content <- list(userlogs)
      log_maps_entity$content <- list(userlogs_tbl_maps_evals)
    } else {
      writeLines(userlogs, file.path(path, log_filename))
      write.csv(
        userlogs_tbl_maps_evals,
        file.path(path, log_maps_filename),
        row.names = FALSE
      )
    }
  }

  # add entities to the RO-Crate
  rocrate <- rocrate |>
    rocrateR::add_entity(log_entity) |>
    rocrateR::add_entity(log_maps_entity) |>
    rocrateR::add_entity_value(
      id = "./",
      key = "hasPart",
      value = c(
        getElement(.get_entity(rocrate, id = "./"), "hasPart"),
        list(
          list(`@id` = log_entity$`@id`),
          list(`@id` = log_maps_entity$`@id`)
        )
      ),
      overwrite = TRUE
    )

  # attach input arguments as attributes
  attr(rocrate, "connection") <- x
  attr(rocrate, "path") <- path
  attr(rocrate, "profile") <- profile
  attr(rocrate, "project") <- project
  attr(rocrate, "resources") <- resources
  attr(rocrate, "tables") <- tables
  attr(rocrate, "user") <- user

  # return RO-Crate with the new entity
  return(rocrate)
}

#' @rdname safe_output
#' @export
safe_output.rocrate <- function(
  x,
  ...,
  path = attr(x, "path"),
  user = attr(x, "user"),
  logs_to = Sys.time(),
  logs_from = logs_to - 24 * 60^2,
  connection = attr(x, "connection"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables")
) {
  # check if the connection was given
  if (is.null(connection)) {
    stop("A `connection` object is required!", call. = FALSE)
  }

  # call method with given `connection` object:
  safe_output(
    connection,
    rocrate = x,
    path = path,
    user = user,
    logs_to = logs_to,
    logs_from = logs_from,
    profile = profile,
    project = project,
    resources = resources,
    tables = tables
  )
}

#' Extract Safe Project entity(ies)
#'
#' @inheritParams safe_data
#' @param ... Other optional arguments. See the full documentation
#'
#' @returns List with Safe Project entity(ies).
#' @rdname extract_safe_project
#' @keywords internal
extract_safe_project <- function(x, ...) {
  UseMethod("extract_safe_project")
}

#' @param rocrate (Optional) RO-Crate object to update with Safe Project
#'     details.
#' @rdname extract_safe_project
#' @export
extract_safe_project.opal <- function(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s()
) {
  # extract all data sources
  ds <- opalr::opal.datasources(x)

  # cycle through the data source (x) and extract project details
  for (i in seq_len(nrow(ds))) {
    project <- ds[i, "name"]
    if (!is.na(project) && !is.null(project)) {
      suppressWarnings({
        rocrate <- rocrate |>
          safe_project(project = project, connection = x)
      })
    }
  }

  # return RO-Crate with Safe Project details
  return(rocrate)
}

#' @param id (Optional) Vector with `@id` strings for Safe Project entity(ies)
#'     to be extracted from the given RO-Crate, `x`.
#' @rdname extract_safe_project
#' @export
extract_safe_project.rocrate <- function(
  x,
  ...,
  id = NULL,
  rocrate = rocrateR::rocrate_5s()
) {
  # validate RO-Crate
  rocrateR::is_rocrate(x)

  # extract Project entities
  entities_lst <- .get_entity(x, type = "Project")

  # extract Dataset entities
  data_entities_lst <- .get_entity(x, type = "Dataset")

  # if `id` was provided, then filter out only those entities
  if (!is.null(id)) {
    idx <- entities_lst |>
      sapply(\(x) getElement(x, "@id") %in% id)
    entities_lst <- entities_lst[idx]
  }

  # check if any entities were found
  if (length(entities_lst) == 0) {
    stop("No matching entities were found!", call. = FALSE)
  } else {
    message(
      length(entities_lst),
      " 'Project' entit",
      ifelse(length(entities_lst) == 1, "y was", "ies were"),
      " found!"
    )
  }

  # add entities to the RO-Crate
  suppressWarnings({
    for (p in seq_along(entities_lst)) {
      # attempt extracting Safe Data entities
      has_part <- getElement(entities_lst[[p]], "hasPart") |>
        unlist()
      # keep only entities linked to the current project
      idx <- data_entities_lst |>
        sapply(\(x) getElement(x, "@id") %in% has_part)
      proj_data_ents_lst <- data_entities_lst[idx]
      if (length(proj_data_ents_lst) > 0) {
        rocrate <- rocrate |>
          rocrateR::add_entity(proj_data_ents_lst, overwrite = TRUE)
      }
      # add current project entity
      rocrate <- rocrate |>
        rocrateR::add_entity(entities_lst[[p]])
    }
  })

  # return RO-Crate with the Safe Data details
  return(rocrate)
}

#' Flatten object with Safe Project details
#'
#' @param x Object (e.g., RO-Crate) with Safe Project details. This
#'     can be generated with the [extract_safe_project()] function.
#' @param ... Other optional arguments (not in used).
#' @param y Object (e.g., RO-Crate) with Safe Data details. This can be
#'     generated with the [extract_safe_data()] function. If not provided, it
#'     uses the `x` by default.
#'
#' @returns Data frame with safe project details.
#' @keywords internal
#' @rdname flatten_safe_project
flatten_safe_project <- function(x, ...) {
  UseMethod("flatten_safe_project")
}

#' @rdname flatten_safe_project
#' @export
flatten_safe_project.default <- function(x, ...) {
  return(tibble::tibble())
}

#' @rdname flatten_safe_project
#' @export
flatten_safe_project.rocrate <- function(x, ..., y = x) {
  tryCatch(
    {
      x |>
        # extract project entities
        .get_entity(type = "Project") |>
        # extract datasets/tables associated with the project
        lapply(function(ent) {
          # extract IDs for datasets
          has_part <- getElement(ent, "hasPart") |>
            unlist()
          project_id <- getElement(ent, "@id")
          project <- getElement(ent, "name")

          # if the current Project entity has any assets, extract them
          if (!is.null(has_part)) {
            # extract assets by @id
            assets_tbl <- y |>
              flatten_safe_data(id = has_part)

            proj_assets_tbl <- tibble::tibble(
              project_id = project_id,
              project = project
            ) |>
              dplyr::bind_cols(assets_tbl)

            return(proj_assets_tbl)
          }
          return(tibble::tibble(
            project_id = project_id,
            project = project,
            asset_id = NA,
            asset = NA
          ))
        }) |>
        dplyr::bind_rows()
    },
    error = function(e) {
      tibble::tibble()
    }
  )
}

# -------------------------------------------------------------------------
# S3 Generic Dispatch
# -------------------------------------------------------------------------

test_that("extract_safe_project dispatches correctly", {
  rocrate <- rocrateR::rocrate_5s()

  # ignore warning about lack of Project entity
  suppressWarnings(
    expect_error(
      extract_safe_project(rocrate),
      "No matching entities were found!"
    )
  )
})

test_that("flatten_safe_project dispatches default method", {
  out <- flatten_safe_project(list())
  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0)
})

# -------------------------------------------------------------------------
# extract_safe_project.rocrate()
# -------------------------------------------------------------------------

test_that("extract_safe_project.rocrate errors when no project entities", {
  rocrate <- rocrateR::rocrate_5s()

  # ignore warning about lack of Project entity
  suppressWarnings(
    expect_error(
      extract_safe_project(rocrate),
      "No matching entities were found!"
    )
  )
})

test_that("extract_safe_project.rocrate returns rocrate with project entities", {
  # create an RO-Crate with Project entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project1",
        type = "Project",
        name = "Project 1"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project2",
        type = "Project",
        name = "Project 2"
      )
    )

  out <- extract_safe_project(rocrate)

  expect_s3_class(out, "rocrate")

  projects <- rocrateR::get_entity(out, type = "Project")
  expect_true(length(projects) >= 1)
})

test_that("extract_safe_project.rocrate filters by id correctly", {
  # create an RO-Crate with Project entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project1",
        type = "Project",
        name = "Project 1"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project2",
        type = "Project",
        name = "Project 2"
      )
    )

  out <- extract_safe_project(rocrate, id = "#project1")

  projects <- rocrateR::get_entity(out, type = "Project")
  project_ids <- vapply(
    projects,
    function(x) getElement(x, "@id"),
    character(1)
  )

  expect_true("#project1" %in% project_ids)
})

test_that("extract_safe_project.rocrate throws error for non-matching id", {
  # create an RO-Crate with Project entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project1",
        type = "Project",
        name = "Project 1"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project2",
        type = "Project",
        name = "Project 2"
      )
    )

  expect_error(
    extract_safe_project(rocrate, id = "nonexistent"),
    "No matching entities were found!"
  )
})

test_that("extract_safe_project.rocrate emits informative message", {
  # create an RO-Crate with Project entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project1",
        type = "Project",
        name = "Project 1"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project2",
        type = "Project",
        name = "Project 2"
      )
    )

  expect_message(
    extract_safe_project(rocrate),
    "'Project' entit"
  )
})

test_that("extract_safe_project.rocrate adds linked dataset entities", {
  # create an RO-Crate with a Project and two Dataset entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project1",
        type = "Project",
        name = "Project 1",
        hasPart = list(
          sapply(c("#ds1", "#ds2"), \(x) list(`id` = x))
        )
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#ds1",
        type = "Dataset",
        name = "Dataset 1"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#ds2",
        type = "Dataset",
        name = "Dataset 2"
      )
    )

  out <- extract_safe_project(rocrate)

  datasets <- rocrateR::get_entity(out, type = "Dataset")
  dataset_ids <- vapply(
    datasets,
    function(x) getElement(x, "@id"),
    character(1)
  )

  expect_true(all(c("#ds1", "#ds2") %in% dataset_ids))
})

# -------------------------------------------------------------------------
# flatten_safe_project.rocrate()
# -------------------------------------------------------------------------

test_that("flatten_safe_project.rocrate returns tibble", {
  # create an RO-Crate with a Project and two Dataset entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project1",
        type = "Project",
        name = "Project 1",
        hasPart = list(
          sapply(c("#ds1", "#ds2"), \(x) list(`id` = x))
        )
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#ds1",
        type = "Dataset",
        name = "Dataset 1"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#ds2",
        type = "Dataset",
        name = "Dataset 2"
      )
    )

  # minimal stub for flatten_safe_data used internally
  flatten_safe_data <- function(x, ..., id = NULL) {
    tibble::tibble(
      asset_id = id,
      asset = c("table1", "table2")
    )
  }

  local_mocked_bindings(flatten_safe_data = flatten_safe_data)

  out <- flatten_safe_project(rocrate)

  expect_s3_class(out, "tbl_df")
  expect_true(all(
    c("project_id", "project", "asset_id", "asset") %in% names(out)
  ))
})

test_that("flatten_safe_project.rocrate extracts project metadata correctly", {
  # create an RO-Crate with a Project and two Dataset entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#project1",
        type = "Project",
        name = "Project 1",
        hasPart = list(
          sapply(c("#ds1", "#ds2"), \(x) list(`id` = x))
        )
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#ds1",
        type = "Dataset",
        name = "Dataset 1"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#ds2",
        type = "Dataset",
        name = "Dataset 2"
      )
    )

  flatten_safe_data <- function(x, ..., id = NULL) {
    tibble::tibble(
      asset_id = id,
      asset = paste0("name_", id)
    )
  }

  local_mocked_bindings(flatten_safe_data = flatten_safe_data)

  out <- flatten_safe_project(rocrate)

  expect_true("Project 1" %in% out$project)
  expect_true(any(grepl("name_", out$asset)))
})

test_that("flatten_safe_project.rocrate handles project without hasPart", {
  rocrate <- rocrateR::rocrate_5s()

  project <- rocrateR::entity(
    id = "project_no_part",
    type = "Project",
    name = "Lonely Project"
  )

  rocrate <- rocrate |>
    rocrateR::add_entity(project)

  out <- flatten_safe_project(rocrate)

  expect_equal(nrow(out), 1)
  expect_true(is.na(out$asset))
})

test_that("flatten_safe_project.rocrate returns empty tibble on error", {
  bad_obj <- structure(list(), class = "rocrate")

  out <- flatten_safe_project(bad_obj)

  expect_s3_class(out, "tbl_df")
})

# -------------------------------------------------------------------------
# extract_safe_project.opal() (INTEGRATION - demo server)
# -------------------------------------------------------------------------

test_that("extract_safe_project.opal works with demo Opal server", {
  # Skip if demo server is unreachable
  con <- try(opal_demo_con(), silent = TRUE)
  skip_if(inherits(con, "try-error"), "Opal demo server not available")

  rocrate <- extract_safe_project(con)

  expect_s3_class(rocrate, "rocrate")

  projects <- rocrateR::get_entity(rocrate, type = "Project")

  # CNSIM project should exist on demo server
  expect_true(length(projects) >= 1)
})

#' Parse user profiles
#'
#' @inheritParams project_exists
#' @param ... Optional arguments, unused.
#'
#' @returns Data frame with given `user`'s profile details, as captured on the
#' server pointed by `x`.
#' @keywords internal
#' @aliases parse_user_profiles,armadillo-method
#' @family Armadillo
#' @usage
#' \S4method{parse_user_profiles}{armadillo}(x, ..., user)
parse_user_profiles <- function(x, ...) {
  UseMethod("parse_user_profiles")
}

# S3 methods ----
#' @rdname parse_user_profiles
#' @family Opal
parse_user_profiles.opal <- function(x, ..., user) {
  # local bindings
  principal <- userInfo <- NULL

  # get user profiles and filter by the current user
  user_prof_tbl <- opalr::oadmin.user_profiles(x, df = FALSE) |>
    dplyr::bind_rows() |>
    dplyr::filter(principal %in% user)
  # extract (if available) `userInfo` which contains additional details
  if (nrow(user_prof_tbl) > 0 && "userInfo" %in% colnames(user_prof_tbl)) {
    user_prof_tbl <- user_prof_tbl |>
      dplyr::mutate(
        userInfo = userInfo |>
          lapply(function(x) {
            if (is.na(x)) {
              tibble::tibble()
            } else {
              x |>
                jsonlite::fromJSON() |>
                tibble::as_tibble()
            }
          })
      )
  }
  return(user_prof_tbl)
}

# S4 methods ----
#' @method parse_user_profiles ArmadilloCredentials
#' @rdname parse_user_profiles
#' @export
parse_user_profiles.ArmadilloCredentials <- function(x, ..., user) {
  message("PLACEHOLDER!")
}

#' Verify if project exists
#'
#' Wrapper for the [opalr::opal.project_exists()] and
#' [MolgenisArmadillo::armadillo.list_projects()] functions.
#'
#' @param x Connection object to backend for DataSHIELD server (e.g., Opal).
#' @param ... Optional arguments, unused.
#' @param project String with project name to be verified.
#'
#' @returns Nothing, call for its side effect. Stop execution of script if
#' `project` does not exist in the given server.
#'
#' @keywords internal
#' @aliases project_exists,armadillo-method
#' @family Armadillo
#' @usage
#' \S4method{project_exists}{armadillo}(
#'   x,
#'   ...,
#'   project
#' )
project_exists <- function(x, ...) {
  UseMethod("project_exists")
}

# S3 methods ----
#' @rdname project_exists
#' @family Opal
project_exists.opal <- function(x, ..., project) {
  if (!opalr::opal.project_exists(x, project)) {
    stop(
      sprintf(
        "The `project = '%s'` was not found in the given Opal connection!",
        project
      ),
      call. = FALSE
    )
  }
}

# S4 methods ----
#' @method project_exists ArmadilloCredentials
#' @rdname project_exists
#' @export
#' @family Armadillo
project_exists.ArmadilloCredentials <-
  function(
    x,
    ...,
    project
  ) {
    if (!(project %in% MolgenisArmadillo::armadillo.list_projects())) {
      stop(
        sprintf(
          "The `project = '%s'` was not found in the given Armadillo connection!",
          project
        ),
        call. = FALSE
      )
    }
  }

test_that("project_exists.opal() succeeds when project exists", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  expect_silent(
    project_exists(opal_con, project = attr(opal_con, "PROJECT"))
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("project_exists.opal() errors when project does not exist", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  expect_error(
    project_exists(opal_con, project = "THIS_PROJECT_DOES_NOT_EXIST"),
    "was not found in the given Opal connection!"
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("project_exists() dispatches to opal method", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  expect_silent(
    project_exists(opal_con, project = attr(opal_con, "PROJECT"))
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

# test_that("project_exists() works for armadillo objects", {
#   skip_if_not_installed("MolgenisArmadillo")
#
#   arm <- methods::new("ArmadilloCredentials")
#
#   local_mocked_bindings(
#     armadillo.list_projects = function() c("proj1", "proj2"),
#     .package = "MolgenisArmadillo"
#   )
#
#   expect_silent(
#     project_exists(arm, project = "proj1")
#   )
#
#   expect_error(
#     project_exists(arm, project = "missing"),
#     "was not found in the given Armadillo connection!"
#   )
# })

test_that("parse_user_profiles() dispatches to opal method", {
  opal_con <- structure(list(), class = "opal")

  local_mocked_bindings(
    `opal.get` = function(...) {
      list(data.frame(principal = "user1", stringsAsFactors = FALSE))
    },
    .package = "opalr"
  )

  res <- parse_user_profiles(opal_con, user = "user1")

  expect_s3_class(res, "data.frame")
  expect_equal(res$principal, "user1")
})

test_that("parse_user_profiles.opal() returns empty tibble when no matching users", {
  opal_con <- structure(list(), class = "opal")

  local_mocked_bindings(
    `opal.get` = function(...) {
      list(data.frame(principal = "other_user", stringsAsFactors = FALSE))
    },
    .package = "opalr"
  )

  res <- parse_user_profiles(opal_con, user = "user1")

  expect_equal(nrow(res), 0)
})

test_that("parse_user_profiles.opal() handles missing userInfo column", {
  opal_con <- structure(list(), class = "opal")

  local_mocked_bindings(
    `opal.get` = function(...) {
      list(
        data.frame(
          principal = c("user1", "user2"),
          stringsAsFactors = FALSE
        )
      )
    },
    .package = "opalr"
  )

  res <- parse_user_profiles(opal_con, user = "user1")

  expect_equal(nrow(res), 1)
  expect_false("userInfo" %in% colnames(res) && is.list(res$userInfo))
})

test_that("parse_user_profiles.opal() parses JSON userInfo and NA correctly", {
  opal_con <- structure(list(), class = "opal")

  json <- '{"givenName":"John","familyName":"Doe"}'

  local_mocked_bindings(
    `opal.get` = function(...) {
      list(
        data.frame(
          principal = c("user1", "user1"),
          userInfo = c(json, NA_character_),
          stringsAsFactors = FALSE
        )
      )
    },
    .package = "opalr"
  )

  res <- parse_user_profiles(opal_con, user = "user1")

  expect_equal(nrow(res), 2)
  expect_true(is.list(res$userInfo))
  expect_s3_class(res$userInfo[[1]], "tbl_df")
  expect_equal(nrow(res$userInfo[[2]]), 0)
})

# test_that("project_exists.opal() succeeds when project exists", {
#   # setup
#   opal_con <- opal_demo_con()
#
#   expect_silent(
#     dsROCrate:::project_exists(opal_con, project = attr(opal_con, "PROJECT"))
#   )
# })
#
# test_that("project_exists.opal() errors when project does not exist", {
#   expect_error(
#     project_exists(opal_con, project = "THIS_PROJECT_DOES_NOT_EXIST"),
#     "was not found in the given Opal connection!"
#   )
# })
#
# test_that("project_exists() dispatches to opal method", {
#   expect_silent(
#     project_exists(opal_con, project = attr(opal_con, "PROJECT"))
#   )
# })
#
# test_that("project_exists() works for armadillo objects", {
#   skip_if_not_installed("MolgenisArmadillo")
#
#   arm <- methods::new("armadillo")
#
#   local_mocked_bindings(
#     armadillo.list_projects = function() c("proj1", "proj2"),
#     .package = "MolgenisArmadillo"
#   )
#
#   expect_silent(
#     project_exists(arm, project = "proj1")
#   )
#
#   expect_error(
#     project_exists(arm, project = "missing"),
#     "was not found in the given Armadillo connection!"
#   )
# })

#' Extract Safe People entity(ies)
#'
#' @inheritParams safe_data
#' @param ... Other optional arguments. See the full documentation
#'
#' @returns RO-Crate with Safe People entity(ies).
#' @rdname extract_safe_people
#' @keywords internal
extract_safe_people <- function(x, ...) {
  UseMethod("extract_safe_people")
}

#' @param rocrate (Optional) RO-Crate object to update with Safe People details.
#' @rdname extract_safe_people
#' @export
extract_safe_people.opal <- function(x, ..., rocrate = rocrateR::rocrate_5s()) {
  # set local binding
  name <- principal <- NULL

  # extract all users
  opal_users <- opalr::oadmin.user_profiles(x, df = FALSE) |>
    dplyr::bind_rows() |>
    dplyr::rename(name = principal) |>
    # exclude system administrators from the report
    dplyr::filter(!(tolower(name) %in% c("admin", "administrator")))

  # cycle through the data source (x) and extract project details
  for (i in seq_len(nrow(opal_users))) {
    username <- opal_users[i, "name"]
    if (!is.na(username) && !is.null(username)) {
      suppressWarnings({
        rocrate <- rocrate |>
          safe_people(user = username, connection = x, set_author = FALSE)
      })
    }
  }

  # return RO-Crate with Safe Project details
  return(rocrate)
}

#' @param id (Optional) Vector with `@id` strings for Safe People entity(ies)
#'     to be extracted from the given RO-Crate, `x`.
#' @rdname extract_safe_people
#' @export
extract_safe_people.rocrate <- function(
  x,
  ...,
  id = NULL,
  rocrate = rocrateR::rocrate_5s()
) {
  # if `id` wasn't provided, then extract from root (./) entity of the RO-Crate
  if (is.null(id)) {
    # extract author `@id`s from the root directory
    id <- .get_entity(x, id = "./", type = "Dataset") |>
      lapply(\(x) getElement(x, "author")) |>
      sapply(\(x) getElement(x, "@id")) |>
      unlist()
  }

  # extract entities with type = 'Person'
  entities_lst <- .get_entity(x, type = "Person")

  # filter out person-entities in the `id`, if `id` is not NULL
  if (length(id) && !is.null(id)) {
    idx <- entities_lst |>
      sapply(\(x) getElement(x, "@id") %in% id)
    entities_lst_v2 <- entities_lst[idx]
  } else {
    entities_lst_v2 <- entities_lst
  }

  # check if any entities were found
  if (length(entities_lst_v2) == 0) {
    stop(
      "No matching entities were found for the Author(s) in the root ",
      "entity (./):\n",
      paste0("  - ", id, collapse = "\n"),
      call. = FALSE
    )
  } else {
    message(
      length(entities_lst_v2),
      " 'Author' entit",
      ifelse(length(entities_lst_v2) == 1, "y was", "ies were"),
      " found!"
    )
  }

  # add entities to the RO-Crate
  rocrate <- rocrate |>
    rocrateR::add_entity(entities_lst_v2, overwrite = TRUE, verbose = FALSE)
  # # link new user entity @id to the root (./) author property
  # rocrateR::add_entity_value(
  #   id = "./",
  #   key = "author",
  #   value = list(`@id` = getElement(entities_lst_v2, "@id"))
  # )

  # return RO-Crate with the Safe People details
  return(rocrate)
}

#' Flatten object with Safe People details
#'
#' @param x Object (e.g., RO-Crate) with Safe People details. This can be
#'     generated with the [extract_safe_data()] function.
#' @param ... Other optional arguments (not in used).
#' @param id Vector of strings with the `@id`s for the users to be extracted.
#'     If not provided, extract all entities with `@type = 'Person'`.
#'
#' @returns Data frame with safe people details.
#' @rdname flatten_safe_people
#' @keywords internal
flatten_safe_people <- function(x, ...) {
  UseMethod("flatten_safe_people")
}

#' @rdname flatten_safe_people
#' @export
flatten_safe_people.default <- function(x, ...) {
  return(tibble::tibble())
}

#' @rdname flatten_safe_people
#' @export
flatten_safe_people.rocrate <- function(x, ..., id = NULL) {
  # local bindings
  person_id <- NULL
  tryCatch(
    {
      # extract Person entities
      entities_tbl <- x |>
        .get_entity(type = "Person") |>
        # extract @id and name for each entity
        lapply(function(ent) {
          tibble::tibble(
            person_id = getElement(ent, "@id"),
            name = getElement(ent, "name"),
            given_name = c(
              getElement(ent, "givenName"),
              getElement(ent, "given_name")
            ),
            family_name = c(
              getElement(ent, "familyName"),
              getElement(ent, "family_name")
            ),
            organisation = getElement(ent, "organisation")
          )
        }) |>
        # combine all rows
        dplyr::bind_rows()

      # if `id` is provided, then only keep those entities
      if (!is.null(id)) {
        entities_tbl <- entities_tbl |>
          dplyr::filter(person_id %in% !!id)
      }

      # return dataset entities
      return(entities_tbl)
    },
    error = function(e) {
      tibble::tibble()
    }
  )
}

test_that("extract_safe_people.opal returns a valid RO-Crate with Person entities", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  # terminate connection when done with tests
  withr::defer(opalr::opal.logout(opal_con))

  roc <- extract_safe_people(opal_con)

  # validate rocrate structure
  expect_no_error(rocrateR::is_rocrate(roc))

  people <- rocrateR::get_entity(roc, type = "Person")

  expect_type(people, "list")
  expect_true(length(people) >= 1)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("extract_safe_people.opal updates an existing RO-Crate", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  # terminate connection when done with tests
  withr::defer(opalr::opal.logout(opal_con))

  base_roc <- rocrateR::rocrate_5s()

  roc <- extract_safe_people(opal_con, rocrate = base_roc)

  # validate rocrate structure
  expect_no_error(rocrateR::is_rocrate(roc))

  people <- rocrateR::get_entity(roc, type = "Person")
  expect_true(length(people) >= 1)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("extract_safe_people.opal excludes admin and administrator accounts", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  # terminate connection when done with tests
  withr::defer(opalr::opal.logout(opal_con))

  roc <- extract_safe_people(opal_con)

  people <- rocrateR::get_entity(roc, type = "Person")
  ids <- vapply(people, function(e) getElement(e, "name")[[1]], character(1))

  expect_false(any(tolower(ids) %in% c("admin", "administrator")))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("extract_safe_people.opal iterates over all returned subject profiles", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  # terminate connection when done with tests
  withr::defer(opalr::opal.logout(opal_con))

  users_raw <- opalr::oadmin.user_profiles(opal_con, df = FALSE)
  users_tbl <- dplyr::bind_rows(users_raw)

  expect_true(nrow(users_tbl) >= 1)

  roc <- extract_safe_people(opal_con)

  # validate rocrate structure
  expect_no_error(rocrateR::is_rocrate(roc))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("extract_safe_people.rocrate copies Person entities", {
  src <- rocrateR::rocrate_5s()

  src <- rocrateR::add_entity(
    src,
    entity = rocrateR::entity(
      id = "#user1",
      type = "Person",
      name = "Alice"
    )
  )

  src <- rocrateR::add_entity(
    src,
    entity = rocrateR::entity(
      id = "#user2",
      type = "Person",
      name = "Bob"
    )
  )

  new_roc <- extract_safe_people(src)

  # validate rocrate structure
  expect_no_error(rocrateR::is_rocrate(src))

  ents <- rocrateR::get_entity(new_roc, type = "Person")
  expect_equal(length(ents), 2)
})

test_that("extract_safe_people.rocrate filters Person entities by id", {
  src <- rocrateR::rocrate_5s()

  src <- rocrateR::add_entity(
    src,
    entity = rocrateR::entity(
      id = "#keep",
      type = "Person",
      name = "Keep Me"
    )
  )

  src <- rocrateR::add_entity(
    src,
    entity = rocrateR::entity(
      id = "#drop",
      type = "Person",
      name = "Drop Me"
    )
  )

  new_roc <- extract_safe_people(src, id = "#keep")

  ents <- rocrateR::get_entity(new_roc, type = "Person")
  ids <- vapply(ents, function(e) getElement(e, "@id")[[1]], character(1))

  expect_equal(ids, "#keep")
})

test_that("extract_safe_people.rocrate infers ids from root Dataset author", {
  src <- rocrateR::rocrate_5s()

  # create person entity
  src <- rocrateR::add_entity(
    src,
    entity = rocrateR::entity(
      id = "#author1",
      type = "Person",
      name = "Author One"
    )
  )

  # attach author to root dataset
  src <- rocrateR::add_entity_value(
    src,
    id = "./",
    key = "author",
    value = list(list(`@id` = "#author1"))
  )

  new_roc <- extract_safe_people(src)

  ents <- rocrateR::get_entity(new_roc, type = "Person")
  expect_equal(length(ents), 1)
  expect_equal(getElement(ents[[1]], "@id"), "#author1")
})

test_that("extract_safe_people.rocrate errors when no matching Person entities found", {
  src <- rocrateR::rocrate_5s()

  # ignore warning about not finding People entities
  suppressWarnings(
    expect_error(
      extract_safe_people(src, id = "#nonexistent"),
      "No matching entities were found for the Author"
    )
  )
})

test_that("flatten_safe_people.default returns empty tibble", {
  res <- flatten_safe_people(list())

  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 0)
})

test_that("flatten_safe_people.rocrate extracts person metadata correctly", {
  roc <- rocrateR::rocrate_5s()

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "#p1",
      type = "Person",
      name = "Jane Doe",
      givenName = "Jane",
      familyName = "Doe",
      organisation = "Test Org"
    )
  )

  res <- flatten_safe_people(roc)

  expect_s3_class(res, "data.frame")
  expect_true(all(
    c("person_id", "name", "given_name", "family_name", "organisation") %in%
      names(res)
  ))
  expect_equal(res$person_id, "#p1")
  expect_equal(res$name, "Jane Doe")
})

test_that("flatten_safe_people.rocrate filters by id argument", {
  roc <- rocrateR::rocrate_5s()

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(id = "#p1", type = "Person", name = "A")
  )

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(id = "#p2", type = "Person", name = "B")
  )

  res <- flatten_safe_people(roc, id = "#p2")

  expect_equal(nrow(res), 1)
  expect_equal(res$person_id, "#p2")
})

test_that("flatten_safe_people.rocrate returns empty tibble on malformed input", {
  bad_obj <- list() # not a rocrate

  res <- flatten_safe_people.rocrate(bad_obj)

  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 0)
})

#' Safe Project details
#'
#' Safe Project details for the RO-Crate.
#'
#' Data must be used ethically, for research that delivers clear public benefit.
#'
#' As part of their application, researchers are asked to provide an overview
#' of their project, including how the data will be used and what outputs will
#' be achieved. This allows data providers to make an informed decision about
#' whether they are comfortable preparing data for the researcher to use for
#' ethical purposes serving a public good.
#'
#' @param ... Other optional arguments. See the full documentation,
#'     [`?dsROCrate::safe_project`][safe_project()].
#' @inheritParams safe_data
#'
#' @returns Updated RO-Crate object with Safe Project information.
#' @export
#'
#' @source
#' \itemize{
#'  \item Research Data Scotland, 2025. "What is the Five Safes framework?".
#'  <https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/>
#' }
#'
#' @aliases safe_project,armadillo-method
#' @usage
#' \S4method{safe_project}{armadillo}(
#'   x,
#'   ...,
#'   profile = "default",
#'   project = NULL,
#'   rocrate = rocrateR::rocrate_5s(),
#'   asset_id_suffix = "#asset:",
#'   project_id_suffix = "#project:",
#'   path = NULL,
#'   resources = NULL,
#'   tables = NULL,
#'   user = NULL
#' )
safe_project <- function(x, ...) {
  UseMethod("safe_project")
}

# S3 methods ----
#' @method safe_project default
#' @export
safe_project.default <- function(x, ...) {
  stop(
    "Unknown class, please try either a file path or",
    " an object with `rocrate` class!"
  )
}

#' @method safe_project character
#' @rdname safe_project
#' @export
safe_project.character <- function(
  x,
  ...,
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  user = attr(x, "user")
) {
  # attempt loading the RO-Crate
  rocrate <- rocrateR::load_rocrate(x)

  # call method with given `rocrate` object:
  safe_project(
    rocrate,
    connection = connection,
    profile = profile,
    project = project,
    asset_id_suffix = asset_id_suffix,
    project_id_suffix = project_id_suffix,
    path = path,
    resources = resources,
    tables = tables,
    user = user
  )
}

#' @method safe_project opal
#' @rdname safe_project
#' @export
safe_project.opal <- function(
  x,
  ...,
  profile = "default",
  project = NULL,
  rocrate = rocrateR::rocrate_5s(),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  path = NULL,
  resources = NULL,
  tables = NULL,
  user = NULL
) {
  # declare local bindings
  created <- lastUpdate <- type <- NULL

  # x is a valid opal connection object
  validate_opal_con(x)

  # validate that connection user has administrative rights
  is_opal_admin_con(x)

  # enforce that `project` is a single value
  if (is.null(project)) {
    stop("A value for `project` is required!", call. = FALSE)
  } else if (length(project) != 1) {
    stop("`project` must be a single value!", call. = FALSE)
  }

  # check if the given `project` exists
  project_exists(x, project = project)

  # create project @id
  project_id <- id_hash(project_id_suffix, project)

  # retrieve details associated to `project`
  project_details_tbl <- opalr::opal.project(x, project)

  # filter out asset entities associated with the project based on the
  # value for `asset_id_suffix`.
  crate_asset_entities <- rocrate$`@graph` |>
    purrr::keep(\(x) grepl(paste0("^", asset_id_suffix), x$`@id`))

  # create project entity
  timestamps <- getElement(project_details_tbl, "timestamps")
  project_entity <- rocrateR::entity(
    id = project_id,
    type = "Project",
    name = getElement(project_details_tbl, "name"),
    dateCreated = getElement(timestamps, "created"),
    dateModified = getElement(timestamps, "lastUpdate"),
    hasPart = if (length(crate_asset_entities) > 0) {
      purrr::map(crate_asset_entities, ~ list("@id" = .x$`@id`))
    } else {
      NULL
    }
  )

  # add new project entity to the RO-Crate
  rocrate <- rocrate |>
    rocrateR::add_entity(project_entity, overwrite = TRUE)

  # Opal permissions ----
  perms <- opalr::opal.project_perm(x, project)

  project_users <- perms |>
    dplyr::filter(type == "user") |>
    purrr::pmap_chr(\(subject, ...) subject) |>
    stats::na.omit()

  # link existing Person entities ----
  people <- .get_entity(rocrate, type = "Person")

  if (!is.null(people)) {
    for (p in people) {
      user <- p$name

      if (user %in% project_users) {
        rocrate <- append_entity_ref(
          rocrate,
          id = p[["@id"]],
          key = "memberOf",
          ref_id = project_id
        )
      }
    }
  }

  # link existing asset entities ----
  # extract assets for the given project
  project_tbl_assets <- get_project_assets(x, project, "tables")
  project_res_assets <- get_project_assets(x, project, "resources")
  # combine assets
  proj_assets_tbl <- dplyr::bind_rows(project_tbl_assets, project_res_assets)

  # filter crate's assets based on the assets associated to the project
  crate_asset_entities <- crate_asset_entities |>
    purrr::keep(\(x) getElement(x, "name") %in% proj_assets_tbl$name)

  if (!is.null(crate_asset_entities) && length(crate_asset_entities) > 0) {
    for (ass in crate_asset_entities) {
      rocrate <- append_entity_ref(
        rocrate,
        id = ass[["@id"]],
        key = "isPartOf",
        ref_id = project_id
      )
    }
  }

  # attach input arguments as attributes
  attr(rocrate, "connection") <- x
  attr(rocrate, "path") <- path
  attr(rocrate, "profile") <- profile
  attr(rocrate, "project") <- project
  attr(rocrate, "resources") <- resources
  attr(rocrate, "tables") <- tables
  attr(rocrate, "user") <- user

  # return RO-Crate with the new entity
  return(rocrate)
}

#' @method safe_project rocrate
#' @rdname safe_project
#' @export
safe_project.rocrate <- function(
  x,
  ...,
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  user = attr(x, "user")
) {
  # check if the connection was given
  if (is.null(connection)) {
    stop("A `connection` object is required!", call. = FALSE)
  }

  # call method with given `connection` object:
  safe_project(
    connection,
    rocrate = x,
    path = path,
    profile = profile,
    project = project,
    asset_id_suffix = asset_id_suffix,
    project_id_suffix = project_id_suffix,
    resources = resources,
    tables = tables,
    user = user
  )
}

# S4 methods ----
#' @method safe_project ArmadilloCredentials
#' @rdname safe_project
#' @export
safe_project.ArmadilloCredentials <- function(
  x,
  ...,
  profile = "default",
  project = NULL,
  rocrate = rocrateR::rocrate_5s(),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  path = NULL,
  resources = NULL,
  tables = NULL,
  user = NULL
) {
  # check if the given `project` exists
  project_exists(x, project = project)

  # retrieve details associated to `project`
  project_details_tbl <- MolgenisArmadillo::armadillo.get_projects_info() |>
    purrr::list_c() |>
    tibble::as_tibble()
}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
# - - - - - - - - - - - - - - - - - ERRORS - - - - - - - - - - - - - - - - - #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
test_that("safe_project fails when attempt calling function with invalid class", {
  expect_error(
    dsROCrate::safe_project(
      structure(list(), class = "InvalidClass"),
      project = NULL
    )
  )
})

test_that("safe_project fails when attempt calling with invalid connection", {
  # setup
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  expect_error(
    basic_rocrate |>
      dsROCrate::safe_project(connection = NULL, project = NULL)
  )
})

test_that("safe_project fails when attempt adding invalid project", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  expect_error(
    basic_rocrate |>
      dsROCrate::safe_project(
        project = "Invalid Project",
        connection = opal_con
      )
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("safe_project fails when attempt calling without a project", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  expect_error(
    basic_rocrate |>
      dsROCrate::safe_project(connection = opal_con)
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("safe_project fails when attempt calling with project = NULL", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  expect_error(
    basic_rocrate |>
      dsROCrate::safe_project(connection = opal_con, project = NULL)
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("safe_project fails when attempt calling with multiple projects", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  expect_error(
    basic_rocrate |>
      dsROCrate::safe_project(connection = opal_con, project = c("A", "B"))
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
# - - - - - - - - - - - - - - - - - VALID  - - - - - - - - - - - - - - - - - #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -#
test_that("safe_project works", {
  # setup
  ## open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  ## create basic RO-Crate
  basic_rocrate <- rocrateR::rocrate_5s()

  # add details for Project without Dataset entities
  basic_rocrate_2 <- basic_rocrate |>
    dsROCrate::safe_project(
      project = attr(opal_con, "PROJECT"),
      connection = opal_con
    )

  # add all datasets for a valid project
  basic_rocrate_3 <- basic_rocrate |>
    dsROCrate::safe_data(
      project = attr(opal_con, "PROJECT"),
      connection = opal_con
    ) |>
    # add the Safe Project details
    dsROCrate::safe_project(
      project = attr(opal_con, "PROJECT"),
      connection = opal_con
    )
  ## extract datasets
  basic_rocrate_3_sd <- basic_rocrate_3 |>
    rocrateR::get_entity(type = "Dataset")
  ## extract datasets' names
  basic_rocrate_3_sd_names <- basic_rocrate_3_sd |>
    sapply(getElement, name = "name")
  ## verify that CNSIM1, CNSIM2 and CNSIM3 are part in the crate
  expect_true(all(
    c("CNSIM1", "CNSIM2", "CNSIM3") %in% basic_rocrate_3_sd_names
  ))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

#' Create an RO-Crate report
#'
#' @param x This can be an RO-Crate ([rocrate][rocrateR::rocrate()] class) or a
#'     string with the path to an RO-Crate.
#' @param ... Other optional arguments. See the full documentation,
#'     [`?dsROCrate::report`][report()].
#'
#' @returns RO-Crate report as markdown (.md) file and/or HTML.
#' @export
report <- function(x, ...) {
  UseMethod("report")
}

#' @rdname report
#' @export
report.character <- function(
  x,
  ...,
  title = "DataSHIELD Report",
  filepath = tempfile(fileext = ".md"),
  render = TRUE,
  doc_format = "html",
  overwrite = FALSE,
  include_user_perm = TRUE,
  diag_title = "DataSHIELD server",
  diag_width = NULL,
  diag_height = NULL,
  max_line_length = 200
) {
  # attempt loading the RO-Crate
  rocrate <- rocrateR::load_rocrate(x)

  # call the next generic method
  rocrate |>
    report(
      title = title,
      filepath = filepath,
      render = render,
      doc_format = doc_format,
      overwrite = overwrite,
      include_user_perm = include_user_perm,
      diag_title = diag_title,
      diag_width = diag_width,
      diag_height = diag_height
    )
}

#' @rdname report
#' @export
report.default <- function(x, ...) {
  stop(
    "Unknown class, please try either a file path or",
    " an object with `rocrate` class!"
  )
}

#' @param study_name String with the study name.
#' @rdname report
#' @export
report.list <- function(
  x,
  ...,
  study_name,
  title = "DataSHIELD Report",
  filepath = tempfile(fileext = ".md"),
  render = TRUE,
  doc_format = "html",
  overwrite = FALSE,
  include_user_perm = TRUE,
  diag_title = "DataSHIELD server",
  diag_width = NULL,
  diag_height = NULL,
  max_line_length = 200
) {
  # local bindings
  asset <- id <- name <- permission <- project <- server <- user <- NULL

  # validate that all the objects in the list, `x`, are valid RO-Crates
  sapply(x, rocrateR::is_rocrate)

  # generate individual reports for each RO-Crate
  report_outputs <- lapply(
    x,
    report,
    title = title,
    filepath = filepath,
    render = FALSE,
    doc_format = doc_format,
    overwrite = TRUE,
    include_user_perm = include_user_perm,
    diag_title = diag_title,
    diag_width = diag_width,
    diag_height = diag_height
  )

  # combine reports ----
  ## Safe People
  safe_people_all <- .extract_srvr_comp(report_outputs, "safe_people")
  ## Safe Data
  safe_data_all <- .extract_srvr_comp(report_outputs, "safe_data")
  ## Safe Projects
  safe_project_all <- .extract_srvr_comp(report_outputs, "safe_project")
  ## Safe Settings
  safe_setting_all <- .extract_srvr_comp(report_outputs, "safe_setting")
  ## Safe Outputs
  safe_output_all <- .extract_srvr_comp(report_outputs, "safe_output")
  ## Safe Data permissions (optional)
  safe_data_permissions_all <- report_outputs |>
    .extract_srvr_comp("safe_data_permissions")

  ## overview table ----
  overview_data_all <- tibble::tibble()
  ## combine aggregated data to generate new overview table
  safe_project_data_all <- safe_project_all #|>
  # dplyr::rename(asset = asset_name)

  # if data permissions were found, then combine with project-data details and
  # generate new overview table
  if (!is.null(safe_data_permissions_all)) {
    overview_data_all <- safe_data_permissions_all |>
      dplyr::left_join(safe_people_all, by = c("person_id", "server")) |>
      dplyr::left_join(safe_project_data_all, by = c("asset_id", "server")) |>
      dplyr::select(server, project, asset, permission, user = name)

    # if any Safe Outputs were found in the inputs, include in the overview
    if (!is.null(safe_output_all) && nrow(safe_output_all)) {
      overview_data_all <- overview_data_all |>
        dplyr::left_join(
          safe_output_all,
          by = c("server", "project", "asset", "user")
        )
    }
  } else {
    # extract previous overview table
    overview_data_all <- tryCatch(
      {
        report_outputs |>
          # extract each component per server
          lapply(getElement, name = "overview_data") |>
          # attach the server name as a new column
          purrr::imap(~ dplyr::mutate(.x, server = .y)) |>
          # combine rows
          dplyr::bind_rows()
      },
      error = function(e) {
        NULL
      }
    )
  }

  ## overview diagram ----
  overview_lst <- overview_data_all |>
    dplyr::rename(name = user) |>
    .overview_diagram(
      include_user_perm,
      filepath,
      render,
      diag_title,
      diag_width,
      diag_height
    )

  ## create tidy overview table ----
  tidy_overview_tbl <- overview_data_all |>
    dplyr::rename(name = user) |>
    .tidy_overview(include_user_perm = include_user_perm)

  ## create markdown report ----
  ## header and overview table
  report_contents <- c(
    .markdown_report_header(title, overview_data_all, overview_lst$diag_path),
    tidy_overview_tbl |>
      # # tidy up duplicated values in `project` and `asset`
      # dplyr::mutate(
      #   Project = unfill_vec(Project),
      #   Data = unfill_vec(Data)
      # ) |>
      dplyr::distinct() |>
      knitr::kable() |>
      paste0(collapse = "\n")
  )

  ## attach entities for 5 safes
  report_contents <- .markdown_report_body(
    report_contents = report_contents,
    overview_tbl = overview_data_all,
    safe_people_tbl = safe_people_all,
    safe_project_tbl = safe_project_all,
    safe_data_tbl = safe_data_all,
    safe_user_perm_tbl = safe_data_permissions_all,
    safe_setting_tbl = safe_setting_all,
    safe_output_tbl = safe_output_all,
    break_by = "server"
  )

  report_contents <- c(report_contents, "\n<hr />\n")

  ## append RO-Crates ----
  for (i in seq_along(x)) {
    # extract current RO-Crate
    rocrate <- x[i][[1]]
    # check if the input object has a `path` attribute
    if (!is.null(attr(rocrate, "path"))) {
      # check if any of the entities with `@type = 'File'` have empty `content`
      rocrate <- rocrate |>
        load_content(roc_path = attr(rocrate, "path"))
    }

    report_contents <- .markdown_report_rocrate(
      report_contents = report_contents,
      rocrate = rocrate,
      section_txt = ifelse(
        i == 1,
        paste0("## RO-Crates\n### '", names(x)[i], "' server"),
        paste0("### '", names(x)[i], "' server")
      ),
      max_line_length = max_line_length
    )
  }

  ## write contents inside file
  ### delete previous version
  if (overwrite) {
    unlink(filepath, recursive = TRUE, force = TRUE)
  }
  writeLines(report_contents, filepath)

  ## render document
  if (render) {
    if (tolower(doc_format) %in% c("html", "html_document")) {
      suppressWarnings({
        rmarkdown::render(
          filepath,
          "html_document",
          sub(".md$", ".html", filepath)
        )
        utils::browseURL(paste0("file://", sub(".md$", ".html", filepath)))
      })
    } else if (tolower(doc_format) %in% c("pdf", "pdf_document")) {
      suppressWarnings({
        rmarkdown::render(
          filepath,
          "pdf_document",
          sub(".md$", ".pdf", filepath)
        )
        utils::browseURL(paste0("file://", sub(".md$", ".pdf", filepath)))
      })
    } else {
      stop("The format `", doc_format, "` is not valid! Try 'html' or 'pdf'.")
    }
  } else {
    return(invisible(
      list(
        overview_diagram = overview_lst$diag_lst,
        overview_data = tidy_overview_tbl,
        safe_people = safe_people_all,
        safe_project_tbl = safe_project_all,
        safe_data = safe_data_all,
        safe_data_permissions = safe_data_permissions_all,
        safe_setting = safe_setting_all,
        safe_output = safe_output_all
      )
    ))
  }

  message("A report has been written to:\n ", filepath)

  # return list of data frames with Safe People, Projects, Data, etc.
  invisible(
    list(
      safe_people = safe_people_all,
      safe_project_tbl = safe_project_all,
      safe_data = safe_data_all,
      safe_data_permissions = safe_data_permissions_all,
      safe_setting = safe_setting_all,
      safe_output = safe_output_all
    )
  )
}

#' @param title String with title for the report (default: 'DataSHIELD Report').
#' @param filepath String with file path for Markdown report with the summary
#'     of the given object, `x`.
#' @param render Boolean flag to indicate whether to render the markdown report.
#' @param doc_format String with file format for the markdown report.
#' @param overwrite Boolean flag to indicate whether to overwrite a previous
#'     version of markdown report.
#' @param include_user_perm Boolean flag to indicate whether to include user
#'     permissions in the report overview's diagram.
#' @param diag_title String with title for the 'root' of the diagram (default:
#'     'DataSHIELD server').
#' @param diag_width Numeric value with width (in inches) for the report
#'     overview's diagram (default: `NULL`, estimated based on number of nodes).
#' @param diag_height Numeric value with height (in inches) for the report
#'     overview's diagram (default: `NULL`, estimated based on number of nodes).
#' @param max_line_length Integer with the maximum number of characters per line
#'     in the RO-Crate to be printed in the report.
#' @rdname report
#' @export
report.rocrate <- function(
  x,
  ...,
  title = "DataSHIELD Report",
  filepath = tempfile(fileext = ".md"),
  render = TRUE,
  doc_format = "html",
  overwrite = FALSE,
  include_user_perm = TRUE,
  diag_title = "DataSHIELD server",
  diag_width = NULL,
  diag_height = NULL,
  max_line_length = 200
) {
  # local bindings ----
  actionStatus <- asset <- asset_id <- description <- fx <- NULL
  encodingFormat <- id <- name <- permission <- perm_id <- project <- NULL
  table_name <- timestamp <- type <- person_id <- user <- NULL

  # validate RO-Crate ----
  rocrateR::is_rocrate(x)

  # ensure the given file path exists
  if (!dir.exists(dirname(filepath))) {
    stop(
      "The given file path directory does not exist. Try running the following",
      "command first:\n\t`mkdir -r",
      dirname(filepath),
      "`",
      call. = FALSE
    )
  }

  # check if the filepath points to an existing file
  if (file.exists(filepath)) {
    if (!overwrite) {
      stop(
        "The given file path points to an existing file. Try setting ",
        "`overwrite = TRUE` or changing the file path!",
        call. = FALSE
      )
    }
  }

  # check if the input object has a `path` attribute
  if (!is.null(attr(x, "path"))) {
    # check if any of the entities with `@type = 'File'` have empty `content`
    x <- x |>
      load_content(roc_path = attr(x, "path"))
  }

  # pre-processing ----
  ## Safe People ----
  # attempt to extract Safe People details
  safe_people_rocrate <- tryCatch(
    {
      extract_safe_people(x)
    },
    error = function(e) {
      NULL
    }
  )

  ## Safe Data ----
  # attempt to extract Safe Data details
  safe_data_rocrate <- tryCatch(
    {
      extract_safe_data(x)
    },
    error = function(e) {
      NULL
    }
  )

  # attempt to extract user permission entities
  user_perm_entity_lst <- tryCatch(
    {
      .get_entity(
        x,
        type = c("ReadAction", "WriteAction", "ControlAction")
      )
    },
    error = function(e) {
      NULL
    }
  )

  ## Safe Projects ----
  # extract project IDs from the Safe People RO-Crate
  member_of <- tryCatch(
    {
      safe_people_rocrate |>
        .get_entity(type = "Person") |>
        sapply(\(x) getElement(x, "memberOf")) |>
        unlist()
    },
    error = function(e) {
      NULL
    }
  )

  # attempt to extract Safe Project details
  safe_project_rocrate <- tryCatch(
    {
      extract_safe_project(x, id = member_of)
    },
    error = function(e) {
      NULL
    }
  )

  ## Safe Settings ----
  # attempt to extract Safe Setting details
  safe_setting_rocrate <- tryCatch(
    {
      extract_safe_setting(x)
    },
    error = function(e) {
      NULL
    }
  )

  ## Safe Outputs ----
  # attempt to extract Safe Output details
  safe_outputs_rocrate <- tryCatch(
    {
      extract_safe_output(x)
    },
    error = function(e) {
      NULL
    }
  )

  # verify contents of RO-Crate ----
  has_project_ents <- !is.null(safe_project_rocrate)
  has_data_ents <- !is.null(safe_data_rocrate)
  has_people_ents <- !is.null(safe_people_rocrate)

  # if not all these flags are `TRUE`, return error message
  if (!all(has_project_ents, has_data_ents, has_people_ents)) {
    stop(
      paste0(
        "The given RO-Crate is missing the following:",
        ifelse(has_project_ents, "", " - Project entity (`@type = 'Project'`)"),
        ifelse(has_data_ents, "", " - Data entity (`@type = 'Dataset'`)"),
        ifelse(has_people_ents, "", " - People entity (`type = 'Person'`)"),
        collapse = "\n"
      ),
      call. = FALSE
    )
  }

  # create overview ----
  overview_tbl <- tibble::tibble()
  ### extract (if available) table with user permissions
  user_perm_tbl <- flatten_user_perm_entity(user_perm_entity_lst)
  ### extract table with Safe People details
  safe_people_tbl <- flatten_safe_people(safe_people_rocrate)
  ### extract table with Safe Project details
  safe_project_tbl <- flatten_safe_project(safe_project_rocrate)
  ### extract table with Safe Data details
  safe_data_tbl <- flatten_safe_data(safe_data_rocrate)
  if (!is.null(user_perm_tbl) && nrow(user_perm_tbl) > 0) {
    overview_tbl <- user_perm_tbl |>
      # combine with Safe People details
      dplyr::left_join(safe_people_tbl, by = "person_id") |>
      # subset columns of interest
      dplyr::select(perm_id, person_id, name, asset_id, permission) |>
      # combine with Safe Data details
      dplyr::left_join(safe_data_tbl, by = "asset_id") |>
      # combine with Safe Project details
      dplyr::left_join(safe_project_tbl, by = c("asset_id", "asset")) #|>
    # # drop unused columns
    # dplyr::select(-id, -person_id, -asset_id, -type) |>
    # dplyr::rename(asset = asset_name)
  } else {
    overview_tbl <- flatten_safe_people(safe_people_rocrate) |>
      purrr::pmap(function(person_id, name, ...) {
        tibble::tibble(person_id, name) |>
          dplyr::bind_cols(flatten_safe_project(safe_project_rocrate))
      }) |>
      purrr::list_c()
  }

  # attempt extracting list of functions executed by the users from Safe Outputs
  safe_output_tbl <- flatten_safe_output(safe_outputs_rocrate) |>
    # extract only entities with mappings and functions, stored in CSV format
    dplyr::filter(encodingFormat == "text/csv") |>
    # extract content
    purrr::pluck("content", .default = list()) |>
    purrr::list_flatten() |>
    purrr::map(tibble::as_tibble) |>
    purrr::list_rbind()

  # initialise safe_output_tbl_v2, to be included in the returned outputs
  safe_output_tbl_v2 <- tibble::tibble()

  if (!is.null(safe_output_tbl) && nrow(safe_output_tbl) > 0) {
    # split `table` into `project` and `table`
    safe_output_tbl_v2 <- safe_output_tbl |>
      dplyr::mutate(
        project = gsub("(?=\\.).*$", "", table, perl = TRUE),
        asset = gsub("^.*(?<=\\.)", "", table, perl = TRUE)
      ) |>
      dplyr::distinct(project, asset, user, fx, timestamp)

    # append the list of functions to the overview table
    overview_tbl <- overview_tbl |>
      dplyr::left_join(
        safe_output_tbl_v2,
        by = c("project" = "project", "asset" = "asset", "name" = "user")
      ) |>
      # replace 'NA' in fx & timestamp with empty string
      dplyr::mutate(
        timestamp = dplyr::case_when(is.na(timestamp) ~ "", T ~ timestamp),
        fx = dplyr::case_when(is.na(fx) ~ "", T ~ fx)
      )
  }

  # overview diagram ----
  overview_lst <- .overview_diagram(
    overview_tbl,
    include_user_perm,
    filepath,
    render,
    diag_title,
    diag_width,
    diag_height
  )

  # create tidy overview table ----
  tidy_overview_tbl <- .tidy_overview(overview_tbl, include_user_perm)

  # create markdown report ----
  ## header and overview table
  report_contents <- c(
    .markdown_report_header(title, overview_tbl, overview_lst$diag_path),
    tidy_overview_tbl |>
      # # tidy up duplicated values in `project` and `table`
      # dplyr::mutate(
      #   Project = unfill_vec(Project),
      #   Data = unfill_vec(Data)
      # ) |>
      dplyr::distinct() |>
      knitr::kable() |>
      paste0(collapse = "\n")
  )

  ## attach entities for 5 safes
  report_contents <- .markdown_report_body(
    report_contents = report_contents,
    overview_tbl = overview_tbl,
    safe_people_tbl = flatten_safe_people(safe_people_rocrate),
    safe_project_tbl = flatten_safe_project(safe_project_rocrate),
    safe_data_tbl = flatten_safe_data(safe_data_rocrate),
    safe_user_perm_tbl = flatten_user_perm_entity(user_perm_entity_lst),
    safe_setting_tbl = flatten_safe_setting(safe_setting_rocrate),
    safe_output_tbl = safe_output_tbl_v2
  )

  report_contents <- c(report_contents, "\n<hr />\n")

  # append input RO-Crate ----
  report_contents <- .markdown_report_rocrate(
    report_contents = report_contents,
    rocrate = x,
    section_txt = "## RO-Crate",
    max_line_length = max_line_length
  )

  ## write contents inside file
  ### delete previous version
  if (overwrite) {
    unlink(filepath, recursive = TRUE, force = TRUE)
  }
  writeLines(report_contents, filepath)

  ## render document
  if (render) {
    if (tolower(doc_format) %in% c("html", "html_document")) {
      suppressWarnings({
        rmarkdown::render(
          filepath,
          "html_document",
          sub(".md$", ".html", filepath)
        )
        utils::browseURL(paste0("file://", sub(".md$", ".html", filepath)))
      })
    } else if (tolower(doc_format) %in% c("pdf", "pdf_document")) {
      suppressWarnings({
        rmarkdown::render(
          filepath,
          "pdf_document",
          sub(".md$", ".pdf", filepath)
        )
        utils::browseURL(paste0("file://", sub(".md$", ".pdf", filepath)))
      })
    } else {
      stop("The format `", doc_format, "` is not valid! Try 'html' or 'pdf'.")
    }
  } else {
    return(invisible(
      list(
        overview_diagram = overview_lst$diag_lst,
        overview_data = tidy_overview_tbl,
        safe_people = flatten_safe_people(safe_people_rocrate),
        safe_project = flatten_safe_project(safe_project_rocrate),
        safe_data = flatten_safe_data(safe_data_rocrate),
        safe_data_permissions = flatten_user_perm_entity(user_perm_entity_lst),
        safe_setting = flatten_safe_setting(safe_setting_rocrate),
        safe_output = safe_output_tbl_v2
      )
    ))
  }

  message("A report has been written to:\n ", filepath)

  # return list of data frames with Safe People, Projects, Data, etc.
  invisible(
    list(
      safe_people = flatten_safe_people(safe_people_rocrate),
      safe_project = flatten_safe_project(safe_project_rocrate),
      safe_data = flatten_safe_data(safe_data_rocrate),
      safe_data_permissions = flatten_user_perm_entity(user_perm_entity_lst),
      safe_setting = flatten_safe_setting(safe_setting_rocrate),
      safe_output = safe_output_tbl_v2
    )
  )
}

.extract_srvr_comp <- function(x, component) {
  tryCatch(
    {
      x |>
        # extract each component per server
        purrr::map(component) |>
        # lapply(getElement, name = component) |>
        # filter out NULL elements
        purrr::keep(\(x) !is.null(x)) |>
        # attach the server name as a new column
        purrr::imap(~ dplyr::mutate(.x, server = .y)) |>
        # combine rows
        dplyr::bind_rows()
    },
    error = function(e) {
      NULL
    }
  )
}

#' Create diagram for RO-Crate overview
#'
#' @param overview_tbl Data frame with overview details for the RO-Crate.
#' @inheritParams report
#'
#' @returns Diagram object
#' @keywords internal
#' @noRd
.overview_diagram <- function(
  overview_tbl,
  include_user_perm,
  filepath,
  render,
  diag_title,
  diag_width,
  diag_height
) {
  # local bindings
  fx <- name <- permission <- project <- asset <- NULL

  ## initialise `vars` and `labelvar`
  vars <- c("project", "asset")
  labelvar <- c(project = "Project", asset = "Data")

  # check if `overview_tbl` has `permission` field AND include_user_perm = TRUE
  if ("permission" %in% colnames(overview_tbl) && include_user_perm) {
    vars <- c(vars, "permission")
    labelvar <- c(labelvar, permission = "Access Level")
  }

  # attach name/user and label it 'People'
  vars <- c(vars, "name")
  labelvar <- c(labelvar, name = "People")

  # check if `overview_tbl` has a `fx` field
  if ("fx" %in% colnames(overview_tbl)) {
    vars <- c(vars, "fx")
    labelvar <- c(labelvar, fx = "Function")

    # replace 'NA' with empty string for `fx`
    overview_tbl <- overview_tbl |>
      dplyr::mutate(fx = ifelse(is.na(fx), "", fx))
  }

  # attach 'server' if found in `overview_tbl`
  if ("server" %in% colnames(overview_tbl)) {
    vars <- c("server", vars)
    labelvar <- c(server = "Server", labelvar)
  }

  # aggregate the data if `permission` exists in `overview_tbl`
  if ("permission" %in% colnames(overview_tbl)) {
    overview_agg <- overview_tbl |>
      dplyr::group_by(dplyr::pick(vars[vars != "permission"])) |>
      dplyr::reframe(
        permission = paste0(unique(permission), collapse = " & "),
      )
  } else {
    overview_agg <- overview_tbl
  }

  ## generate diagram
  diagram_lst <- overview_agg |>
    vtree::vtree(
      vars = vars,
      labelvar = labelvar,
      showpct = FALSE,
      showcount = FALSE,
      horiz = FALSE,
      varnamebold = TRUE,
      splitwidth = 1,
      vsplitwidth = 1,
      folder = dirname(filepath),
      title = diag_title,
      # imageFileOnly = render,
      pngknit = render,
      # pxheight = min(80 * nrow(overview_tbl), 500),
      # pxwidth = 200 * nrow(overview_tbl),
      prune = list(fx = "")
    )

  ## if `render = TRUE`, then render diagram as PNG
  diagram_filepath <- NULL
  ### estimate number of nodes
  dot <- diagram_lst$x$diagram
  nodes <- dot |>
    gregexpr(pattern = "Node_[A-Za-z0-9_]+", perl = TRUE) |>
    (\(.) regmatches(dot, .))() |>
    unlist() |>
    unique()
  num_nodes <- length(nodes)
  ### scale width/height based on nodes
  scale_factor <- 0.5 # inches per node
  if (is.null(diag_width)) {
    width <- max(12, num_nodes * scale_factor)
  } else {
    width <- diag_width
  }
  if (is.null(diag_height)) {
    height <- max(6, num_nodes * scale_factor)
  } else {
    height <- diag_height
  }
  if (render) {
    diagram_filepath <- diagram_lst |>
      vtree::grVizToImageFile(
        folder = dirname(filepath),
        filename = gsub("md$", "png", basename(filepath))
      )
  }

  # # find path to latest PNG generated with `vtree`
  # diagram_filepath <- list.files(dirname(filepath), "^vtree")
  # diagram_filepath <- diagram_filepath[length(diagram_filepath)]

  return(list(diag_lst = diagram_lst, diag_path = diagram_filepath))
}

#' Create tidy version of the overview table
#'
#' @inheritParams .overview_diagram
#'
#' @returns Data frame with tidy overview table.
#' @keywords internal
#' @noRd
.tidy_overview <- function(overview_tbl, include_user_perm) {
  # local bindings
  fx <- permission <- timestamp <- NULL

  ## initialise `vars` and `varslab`
  vars <- c("project", "asset")
  varslab <- c("Project" = "project", "Data" = "asset")

  # check if `overview_tbl` has `permission` field AND include_user_perm = TRUE
  if ("permission" %in% colnames(overview_tbl) && include_user_perm) {
    vars <- c(vars, "permission")
    varslab <- c(varslab, "Access Level" = "permission")
  }

  # attach name/user and label it 'People'
  vars <- c(vars, "name")
  varslab <- c(varslab, "People" = "name")

  # check if `overview_tbl` has a `fx` field
  if ("fx" %in% colnames(overview_tbl)) {
    vars <- c(vars, "fx", "timestamp")
    varslab <- c(
      varslab,
      "Function" = "fx",
      "Timestamp" = "timestamp"
    )

    # replace 'NA' with empty string for `fx` and `timestamp`
    overview_tbl <- overview_tbl |>
      dplyr::mutate(
        fx = ifelse(is.na(fx), "", fx),
        timestamp = ifelse(is.na(timestamp), "", timestamp)
      )
  }

  # attach 'server' if found in `overview_tbl`
  if ("server" %in% colnames(overview_tbl)) {
    vars <- c("server", vars)
    varslab <- c("Server" = "server", varslab)
  }

  # create tidy data frame
  tidy_overview_tbl <- overview_tbl |>
    dplyr::distinct(dplyr::pick(dplyr::all_of(vars)))

  if ("permission" %in% colnames(overview_tbl)) {
    if ("fx" %in% colnames(overview_tbl)) {
      vars_agg <- vars[!vars %in% c("permission", "fx", "timestamp")]
      tidy_overview_tbl <- tidy_overview_tbl |>
        dplyr::group_by(dplyr::pick(vars_agg)) |>
        dplyr::reframe(
          permission = paste0(unique(permission), collapse = " & "),
          fx = fx,
          timestamp = timestamp,
          .groups = "drop"
        )
    } else {
      tidy_overview_tbl <- tidy_overview_tbl |>
        dplyr::group_by(dplyr::pick(vars[vars != "permission"])) |>
        dplyr::reframe(
          permission = paste0(unique(permission), collapse = " & ")
        )
    }
  }

  tidy_overview_tbl <- tidy_overview_tbl |>
    dplyr::select(dplyr::all_of(varslab)) |>
    dplyr::distinct()

  return(tidy_overview_tbl)
}


#' Generate Markdown report's header
#'
#' @inheritParams .overview_diagram
#' @inheritParams report
#'
#' @returns String with report's header
#' @keywords internal
#' @noRd
.markdown_report_header <- function(title, overview_tbl, diagram_filepath) {
  # initialise variables for the report header
  unique_users_vct <- unique(c(
    getElement(overview_tbl, "name"),
    getElement(overview_tbl, "user")
  ))
  unique_project_vct <- unique(getElement(overview_tbl, "project"))
  unique_servers_vct <- unique(getElement(overview_tbl, "server"))
  ## initialise markdown header (see https://rmarkdown.rstudio.com/lesson-9.html)
  paste0(
    "---\n",
    paste0("title: ", title),
    "\noutput:\n",
    "  pdf_document:\n",
    "    df_print: kable\n",
    "    highlight: tango\n",
    "  html_document: default\n",
    "fontsize: 11pt\n",
    "geometry:\n",
    "  - margin=.5in\n",
    "  - landscape\n",
    "header-includes: \n",
    "  - \\usepackage{array}\n",
    "  - \\usepackage{longtable}\n",
    "  - \\usepackage{xurl}\n",
    "  - \\usepackage{hyphenat}\n",
    "  - \\usepackage{microtype}\n",
    "  - \\sloppy\n",
    "  - \\setlength{\\emergencystretch}{3em}\n",
    "  - \\usepackage{float}\n",
    "  - \\usepackage{titling}\n",
    "  - \\setlength{\\droptitle}{-1.5cm}\n",
    "  - \\usepackage[breakable]{tcolorbox}\n",
    "  - \\usepackage{tcolorbox}\n",
    "  - \\tcbuselibrary{breakable}\n",
    "  - \\renewenvironment{Shaded}{\\begin{tcolorbox}[breakable,boxrule=0pt,",
    "colback=gray!10]}{\\end{tcolorbox}}\n",
    paste0("date: ", format(Sys.time(), '%Y-%m-%d %H:%M:%S')),
    "\n---\n",
    paste0(
      "This report contains details for ",
      length(unique_users_vct),
      " user",
      ifelse(length(unique_users_vct) > 1, "s", ""),
      " and ",
      length(unique_project_vct),
      " project",
      ifelse(length(unique_project_vct) > 1, "s", ""),
      ". In addition, the tables they have access to within",
      ifelse(length(unique_project_vct) > 1, " a ", " the "),
      "project. \n",
      ifelse(
        "server" %in% colnames(overview_tbl),
        paste0(
          "\n**Note:** the data shown in this report was extracted from ",
          length(unique_servers_vct),
          " server",
          ifelse(length(unique_servers_vct) > 1, "s", ""),
          "."
        ),
        ""
      ),
      "\n\n"
    ),
    "## Overview\n\n",
    "<div style=\"margin:0;\">\n",
    # "::: {style=\"text-align: center;\"}\n",
    "<!-- PDF-only -->\n",
    "```{=latex}\n",
    paste0(
      "\\begin{figure}[H]\n",
      "\\centering\n",
      "\\includegraphics[height=0.75\\textheight, keepaspectratio, width=0.9\\textwidth]{",
      diagram_filepath,
      "}",
      "\\caption{RO-Crate Overview}\n",
      "\\end{figure}\n```\n"
    ),
    "<!-- HTML-only -->",
    "```{=html}\n",
    paste0(
      "<img src=\"",
      diagram_filepath,
      "\" alt=\"RO-Crate Overview\" ",
      "style=\"display:block; margin-left:auto; margin-right:auto;\" />\n"
    ),
    "```\n</div>\n\n\\newpage"
  )
}

#' Generate Markdown report's body
#'
#' @inheritParams .overview_diagram
#' @param report_contents String with Markdown report (e.g., header).
#' @param safe_people_tbl Data frame with Safe People details.
#' @param safe_project_tbl Data frame with Safe Project details.
#' @param safe_data_tbl Data frame with Safe Data details.
#' @param safe_user_perm_tbl Data frame with Safe Data user permissions.
#' @param safe_setting_tbl Data frame with Safe Setting details.
#' @param safe_output_tbl Data frame with Safe Output details.
#' @param break_by Optional string with variable to be used for breaking down
#'     each table (e.g., server), as opposed to display all the results in a
#'     single table.
#'
#' @returns String with updated Markdown report.
#' @keywords internal
#' @noRd
.markdown_report_body <- function(
  report_contents,
  overview_tbl,
  safe_people_tbl,
  safe_project_tbl,
  safe_data_tbl,
  safe_user_perm_tbl,
  safe_setting_tbl,
  safe_output_tbl,
  break_by = NULL
) {
  # local bindings
  actionStatus <- description <- permission <- NULL

  report_contents <- c(report_contents, "\n<hr />\n", "## Entities")

  ## append Safe People details
  report_contents <- c(
    report_contents,
    "\n### People\n",
    safe_people_tbl |>
      .break_tibble(varname = break_by) |>
      paste0(collapse = "\n")
  )

  ## append Safe Project & Safe Data details
  report_contents <- c(
    report_contents,
    "### Project(s)\n",
    safe_project_tbl |>
      .break_tibble(varname = break_by) |>
      paste0(collapse = "\n")
  )

  ## append Safe Data details
  report_contents <- c(
    report_contents,
    "### Data\n",
    safe_data_tbl |>
      .break_tibble(varname = break_by) |>
      paste0(collapse = "\n")
  )

  ## append Safe Data permissions
  #### check if `overview_tbl` has `permission` field
  if ("permission" %in% colnames(overview_tbl)) {
    report_contents <- c(
      report_contents,
      "#### Data permissions\n",
      safe_user_perm_tbl |>
        dplyr::select(-actionStatus, -description, -permission) |>
        .break_tibble(varname = break_by) |>
        paste0(collapse = "\n")
    )
  }

  ## append Safe Settings details
  report_contents <- c(
    report_contents,
    "### Settings\n",
    safe_setting_tbl |>
      .break_tibble(varname = break_by) |>
      paste0(collapse = "\n")
  )

  if (!is.null(safe_output_tbl) && nrow(safe_output_tbl) > 0) {
    ## append Safe Outputs details
    report_contents <- c(
      report_contents,
      "### Outputs\n",
      safe_output_tbl |>
        .break_tibble(varname = break_by) |>
        paste0(collapse = "\n")
    )
  }

  return(report_contents)
}

#' Embed RO-Crate in Markdown report
#'
#' @inheritParams .markdown_report_body
#' @param rocrate RO-Crate object (see [rocrateR::rocrate]).
#' @param section_txt String with to be used as the section header
#'     (e.g., RO-Crate).
#'
#' @returns String with update Markdown report.
#' @keywords internal
#' @noRd
.markdown_report_rocrate <- function(
  report_contents,
  rocrate,
  section_txt,
  max_line_length = 200
) {
  # save the input into intermediate JSON file
  tmp_file <- tempfile(fileext = ".json")
  # delete temporary file
  on.exit(unlink(tmp_file, recursive = TRUE, force = TRUE))
  # store RO-Crate in JSON format
  jsonlite::write_json(
    rocrate,
    path = tmp_file,
    pretty = TRUE,
    auto_unbox = TRUE
  )
  # load formatted RO-Crate as text
  rocrate_txt <- readLines(tmp_file) |>
    # shorten long lines, beyond `max_line_length`
    sapply(function(x) {
      if (nchar(x) > max_line_length && max_line_length > 10) {
        return(paste0(
          substr(x, 1, max_line_length - 10),
          "... <line truncated> ...",
          substr(x, nchar(x) - 10, nchar(x))
        ))
      }
      return(x)
    })
  report_contents <- c(
    report_contents,
    "\n\\newpage\n",
    section_txt,
    "\n```json",
    # display formatted RO-Crate
    rocrate_txt,
    "\n```"
  )
}

#' Break tibble by group, `varname`
#'
#' @param df Data frame to be broken down into groups.
#' @param varname String with variable name
#'
#' @returns String with data frame rendered with `kable`.
#' @keywords internal
#' @noRd
.break_tibble <- function(df, varname) {
  # local bindings
  temp <- NULL

  # if `varname` is NULL or not a column in `df`, render all the data
  if (is.null(varname) || !(varname %in% colnames(df))) {
    return(knitr::kable(df))
  }

  # extract unique values for `varname` in `df`
  unique_vals <- unique(getElement(df, varname))

  # render data frame based on groups
  sapply(unique_vals, function(v) {
    c(
      paste0("\n#### '", v, "' ", varname, "\n"),
      df |>
        dplyr::rename(temp = !!varname) |>
        dplyr::filter(temp == v) |>
        dplyr::select(-temp) |>
        # drop out empty columns, equivalent to `janitor::remove_empty`
        Filter(f = function(x) any(!is.na(x))) |>
        knitr::kable()
    ) |>
      paste0(collapse = "\n")
  })
}

test_that("report.default errors", {
  expect_error(
    report(123),
    "Unknown class"
  )
})

test_that("report.character errors for invalid path", {
  expect_error(
    report("nonexistent_path", render = FALSE),
    "The provided path does not exist"
  )
})

test_that("report.character errors if read_rocrate fails", {
  # create temporary file
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  tmp_file <- tempfile(tmpdir = tmp_dir)
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))
  file.create(tmp_file)

  testthat::with_mocked_bindings(
    read_rocrate = function(...) stop("fail"),
    code = {
      expect_error(
        report(tmp_file, render = FALSE),
        "Could not determine how to load RO-Crate from provided input"
      )
    },
    .package = "rocrateR"
  )
})

test_that("report.character errors if input is invalid .zip", {
  # create temporary file
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  tmp_file <- tempfile(tmpdir = tmp_dir, fileext = ".zip")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))
  file.create(tmp_file)

  expect_error(report(tmp_file, render = FALSE))
})

test_that("report.character dispatches to rocrate method", {
  # create temporary file
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))
  json_file <- file.path(tmp_dir, "ro-crate-metadata.json")
  writeLines("{}", json_file)

  fake_rocrate <- structure(list(), class = "rocrate")

  testthat::local_mocked_bindings(
    read_rocrate = function(...) fake_rocrate,
    .package = "rocrateR"
  )
  testthat::with_mocked_bindings(
    load_content = function(x, ...) x,
    report.character = function(x, ...) "OK",
    code = {
      result <- report(tmp_dir, render = FALSE)
      expect_equal(result, "OK")
    },
    .package = "dsROCrate"
  )
})

test_that("report.rocrate errors if required entities missing", {
  fake_rocrate <- structure(list(), class = "rocrate")

  testthat::local_mocked_bindings(
    is_rocrate = function(...) TRUE,
    .package = "rocrateR"
  )

  testthat::with_mocked_bindings(
    extract_safe_people = function(...) NULL,
    extract_safe_project = function(...) NULL,
    extract_safe_data = function(...) NULL,
    code = {
      expect_error(
        report(fake_rocrate, render = FALSE),
        "missing"
      )
    },
    .package = "dsROCrate"
  )
})

test_that("report.rocrate returns expected structure", {
  fake_rocrate <- structure(list(), class = "rocrate")

  safe_people <- list()
  safe_project <- list()
  safe_data <- list()

  testthat::local_mocked_bindings(
    is_rocrate = function(...) TRUE,
    .package = "rocrateR"
  )

  testthat::with_mocked_bindings(
    extract_safe_people = function(...) safe_people,
    extract_safe_project = function(...) safe_project,
    extract_safe_data = function(...) safe_data,
    extract_safe_setting = function(...) NULL,
    extract_safe_output = function(...) NULL,
    flatten_safe_people = function(...) {
      tibble::tibble(person_id = "u1", name = "dsuser", project = "P1")
    },
    flatten_safe_project = function(...) {
      tibble::tibble(
        project_id = "p1",
        project = "P1",
        asset_id = "t1",
        asset = "T1"
      )
    },
    flatten_safe_data = function(...) {
      tibble::tibble(asset_id = "t1", asset = "T1")
    },
    flatten_safe_output = function(...) tibble::tibble(),
    flatten_user_perm_entity = function(...) NULL,
    .overview_diagram = function(...) list(diag_lst = "diag", diag_path = NULL),
    .tidy_overview = function(...) tibble::tibble(project = "P1"),
    .markdown_report_header = function(...) "HEADER",
    .markdown_report_body = function(...) "BODY",
    .markdown_report_rocrate = function(...) "ROCRATE",
    code = {
      result <- report(
        fake_rocrate,
        render = FALSE,
        filepath = tempfile(fileext = ".md")
      )

      expect_type(result, "list")
      expect_named(
        result,
        c(
          "overview_diagram",
          "overview_data",
          "safe_people",
          "safe_project",
          "safe_data",
          "safe_data_permissions",
          "safe_setting",
          "safe_output"
        )
      )
    },
    .package = "dsROCrate"
  )
})

test_that("report.rocrate prevents overwrite", {
  fake_rocrate <- structure(list(), class = "rocrate")
  # create temporary file
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  tmp_file <- tempfile(tmpdir = tmp_dir, fileext = ".md")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))
  file.create(tmp_file)

  testthat::local_mocked_bindings(
    is_rocrate = function(...) TRUE,
    .package = "rocrateR"
  )

  testthat::with_mocked_bindings(
    extract_safe_people = function(...) list(),
    extract_safe_project = function(...) list(),
    extract_safe_data = function(...) list(),
    code = {
      expect_error(
        report(fake_rocrate, filepath = tmp_file, overwrite = FALSE),
        "existing file"
      )
    },
    .package = "dsROCrate"
  )
})

test_that("report.list aggregates multiple servers", {
  fake_rocrate <- structure(list(), class = "rocrate")

  server_output <- list(
    safe_people = tibble::tibble(id_person = "u1", name = "dsuser"),
    safe_project = tibble::tibble(
      id_project = "p1",
      project = "P1",
      asset_id = "t1",
      asset = "T1"
    ),
    safe_data = tibble::tibble(asset_id = "t1", asset = "T1"),
    safe_data_permissions = NULL,
    safe_setting = tibble::tibble(),
    safe_output = tibble::tibble(),
    overview_data = tibble::tibble(
      project = "P1",
      asset = "T1",
      person = "dsuser"
    )
  )

  testthat::local_mocked_bindings(
    is_rocrate = function(...) TRUE,
    .package = "rocrateR"
  )

  testthat::with_mocked_bindings(
    report = function(...) server_output,
    .overview_diagram = function(...) list(diag_lst = "diag", diag_path = NULL),
    .tidy_overview = function(...) tibble::tibble(Project = "P1"),
    .markdown_report_header = function(...) "HEADER",
    .markdown_report_body = function(...) "BODY",
    .markdown_report_rocrate = function(...) "ROCRATE",
    code = {
      result <- report(
        list(server1 = fake_rocrate, server2 = fake_rocrate),
        study_name = "StudyX",
        render = FALSE,
        filepath = tempfile(fileext = ".md")
      )

      expect_type(result, "list")
      expect_true("safe_people" %in% names(result))
    },
    .package = "dsROCrate"
  )
})

test_that(".tidy_overview collapses permissions correctly", {
  df <- tibble::tibble(
    project = "P1",
    asset = "T1",
    name = "dsuser",
    permission = c("read", "write")
  )

  result <- .tidy_overview(df, include_user_perm = TRUE)

  expect_true("Access Level" %in% colnames(result))
  expect_match(result$`Access Level`, "read")
})

test_that(".break_tibble splits by group", {
  df <- tibble::tibble(
    server = c("A", "B"),
    value = 1:2
  )

  result <- .break_tibble(df, "server")

  expect_length(result, 2)
})

test_that(".break_tibble returns input table when varname = NULL", {
  df <- tibble::tibble(
    server = c("A", "B"),
    value = 1:2
  )

  expect_equal(knitr::kable(df), .break_tibble(df, NULL))
})

test_that("report fails if the given path points to a non-existing directory", {
  expect_error(
    report(rocrateR::rocrate_5s(), filepath = "path/to/dir")
  )
})

test_that("report works end-to-end with real dsROCrate audit outputs", {
  # create temporary file
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))

  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  # generate audit RO-Crate for a Safe Project
  # suppress warnings about missing logs
  suppressWarnings({
    roc <- audit(
      opal_con,
      project = "CNSIM",
      path = tmp_dir
    )
  })

  out_file <- file.path(tmp_dir, "report.md")

  # ignore warnings about missing permission entities (e.g., @type = 'ControlAction')
  suppressWarnings(
    result <- report(
      roc,
      filepath = out_file,
      render = FALSE,
      overwrite = TRUE
    )
  )

  # ---- assertions ----
  expect_true(file.exists(out_file))
  expect_type(result, "list")

  expect_true("safe_people" %in% names(result))
  expect_true("safe_project" %in% names(result))
  expect_true("safe_data" %in% names(result))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("report.character loads crate from disk", {
  # create temporary file
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))

  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  # generate audit RO-Crate for a Safe Project
  # suppress warnings about missing logs
  suppressWarnings({
    roc <- audit(
      opal_con,
      project = "CNSIM",
      path = tmp_dir
    )
  })

  # write crate to disk (real metadata file)
  path_to_rocrate_bag <- rocrateR::bag_rocrate(roc, path = tmp_dir)

  out_file <- file.path(tmp_dir, "report.md")

  # ignore warnings about missing permission entities (e.g., @type = 'ControlAction')
  suppressWarnings(
    result <- report(
      path_to_rocrate_bag,
      filepath = out_file,
      render = FALSE,
      overwrite = TRUE
    )
  )

  # ---- assertions ----
  expect_true(file.exists(out_file))
  expect_type(result, "list")

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("report.list aggregates outputs from a study audit", {
  # create temporary file
  tmp_dir <- file.path(tempdir(), "dsROCRate_tests")
  dir.create(tmp_dir, recursive = TRUE)
  on.exit(unlink(tmp_dir, recursive = TRUE, force = TRUE))

  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  # generate audit RO-Crate for a study
  ## ignore warnings about missing logs
  suppressWarnings(
    roc <- audit(
      list(demo = opal_con),
      project = "CNSIM",
      path = tmp_dir
    )
  )

  out_file <- file.path(tmp_dir, "report.md")

  # ignore warnings about missing permission entities (e.g., @type = 'ControlAction')
  suppressWarnings(
    result <- report(
      roc,
      filepath = out_file,
      study_name = "StudyX",
      render = FALSE,
      overwrite = TRUE
    )
  )

  # ---- assertions ----
  expect_true(file.exists(out_file))
  expect_type(result, "list")

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("report.list handles missing outputs from a study audit", {
  testthat::with_mocked_bindings(
    report = function(...) list(),
    code = {
      result <- report(
        list(server1 = rocrateR::rocrate_5s()),
        study_name = "StudyX",
        render = FALSE,
        filepath = tempfile(fileext = ".md")
      )

      expect_type(result, "list")
    },
    .package = "dsROCrate"
  )
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  cache = FALSE,
  out.width = "100%",
  cache = FALSE,
  screenshot.opts = list(vwidth = 2000, vheight = 600, zoom = 3, selector = "div.html-widget")
)

# save the built-in output hook
hook_output <- knitr::knit_hooks$get("output")
hook_message <- knitr::knit_hooks$get("message")

# define a truncation helper
truncate_lines_tail <- function(x, lines) {
  x <- unlist(strsplit(x, "\n"))
  more <- "..."
  if (length(lines) == 1) {
    if (length(x) > lines) {
      # truncate the output, but add `more`
      # x <- c(head(x, n = lines), more)
      x <- c(more, tail(x, n = lines))
    }
  } else {
    x <- c(more, x[lines], more)
  }
  paste(c(x, ""), collapse = "\n")
}

# set the new hooks
knitr::knit_hooks$set(
  output = function(x, options) {
    lines <- options$out.lines
    if (is.null(lines)) return(hook_output(x, options))
    x <- truncate_lines_tail(x, lines)
    return(hook_output(x, options))
  },
  message = function(x, options) {
    lines <- options$out.lines
    if (is.null(lines)) return(hook_message(x, options))
    x <- truncate_lines_tail(x, lines)
    return(hook_message(x, options))
  }
)

# helper function to get the number of rows of an RO-Crate
rocrate_lines <- function(rocrate) {
  rocrate_txt(rocrate) |>
    length()
}

# helper function to read RO-Crate as text file
rocrate_txt <- function(rocrate) {
  # create tmp file to write a JSON file with the output
  tmp_file <- tempfile(fileext = ".json")
  on.exit(unlink(tmp_file, force = TRUE))
  # write RO-Crate in tmp file
  rocrateR::write_rocrate(rocrate, tmp_file)
  # read RO-Crate as text file
  readLines(tmp_file)
}

## ----setup--------------------------------------------------------------------
library(dsROCrate)

# show all the lines in the RO-Crate
oopt <- options(max_lines = Inf)
on.exit(options(oopt), add = TRUE)

## ----eval = FALSE-------------------------------------------------------------
# vignette("deploy-local-datashield-server-with-opal", package = "dsROCrate")

## -----------------------------------------------------------------------------
# define global variables
## Opal server access
USERNAME <- "administrator"
USERPASS <- "password"
SERVER <- "https://opal-demo.obiba.org"
## Credentials for `dsuser`
### NOTE: this is only used to simulate an analysis and generate logs
DSUSERPASS <- "P@ssw0rd"

## -----------------------------------------------------------------------------
## Five safes variables
PEOPLE <- "dsuser"
PROJECT <- "CNSIM"
TABLES <- c("CNSIM1")

## -----------------------------------------------------------------------------
# login to local server with `USERNAME` and `USERPASS`.
o <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
)

print(o)

## ----eval = FALSE, echo = TRUE------------------------------------------------
# # install.packages("pak")
# pak::pak("rocrateR")
# 
# # for development version use
# pak::pak("ResearchObject/ro-crate-r@dev")

## -----------------------------------------------------------------------------
basic_rocrate <- rocrateR::rocrate_5s()

## -----------------------------------------------------------------------------
print(basic_rocrate)

## ----out.lines=-(rocrate_lines(basic_rocrate) - 2)----------------------------
basic_rocrate <- o |>
  dsROCrate::safe_data(rocrate = basic_rocrate,
                       project = PROJECT,
                       tables = TABLES)

print(basic_rocrate) # note that the output will be truncated

## ----out.lines=-(rocrate_lines(basic_rocrate) - 2)----------------------------
basic_rocrate <- o |>
  dsROCrate::safe_project(rocrate = basic_rocrate,
                          project = PROJECT)

print(basic_rocrate) # note that the output will be truncated

## ----safe_people, out.lines=-(rocrate_lines(basic_rocrate) + 1)---------------
basic_rocrate <- o |>
  dsROCrate::safe_people(rocrate = basic_rocrate, user = PEOPLE)

print(basic_rocrate) # note that the output will be truncated

## ----eval = FALSE-------------------------------------------------------------
# # close previous connection
# opalr::opal.logout(o)
# 
# # open new connection as administrator
# o <- opalr::opal.login(
#   username = "administrator",
#   password = "password",
#   url = SERVER
# )

## ----out.lines=-(rocrate_lines(basic_rocrate) - 2)----------------------------
basic_rocrate <- o |>
  dsROCrate::safe_setting(rocrate = basic_rocrate)

print(basic_rocrate) # note that the output will be truncated

## ----eval = FALSE-------------------------------------------------------------
# # close previous connection
# opalr::opal.logout(o)
# 
# # open new connection as administrator
# o <- opalr::opal.login(
#   username = "administrator",
#   password = "password",
#   url = SERVER
# )

## ----eval = FALSE-------------------------------------------------------------
# pak::pak("DSI")
# pak::pak("DSOpal")
# pak::pak("dsBaseClient")

## ----eval = TRUE--------------------------------------------------------------
# run some test commands with dsBaseClient
## needed to defined the OpalDriver class in the current environment
DSOpal::Opal()
## create new login object, note that here we use the `dsuser`
builder <- DSI::newDSLoginBuilder()
builder$append(server = "study1",
               url = SERVER,
               user = "dsuser",
               password = DSUSERPASS,
               driver = "OpalDriver")
logindata <- builder$build()
conns <- DSI::datashield.login(logins = logindata)

## ----eval = TRUE--------------------------------------------------------------
## assign data
DSI::datashield.assign.table(conns["study1"], 
                             symbol = "dsROCrate_test",
                             table = paste0(PROJECT, ".", TABLES[1]),
                             errors.print = TRUE)

dsBaseClient::ds.ls(datasources = conns["study1"])
dsBaseClient::ds.summary("dsROCrate_test")

## ----echo = FALSE, warning = FALSE, message = FALSE---------------------------
# check if there are any logs available, if not simulate some operations

## ----safe_outputs_internal, echo=FALSE----------------------------------------
lines_before_safe_outputs <- rocrate_lines(basic_rocrate)

## ----safe_outputs-------------------------------------------------------------
basic_rocrate <- o |>
  dsROCrate::safe_output(rocrate = basic_rocrate,
                         logs_from = Sys.time() - 60, # capture the last minute
                         logs_to = Sys.time())

## ----out.lines=-(lines_before_safe_outputs + 1)-------------------------------
print(basic_rocrate) # note that the output will be truncated

## -----------------------------------------------------------------------------
opalr::opal.logout(o)

## -----------------------------------------------------------------------------
# create temp directory
tmp_path_bag <- file.path(tempdir(), "dsROCrate-getting-started")
dir.create(tmp_path_bag, showWarnings = FALSE)

# create RO-Crate bag
path_to_rocrate_bag <- basic_rocrate |>
  rocrateR::bag_rocrate(path = tmp_path_bag, overwrite = TRUE)

## -----------------------------------------------------------------------------
# extract files in temporary directory
path_to_rocrate_bag |>
  # extract contents inside /tmp_path_bag/ROC
  rocrateR::unbag_rocrate(output = file.path(tmp_path_bag, "ROC"), quiet = TRUE) |>
  # create tree with the files
  fs::dir_tree()

## -----------------------------------------------------------------------------
unlink(tmp_path_bag, recursive = TRUE, force = TRUE)

## ----warning=FALSE------------------------------------------------------------
safe_people_crate_v1 <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
) |>
  dsROCrate::audit(user = "dsuser", project = "CNSIM")

print(safe_people_crate_v1)

## ----safe_people_crate_audit_v1-----------------------------------------------
safe_people_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_people_crate_contents <- safe_people_crate_v1 |>
  dsROCrate::report(filepath = safe_people_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_people_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_people_crate_contents$overview_data |>
  knitr::kable()

## ----eval = FALSE-------------------------------------------------------------
# safe_people_crate_v1 |>
#   dsROCrate::report(filepath = safe_people_crate_v1_rmd,
#                             title = "DataSHIELD Safe People - Audit Report",
#                             render = TRUE,
#                             overwrite = TRUE)

## ----warning=FALSE------------------------------------------------------------
safe_project_crate_v1 <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
) |>
  dsROCrate::audit(project = "CNSIM")

print(safe_project_crate_v1)

## ----safe_project_crate_audit_v1----------------------------------------------
safe_project_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_project_crate_contents <- safe_project_crate_v1 |>
  dsROCrate::report(filepath = safe_project_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_project_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_project_crate_contents$overview_data |>
  knitr::kable()

## ----eval = FALSE-------------------------------------------------------------
# safe_project_crate_v1 |>
#   dsROCrate::report(filepath = safe_project_crate_v1_rmd,
#                             title = "DataSHIELD Safe Project - Audit Report",
#                             render = TRUE,
#                             overwrite = TRUE)

## ----warning=FALSE------------------------------------------------------------
study_crate_v1 <- 
  list(
    "opal_test" = opalr::opal.login(
      username = USERNAME,
      password = USERPASS,
      url = "https://opal-test.obiba.org"
    ),
    "opal_demo" = opalr::opal.login(
      username = USERNAME,
      password = USERPASS,
      url = "https://opal-demo.obiba.org"
    )
  ) |>
  dsROCrate::audit(project = "CNSIM")

print(study_crate_v1)

## ----study_crate_audit_v1-----------------------------------------------------
study_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_project_crate_contents <- study_crate_v1 |>
  dsROCrate::report(filepath = study_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_project_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_project_crate_contents$overview_data |>
  knitr::kable()

## ----eval = FALSE-------------------------------------------------------------
# study_crate_v1 |>
#   dsROCrate::report(filepath = study_crate_v1_rmd,
#                             title = "DataSHIELD Study audit",
#                             render = TRUE,
#                             overwrite = TRUE)

## ----echo = FALSE, message = FALSE, error = FALSE-----------------------------
unlink(safe_people_crate_v1_rmd, TRUE, TRUE)
unlink(safe_project_crate_v1_rmd, TRUE, TRUE)
unlink(study_crate_v1_rmd, TRUE, TRUE)
# reverse options
options(oopt)

#' Add Safe People entities to RO-Crate
#'
#' Creates Person entities and links them to their Projects using
#' the `memberOf` property.
#'
#' @param rc RO-Crate object, see [rocrateR::rocrate].
#' @param people_tbl Tibble with user metadata.
#'   Columns: username, given_name, family_name, email, affiliation.
#' @param membership_tbl Tibble mapping users to projects.
#'   Columns: username, project.
#'
#' @return Updated RO-Crate
#' @noRd
add_safe_people_entities_cr8tor <- function(rc, people_tbl, membership_tbl) {
  # local bindings
  project <- username <- NULL

  for (i in seq_len(nrow(people_tbl))) {
    p <- people_tbl[i, ]
    person_id <- id_hash("#person:", p$username)

    # projects user belongs to
    projs <- membership_tbl |>
      dplyr::filter(username == p$username) |>
      dplyr::pull(project) |>
      unique()

    memberOf <- purrr::map(projs, \(pr) {
      list(`@id` = id_hash("#project:", pr))
    })

    rc <- rc |>
      rocrateR::add_entity(
        rocrateR::entity(
          id = person_id,
          type = "Person",
          name = p$username,
          givenName = p$given_name,
          familyName = p$family_name,
          email = p$email,
          affiliation = p$affiliation,
          memberOf = memberOf
        )
      )
  }

  rc
}

#' Add Safe Project entities
#'
#' Creates Project entities and links datasets using `hasPart`.
#'
#' @param rc RO-Crate object, see [rocrateR::rocrate].
#' @param proj_tbl Tibble with project metadata.
#'   Columns: id, name.
#' @param data_tbl Tibble with dataset metadata.
#'   Columns: project, asset.
#'
#' @return Updated RO-Crate
#' @noRd
add_safe_project_entities_cr8tor <- function(rc, proj_tbl, data_tbl) {
  # local bindings
  asset <- project <- NULL

  now <- format(Sys.time(), "%Y-%m-%dT%H:%M:%OS3Z", tz = "UTC")

  for (i in seq_len(nrow(proj_tbl))) {
    project_id <- proj_tbl$project_id[i]
    project <- proj_tbl$name[i]

    ds_ids <- data_tbl |>
      dplyr::filter(project == project_id) |>
      dplyr::pull(asset) |>
      unique() |>
      (\(x) id_hash("#asset:", paste0(project, x)))()

    has_part <- purrr::map(ds_ids, \(x) list(`@id` = x))

    rc <- rc |>
      rocrateR::add_entity(
        rocrateR::entity(
          id = id_hash("#project:", project),
          type = "Project",
          name = project,
          dateCreated = now,
          dateModified = now,
          hasPart = has_part
        )
      )
  }

  rc
}

#' Add Safe Dataset entities
#'
#' Creates Dataset entities and links them back to Projects
#' using `isPartOf`.
#'
#' @param rc RO-Crate object, see [rocrateR::rocrate].
#' @param tbl Tibble with dataset metadata.
#'   Columns: project, asset.
#'
#' @return Updated RO-Crate
#' @noRd
add_safe_data_entities_cr8tor <- function(rc, tbl) {
  for (i in seq_len(nrow(tbl))) {
    asset_id <- id_hash("#asset:", paste0(tbl$project[i], tbl$asset[i]))
    project_id <- id_hash("#project:", tbl$project[i])

    rc <- rc |>
      rocrateR::add_entity(
        rocrateR::entity(
          id = asset_id,
          type = "Dataset",
          name = tbl$asset[i],
          isPartOf = list(`@id` = project_id)
        )
      )
  }

  rc
}

#' Add Group entities
#'
#' Represents governance groups as organisations.
#'
#' @param rc RO-Crate object, see [rocrateR::rocrate].
#' @param groups_tbl Tibble with group metadata.
#'   Columns: group_id, description, project.
#'
#' @return Updated RO-Crate
#' @noRd
add_group_entities_cr8tor <- function(rc, groups_tbl) {
  for (i in seq_len(nrow(groups_tbl))) {
    g <- groups_tbl[i, ]
    gid <- id_hash("#group:", g$group_id)

    rc <- rc |>
      rocrateR::add_entity(
        rocrateR::entity(
          id = gid,
          type = "Organization",
          name = g$group_id,
          description = g$description,
          parentOrganization = list(`@id` = id_hash("#project:", g$project))
        )
      )
  }

  rc
}

#' Add Permission entities (Opal-level)
#'
#' Expands user-asset permission matrix into RO-Crate Action entities.
#' Uses `user_perm_entity()` to generate Read/Write/Control actions.
#'
#' @param rc RO-Crate object, see [rocrateR::rocrate].
#' @param perm_expanded_tbl Tibble with user permissions expanded.
#'   Columns: username, asset, permission.
#'
#' @return Updated RO-Crate
#' @noRd
add_permission_entities_cr8tor <- function(rc, perm_expanded_tbl) {
  for (i in seq_len(nrow(perm_expanded_tbl))) {
    row <- perm_expanded_tbl[i, ]

    person_id <- id_hash("#person:", row$username)
    asset_id <- id_hash("#asset:", paste0(row$project, row$asset))

    ents <- user_perm_entity(
      person = row$username,
      person_id = person_id,
      asset = row$asset,
      asset_id = asset_id,
      permission = row$permission
    )

    if (!is.null(ents)) {
      for (ent in ents) {
        rc <- rc |> rocrateR::add_entity(ent)
      }
    }
  }

  rc
}

add_safe_setting_entities_cr8tor <- function(rc, tbl) {
  for (i in seq_len(nrow(tbl))) {
    nm <- names(tbl)[i]

    rc <- rc |>
      rocrateR::add_entity(
        rocrateR::entity(
          id = id_hash("#setting:", nm),
          type = "PropertyValue",
          name = nm,
          value = as.character(tbl[[i]])
        )
      )
  }

  rc
}

add_safe_output_entities_cr8tor <- function(rc, tbl) {
  rc |>
    rocrateR::add_entity(
      rocrateR::entity(
        id = "#output:audit",
        type = "AssessAction",
        name = "cr8tor Compliance Assessment",
        actionStatus = if (tbl$validated) {
          "CompletedActionStatus"
        } else {
          "PotentialActionStatus"
        }
      )
    )
}

#' Convert audit result into RO-Crate
#'
#' @param audit list returned by audit.cr8tor()
#' @return rocrate object
#' @noRd
as_rocrate_audit <- function(audit) {
  rc <- rocrateR::rocrate_5s()

  # Root dataset describing the audit
  rc <- rc |>
    rocrateR::add_entity_value(
      "./",
      "name",
      "cr8tor 5 Safes Audit",
      overwrite = TRUE
    ) |>
    rocrateR::add_entity_value(
      "./",
      "description",
      "Audit report generated from cr8tor archive",
      overwrite = TRUE
    )

  # ---- Safe People ----
  rc <- rc |>
    add_group_entities_cr8tor(audit$groups) |>
    add_safe_people_entities_cr8tor(
      audit$safe_people$users,
      audit$user_projects
    ) |>
    link_people_to_root(audit$safe_people$users$username)

  # ---- Safe Projects ----
  rc <- rc |>
    add_safe_project_entities_cr8tor(
      audit$safe_projects,
      audit$safe_data$assets
    )

  # ---- Safe Data ----
  rc <- rc |>
    add_safe_data_entities_cr8tor(audit$safe_data$assets)

  # ---- Permissions ----
  rc <- rc |>
    add_permission_entities_cr8tor(
      expand_group_permissions_to_users(
        perm_tbl = audit$permissions |>
          dplyr::left_join(
            audit$safe_projects,
            by = c("project" = "project_id")
          ),
        membership_tbl = audit$user_groups,
        data_tbl = audit$safe_data$assets
      ) |>
        dedupe_effective_permissions()
    )

  # ---- Safe Settings ----
  rc <- rc |>
    add_safe_setting_entities_cr8tor(audit$safe_settings)

  # ---- Safe Outputs ----
  rc <- rc |>
    add_safe_output_entities_cr8tor(audit$safe_outputs)

  rc
}


#' Keep strongest permission per user-asset pair
#'
#' @param perm_tbl Tibble with username, asset, permission.
#' @return Deduplicated tibble
#' @noRd
dedupe_effective_permissions <- function(perm_tbl) {
  # local bindings
  asset <- permission <- strength <- username <- NULL
  strength_order <- c(
    "view",
    "view-values",
    "edit",
    "edit-values",
    "administrate"
  )

  perm_tbl |>
    dplyr::mutate(
      strength = match(permission, strength_order)
    ) |>
    dplyr::group_by(username, asset) |>
    dplyr::slice_max(strength, n = 1, with_ties = FALSE) |>
    dplyr::ungroup() |>
    dplyr::select(-strength)
}

#' Expand cr8tor group permissions to user-asset permissions
#'
#' Converts:
#'   Group → Project permissions
#' into:
#'   User → Table permissions
#'
#' @param perm_tbl Tibble from extract_permissions_cr8tor().
#' @param membership_tbl Tibble from extract_user_groups_cr8tor().
#' @param data_tbl Tibble from extract_safe_data_cr8tor()$tables.
#'
#' @return Tibble with username, project, asset, permission
#' @noRd
expand_group_permissions_to_users <- function(
  perm_tbl,
  membership_tbl,
  data_tbl
) {
  # local bindings
  asset <- name <- project <- role <- user <- username <- NULL

  # map cr8tor role to Opal permission
  role_to_permission <- function(role) {
    dplyr::case_when(
      grepl("admin", role, ignore.case = TRUE) ~ "administrate",
      grepl("analyst", role, ignore.case = TRUE) ~ "edit-values",
      grepl("editor", role, ignore.case = TRUE) ~ "edit",
      grepl("viewer", role, ignore.case = TRUE) ~ "view-values",
      TRUE ~ "view"
    )
  }

  # 1. expand group to users
  perm_users <- perm_tbl |>
    dplyr::left_join(
      membership_tbl,
      by = c("group" = "group")
    ) |>
    dplyr::mutate(
      username = dplyr::coalesce(user, username)
    ) |>
    dplyr::filter(!is.na(username))

  # 2. expand project to tables
  perm_tables <- perm_users |>
    dplyr::inner_join(
      data_tbl,
      by = "project"
    )

  # 3. Map roles to permissions
  perm_tables |>
    dplyr::mutate(project = name) |>
    dplyr::transmute(
      username,
      project,
      asset,
      permission = role_to_permission(role)
    ) |>
    dplyr::distinct()
}

#' Extract high-level cr8tor metadata
#'
#' @param bundle cr8tor_bundle
#' @return tibble with crate-level metadata
#' @noRd
extract_cr8tor_metadata <- function(bundle) {
  # root dataset (the crate itself)
  crate <- bundle$rocrate
  ds <- .get_entity(crate, id = "./")[[1]]

  # publisher may be nested
  publisher <- tryCatch(
    ds$publisher[["@id"]],
    error = function(e) NA_character_
  )

  tibble::tibble(
    crate_name = ds$name %||% NA_character_,
    description = ds$description %||% NA_character_,
    date_published = ds$datePublished %||% NA_character_,
    license = ds$license %||% NA_character_,
    publisher = publisher,
    main_project_id = ds$mainEntity[["@id"]] %||% NA_character_,
    rocrate_context = crate$`@context`,
    n_entities = length(crate$`@graph`)
  )
}

#' Extract integrity information
#'
#' BagIt validation handled during load_rocrate()
#'
#' @param bundle cr8tor_bundle
#' @noRd
extract_integrity_cr8tor <- function(bundle) {
  tibble::tibble(
    bagit_validated = TRUE,
    validation_source = "rocrateR::load_rocrate()"
  )
}

#' Extract Safe Data (datasets & tables)
#'
#' @param bundle cr8tor_bundle
#' @return list(tables tibble, n_tables)
#' @noRd
extract_safe_data_cr8tor <- function(bundle) {
  crate <- bundle$rocrate
  ing <- .get_entity(crate, id = "data/cr8-ingress.yaml")[[1]]

  # check if contents of the ingress entity are missing
  if (is.null(ing$content)) {
    return(list(tables = tibble::tibble(), n_tables = 0))
  }

  ing_yaml <- yaml::yaml.load(paste(ing$content[[1]], collapse = "\n"))
  # extract groups and projects
  group_tbl <- extract_groups_cr8tor(bundle)
  proj_tbl <- extract_safe_projects_cr8tor(bundle)

  assets <- ing_yaml$datasets |>
    purrr::map(function(ds) {
      # project name from locations
      yaml_project <- ds$locations[[1]]$opal_project_name %||% NA_character_
      # canonical project id
      project_id <- map_project_name_to_id(yaml_project, proj_tbl)

      purrr::map(ds$tables, function(tbl) {
        tibble::tibble(
          project = project_id,
          dataset = ds$name,
          asset = tbl$name,
          n_cols = length(tbl$columns)
        )
      }) |>
        purrr::list_c()
    }) |>
    purrr::list_c()

  tibble::tibble(
    assets = assets,
    n_assets = nrow(assets)
  )
}

#' Extract Safe Outputs
#'
#' @param bundle cr8tor_bundle
#' @return tibble
#' @noRd
extract_safe_outputs_cr8tor <- function(bundle) {
  acts <- .get_entity(
    bundle$rocrate,
    type = c("AssessAction", "CreateAction")
  )

  ids <- acts |>
    purrr::map(\(x) x[["@id"]]) |>
    purrr::list_c()

  tibble::tibble(
    n_actions = length(ids),
    validated = any(grepl("Validate", ids)),
    signoff = any(grepl("Sign-Off", ids)),
    disclosure = any(grepl("Disclosure", ids)),
    published = any(grepl("Publish", ids))
  )
}

#' Extract Safe People (users)
#'
#' @param bundle cr8tor_bundle
#' @return list(users tibble, n_users)
#' @noRd
extract_safe_people_cr8tor <- function(bundle) {
  # local bindings
  username <- NULL

  user_docs <- bundle$resources[
    grepl("user-.*\\.ya?ml$", names(bundle$resources))
  ]

  users <- purrr::map(user_docs, function(doc) {
    tibble::tibble(
      username = doc$metadata$name %||% NA_character_,
      given_name = doc$spec$given_name %||% NA_character_,
      family_name = doc$spec$family_name %||% NA_character_,
      email = doc$spec$email %||% NA_character_,
      affiliation = doc$spec$affiliation %||% NA_character_,
      enabled = doc$spec$enabled %||% FALSE
    )
  }) |>
    purrr::list_c() |>
    dplyr::distinct(username, .keep_all = TRUE)

  list(users = users, n_users = nrow(users))
}

extract_user_groups_cr8tor <- function(bundle) {
  user_docs <- bundle$resources[
    grepl("user-.*\\.ya?ml$", names(bundle$resources))
  ]

  purrr::map(user_docs, function(doc) {
    username <- doc$spec$username
    groups <- doc$spec$groups #%||% list()

    if (length(groups) == 0) {
      return(NULL)
    }

    purrr::map(groups, function(g) {
      tibble::tibble(
        username = username,
        group = g$value %||% NA_character_
      )
    }) |>
      purrr::list_c()
  }) |>
    purrr::list_c()
}

extract_user_projects_cr8tor <- function(bundle) {
  # local bindings
  description <- group_id <- NULL
  # extract group details
  grp_tbl <- extract_groups_cr8tor(bundle)
  # extract project details
  prj_tbl <- extract_projects_cr8tor(bundle)
  # extract user docs
  user_docs <- bundle$resources[
    grepl("user-.*\\.ya?ml$", names(bundle$resources))
  ]

  purrr::map(user_docs, function(doc) {
    username <- doc$spec$username
    groups <- doc$spec$groups # %||% list()

    if (length(groups) == 0) {
      return(NULL)
    }

    usr_grp_tbl <- purrr::map(groups, function(g) {
      tibble::tibble(
        username = username,
        group_id = g$value
      )
    }) |>
      purrr::list_c()

    # combine user x groups x projects
    usr_grp_tbl |>
      dplyr::left_join(grp_tbl, by = "group_id") |>
      dplyr::select(-description, -group_id) |>
      dplyr::left_join(prj_tbl, by = "project")
  }) |>
    purrr::list_c() |>
    dplyr::distinct()
}

extract_groups_cr8tor <- function(bundle) {
  grp_docs <- bundle$resources[
    grepl("group-.*\\.ya?ml$", names(bundle$resources))
  ]

  purrr::map(grp_docs, function(doc) {
    tibble::tibble(
      group_id = doc$metadata$name,
      description = doc$spec$description %||% NA_character_,
      project = doc$spec$projects %||% NA_character_
    )
  }) |>
    purrr::list_c()
}

#' Extract project permissions
#'
#' @param bundle cr8tor_bundle
#' @return tibble
#' @noRd
extract_permissions_cr8tor <- function(bundle) {
  grp_docs <- bundle$resources[
    grepl("group-.*\\.ya?ml$", names(bundle$resources))
  ]

  # role inference from group name
  role_from_group <- function(group_name) {
    dplyr::case_when(
      grepl("-admin$", group_name) ~ "admin",
      grepl("-analyst$", group_name) ~ "analyst",
      grepl("-editor$", group_name) ~ "editor",
      grepl("-viewer$", group_name) ~ "viewer",
      TRUE ~ "member"
    )
  }

  purrr::map(grp_docs, function(doc) {
    group_name <- doc$metadata$name %||% NA_character_
    role <- role_from_group(group_name)

    projects <- doc$spec$projects %||% NA_character_
    members <- doc$spec$members %||% NA_character_

    # normalise to vectors
    projects <- unlist(projects)
    members <- unlist(members)

    # if no explicit members, return group-level rule
    if (all(is.na(members))) {
      return(tibble::tibble(
        group = group_name,
        project = projects,
        user = NA_character_,
        role = role
      ))
    }

    # expand grid: user × project
    expand.grid(
      group = group_name,
      project = projects,
      user = members,
      role = role,
      stringsAsFactors = FALSE
    ) |>
      tibble::as_tibble()
  }) |>
    purrr::list_c()
}

extract_projects_cr8tor <- function(bundle) {
  prj_docs <- bundle$resources[
    grepl("project-.*\\.ya?ml$", names(bundle$resources))
  ]

  purrr::map(prj_docs, function(doc) {
    tibble::tibble(
      project = doc$metadata$name,
      description = doc$spec$description %||% NA_character_
    )
  }) |>
    purrr::list_c()
}

#' Extract Safe Projects
#'
#' @param bundle cr8tor_bundle
#' @return tibble
#' @noRd
extract_safe_projects_cr8tor <- function(bundle) {
  # RO-Crate identity
  proj <- .get_entity(bundle$rocrate, type = "Project")[[1]]

  # deployment spec
  proj_docs <- bundle$resources[
    grepl(
      sprintf("project-%s.*\\.ya?ml$", proj$identifier),
      names(bundle$resources)
    )
  ]

  tibble::tibble(
    project_id = proj$identifier,
    name = proj$identifier %||% NA_character_,
    description = proj$name %||% NA_character_,
    uuid = proj$`@id` %||% NA_character_,
    deployment_defined = length(proj_docs) > 0
  )
}

#' Extract Safe Settings
#'
#' @param bundle cr8tor_bundle
#' @return tibble
#' @noRd
extract_safe_settings_cr8tor <- function(bundle) {
  has_deployment <- any(grepl("deployment", names(bundle$resources)))

  tibble::tibble(
    deployment_defined = has_deployment,
    environment_type = "Containerised / Kubernetes",
    config_present = !is.null(bundle$config)
  )
}

#' Find BagIt root for an RO-Crate
#'
#' @param path String with path to RO-Crate bag.
#'
#' @returns String with path to RO-Crate bag root (if any).
#' @noRd
find_bagit_root <- function(path) {
  path <- normalizePath(path, mustWork = TRUE)

  # candidate directories: root + all subdirectories
  candidate_dirs <- c(
    path,
    list.dirs(path, recursive = TRUE, full.names = TRUE)
  )

  for (dir in candidate_dirs) {
    if (
      file.exists(file.path(dir, "bagit.txt")) &&
        dir.exists(file.path(dir, "data"))
    ) {
      return(dir)
    }
  }

  return(NULL)
}

link_people_to_root <- function(rc, usernames) {
  authors <- lapply(usernames, \(u) {
    list(`@id` = id_hash("#person:", u))
  })

  rocrateR::add_entity_value(
    rc,
    id = "./",
    key = "author",
    value = authors,
    overwrite = TRUE
  )
}

#' Load cr8tor governance bundle / project archive
#'
#' A cr8tor project archive contains:
#' * bagit/      → RO-Crate metadata layer.
#' * resources/  → deployment & governance YAML specs.
#' * config.toml → platform configuration.
#'
#' @param x Path to cr8tor ZIP archive.
#' @param ... Unused.
#'
#' @return Object of class `cr8tor`.
#' @keywords internal
load_cr8tor_bundle <- function(x, ...) {
  tmp <- tempfile("cr8tor_")
  dir.create(tmp, recursive = TRUE, showWarnings = FALSE)

  utils::unzip(x, exdir = tmp)

  # load RO-Crate layer
  rocrate <- rocrateR::load_rocrate(
    x, #file.path(tmp, "bagit"),
    load_content = TRUE,
    ...
  )

  tmp_root <- find_bagit_root(tmp) |>
    dirname()

  # parse resources layer
  res_dir <- file.path(tmp_root, "resources")

  yaml_files <- list.files(
    res_dir,
    pattern = "\\.ya?ml$",
    recursive = TRUE,
    full.names = TRUE
  )

  resources <- purrr::map(yaml_files, yaml::read_yaml)
  names(resources) <- sub(paste0("^", res_dir, "/?"), "", yaml_files)

  # parse config
  cfg_path <- file.path(tmp_root, "config.toml")
  config <- if (file.exists(cfg_path)) RcppTOML::parseTOML(cfg_path) else NULL

  structure(
    list(
      rocrate = rocrate,
      resources = resources,
      config = config,
      root = tmp_root
    ),
    class = c("cr8tor", "cr8tor_bundle", "list")
  )
}

map_project_name_to_id <- function(project, proj_tbl) {
  idx <- which(proj_tbl$description == project)
  if (length(idx) == 0) {
    idx <- which(proj_tbl$name == project)
  }
  if (length(idx)) proj_tbl$project_id[idx[1]] else project
}

#' Extract Safe Setting entity(ies)
#'
#' @inheritParams safe_data
#' @param ... Other optional arguments. See the full documentation
#'
#' @returns RO-Crate with Safe Setting entity(ies).
#' @rdname extract_safe_setting
#' @keywords internal
extract_safe_setting <- function(x, ...) {
  UseMethod("extract_safe_setting")
}

#' @param rocrate (Optional) RO-Crate object to update with Safe Setting details.
#' @rdname extract_safe_setting
#' @export
extract_safe_setting.opal <- function(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s()
) {
  # extract all the Safe Settings for the current Opal connection
  rocrate <- safe_setting(x, rocrate = rocrate)

  # return RO-Crate with Safe Setting details
  return(rocrate)
}

#' @param id (Optional) Vector with `@id` strings for Safe Setting entity(ies)
#'     to be extracted from the given RO-Crate, `x`.
#' @rdname extract_safe_setting
#' @export
extract_safe_setting.rocrate <- function(
  x,
  ...,
  id = NULL,
  rocrate = rocrateR::rocrate_5s()
) {
  # validate RO-Crate
  rocrateR::is_rocrate(x)

  # extract Safe Setting root entity, type = 'CreativeWork'
  ss_root <- .get_entity(x, type = "CreativeWork") |>
    purrr::keep(\(x) grepl("^#safesetting", x[["@id"]]))

  if (is.null(ss_root)) {
    stop("No Safe Setting root entity was found!", call. = FALSE)
  }

  # extract entities linked to the root
  part_ids <- ss_root |>
    purrr::map("hasPart") |>
    purrr::list_c() |>
    purrr::map(`[[`, "@id") |>
    # purrr::map(~ purrr::map(.x, "@id")) |>
    purrr::list_c()
  # ss_root_ids <- ss_root[[1]]$hasPart |>
  #   purrr::map(`[[`, "@id") |>
  #   purrr::list_c()

  # extract sub-entities linked to entities directly linked to the root entity
  if (length(part_ids) == 0) {
    stop("The Safe Setting root entity has no entities linked!", call. = FALSE)
  }

  entities_lst <- .get_entity(x, id = part_ids)
  sub_part_ids <- entities_lst |>
    purrr::map("hasPart") |>
    purrr::list_c() |>
    purrr::map(`[[`, "@id") |>
    purrr::list_c()
  if (length(sub_part_ids) > 0) {
    entities_lst <- c(entities_lst, .get_entity(x, id = sub_part_ids))
  }

  # if `id` was provided, then filter out only those entities
  if (!is.null(id)) {
    idx <- entities_lst |>
      sapply(\(x) getElement(x, "@id") %in% id)
    entities_lst <- entities_lst[idx]
  }

  # check if any entities were found
  if (length(entities_lst) == 0) {
    stop("No matching entities were found!", call. = FALSE)
  } else {
    message(
      length(entities_lst),
      " 'CreativeWork', 'PropertyValue' OR 'SoftwareApplication' entit",
      ifelse(length(entities_lst) == 1, "y was", "ies were"),
      " found!"
    )
  }

  # add entities to the RO-Crate and
  # ignore warnings about existing permission entities
  # and return RO-Crate with the Safe Setting details
  suppressWarnings({
    purrr::reduce(
      entities_lst,
      rocrateR::add_entity,
      overwrite = TRUE,
      .init = rocrate
    )
  })
}

#' Flatten object with Safe Setting details
#'
#' @param x Object (e.g., RO-Crate) with Safe Setting details. This can be
#'     generated with the [extract_safe_setting()] function.
#' @param id Vector of strings with the `@id`s for the settings to be extracted.
#'     If not provided, extract all entities with `@type = 'PropertyValue'` or
#'     `@type = 'SoftwareApplication'`.
#'
#' @returns Data frame with Safe Settings.
#' @rdname flatten_safe_setting
#' @keywords internal
flatten_safe_setting <- function(x, ...) {
  UseMethod("flatten_safe_setting")
}

#' @rdname flatten_safe_setting
#' @export
flatten_safe_setting.default <- function(x, ...) {
  return(tibble::tibble())
}

#' @rdname flatten_safe_setting
#' @export
flatten_safe_setting.rocrate <- function(x, ..., id = NULL) {
  tryCatch(
    {
      # extract entities
      entities_tbl <- x |>
        .get_entity(type = c("PropertyValue", "SoftwareApplication")) |>
        # extract @id, name, value and version for each entity
        lapply(function(ent) {
          tibble::tibble(
            id = getElement(ent, "@id"),
            name = getElement(ent, "name"),
            value = getElement(ent, "value"),
            version = getElement(ent, "version")
          )
        }) |>
        # combine all rows
        dplyr::bind_rows()

      # if `id` is provided, then only keep those entities
      if (!is.null(id)) {
        entities_tbl <- entities_tbl |>
          dplyr::filter(id %in% !!id)
      }

      # return dataset entities
      return(entities_tbl)
    },
    error = function(e) {
      tibble::tibble()
    }
  )
}

#' Link Safe Setting entity to all Safe Projects
#'
#' Creates explicit graph relationships so Safe Settings are
#' discoverable from Project-level governance reports.
#'
#' Relationship model:
#'   Safe Setting <root> --> isPartOf --> Project
#'
#' @param rocrate RO-Crate object, see [rocrateR::rocrate].
#' @return Updated rocrate.
#' @keywords internal
#' @noRd
link_safe_settings_to_projects <- function(rocrate) {
  ents <- rocrate$`@graph`

  setting_ids <- purrr::map(
    purrr::keep(ents, ~ grepl("^#safesetting", .x$`@id` %||% "")),
    "@id"
  ) |>
    purrr::list_c()

  project_ids <- purrr::map(
    purrr::keep(ents, ~ identical(.x$`@type`, "Project")),
    "@id"
  ) |>
    purrr::list_c()

  if (length(setting_ids) == 0 || length(project_ids) == 0) {
    return(rocrate)
  }

  link_entities <- expand.grid(setting = setting_ids, proj = project_ids) |>
    purrr::pmap(
      function(setting, proj) {
        rocrateR::entity(
          id = id_hash("#link:", paste0(setting, proj)),
          type = "CreativeWork",
          name = "Safe Settings x Safe Project Link",
          about = list("@id" = setting),
          isPartOf = list("@id" = proj)
        )
      }
    )

  purrr::reduce(link_entities, rocrateR::add_entity, .init = rocrate)
}

# -------------------------------------------------------------------------
# S3 generic dispatch
# -------------------------------------------------------------------------
test_that("extract_safe_setting dispatch works for rocrate", {
  # create an RO-Crate with setting entities
  disc_setting <- rocrateR::entity(
    id = "#setting-1",
    type = "PropertyValue",
    name = "analysis.threshold",
    value = "0.05"
  )
  disc_env <- rocrateR::entity(
    id = paste0("#env:disclosure_settings:", "default"),
    type = "CreativeWork",
    name = "Disclosure Control Environment",
    hasPart = purrr::map(list(disc_setting), ~ list("@id" = .x$`@id`))
  )
  safe_setting_root <- rocrateR::entity(
    id = paste0("#safesetting:", "opal"),
    type = "CreativeWork",
    name = "Safe Setting Controls (Opal)",
    description = paste(
      "Technical, physical and organisational safeguards applied to minimise",
      "disclosure risk."
    ),
    hasPart = list(disc_env) |>
      purrr::map(\(x) list("@id" = x$`@id`))
  )
  rocrate <- list(disc_setting, disc_env, safe_setting_root) |>
    purrr::reduce(rocrateR::add_entity, .init = rocrateR::rocrate_5s())

  out <- dsROCrate:::extract_safe_setting(rocrate)

  expect_s3_class(out, "rocrate")
})

test_that("flatten_safe_setting.default returns empty tibble", {
  out <- flatten_safe_setting(list())

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0)
})

# -------------------------------------------------------------------------
# extract_safe_setting.rocrate()
# -------------------------------------------------------------------------
test_that("extract_safe_setting.rocrate errors when no matching entities", {
  rocrate <- rocrateR::rocrate_5s()

  # ignore warning about missing entity of @type = 'PropertyValue'!
  suppressWarnings(
    expect_error(
      extract_safe_setting(rocrate),
      "The Safe Setting root entity has no entities linked!"
    )
  )
})

test_that("extract_safe_setting.rocrate extracts PropertyValue and SoftwareApplication", {
  # create an RO-Crate with setting entities
  software_ent <- rocrateR::entity(
    id = "#software-1",
    type = "SoftwareApplication",
    name = "DataSHIELD",
    version = "6.2.0"
  )
  disc_setting <- rocrateR::entity(
    id = "#setting-1",
    type = "PropertyValue",
    name = "analysis.threshold",
    value = "0.05"
  )
  disc_env <- rocrateR::entity(
    id = paste0("#env:disclosure_settings:", "default"),
    type = "CreativeWork",
    name = "Disclosure Control Environment",
    hasPart = purrr::map(list(disc_setting), ~ list("@id" = .x$`@id`))
  )
  safe_setting_root <- rocrateR::entity(
    id = paste0("#safesetting:", "opal"),
    type = "CreativeWork",
    name = "Safe Setting Controls (Opal)",
    description = paste(
      "Technical, physical and organisational safeguards applied to minimise",
      "disclosure risk."
    ),
    hasPart = list(disc_env, software_ent) |>
      purrr::map(\(x) list("@id" = x$`@id`))
  )
  rocrate <- list(disc_setting, disc_env, software_ent, safe_setting_root) |>
    purrr::reduce(rocrateR::add_entity, .init = rocrateR::rocrate_5s())

  out <- extract_safe_setting(rocrate)

  expect_s3_class(out, "rocrate")

  pv <- .get_entity(out, type = "PropertyValue")
  sa <- .get_entity(out, type = "SoftwareApplication")

  expect_equal(length(pv), 1)
  expect_equal(length(sa), 1)
})

test_that("extract_safe_setting.rocrate filters correctly by id", {
  # create an RO-Crate with setting entities
  disc_setting <- rocrateR::entity(
    id = "#setting-1",
    type = "PropertyValue",
    name = "analysis.threshold",
    value = "0.05"
  )
  disc_env <- rocrateR::entity(
    id = paste0("#env:disclosure_settings:", "default"),
    type = "CreativeWork",
    name = "Disclosure Control Environment",
    hasPart = purrr::map(list(disc_setting), ~ list("@id" = .x$`@id`))
  )
  safe_setting_root <- rocrateR::entity(
    id = paste0("#safesetting:", "opal"),
    type = "CreativeWork",
    name = "Safe Setting Controls (Opal)",
    description = paste(
      "Technical, physical and organisational safeguards applied to minimise",
      "disclosure risk."
    ),
    hasPart = list(disc_env) |>
      purrr::map(\(x) list("@id" = x$`@id`))
  )
  rocrate <- list(disc_setting, disc_env, safe_setting_root) |>
    purrr::reduce(rocrateR::add_entity, .init = rocrateR::rocrate_5s())

  out <- extract_safe_setting(rocrate, id = "#setting-1")

  pv <- .get_entity(out, type = "PropertyValue")
  # ignore warning about missing entity of @type = 'SoftwareApplication'!
  suppressWarnings(
    sa <- .get_entity(out, type = "SoftwareApplication")
  )

  expect_equal(length(pv), 1)
  expect_equal(getElement(pv[[1]], "@id"), "#setting-1")
  # SoftwareApplication should not be re-added when filtered out
  expect_true(length(sa) >= 0)
})

test_that("extract_safe_setting.rocrate emits informative message", {
  # create an RO-Crate with setting entities
  disc_setting <- rocrateR::entity(
    id = "#setting-1",
    type = "PropertyValue",
    name = "analysis.threshold",
    value = "0.05"
  )
  disc_env <- rocrateR::entity(
    id = paste0("#env:disclosure_settings:", "default"),
    type = "CreativeWork",
    name = "Disclosure Control Environment",
    hasPart = purrr::map(list(disc_setting), ~ list("@id" = .x$`@id`))
  )
  safe_setting_root <- rocrateR::entity(
    id = paste0("#safesetting:", "opal"),
    type = "CreativeWork",
    name = "Safe Setting Controls (Opal)",
    description = paste(
      "Technical, physical and organisational safeguards applied to minimise",
      "disclosure risk."
    ),
    hasPart = list(disc_env) |>
      purrr::map(\(x) list("@id" = x$`@id`))
  )
  rocrate <- list(disc_setting, disc_env, safe_setting_root) |>
    purrr::reduce(rocrateR::add_entity, .init = rocrateR::rocrate_5s())

  expect_message(
    extract_safe_setting(rocrate),
    "'PropertyValue' OR 'SoftwareApplication' entit"
  )
})

test_that("extract_safe_setting.rocrate errors when id filter removes all", {
  # create an RO-Crate with setting entities
  disc_setting <- rocrateR::entity(
    id = "#setting-1",
    type = "PropertyValue",
    name = "analysis.threshold",
    value = "0.05"
  )
  disc_env <- rocrateR::entity(
    id = paste0("#env:disclosure_settings:", "default"),
    type = "CreativeWork",
    name = "Disclosure Control Environment",
    hasPart = purrr::map(list(disc_setting), ~ list("@id" = .x$`@id`))
  )
  safe_setting_root <- rocrateR::entity(
    id = paste0("#safesetting:", "opal"),
    type = "CreativeWork",
    name = "Safe Setting Controls (Opal)",
    description = paste(
      "Technical, physical and organisational safeguards applied to minimise",
      "disclosure risk."
    ),
    hasPart = list(disc_env) |>
      purrr::map(\(x) list("@id" = x$`@id`))
  )
  rocrate <- list(disc_setting, disc_env, safe_setting_root) |>
    purrr::reduce(rocrateR::add_entity, .init = rocrateR::rocrate_5s())

  expect_error(
    extract_safe_setting(rocrate, id = "#nonexistent"),
    "No matching entities were found!"
  )
})

test_that("extract_safe_setting.rocrate validates rocrate input", {
  bad_rc <- structure(list(), class = "rocrate")

  expect_error(
    extract_safe_setting.rocrate(bad_rc)
  )
})

# -------------------------------------------------------------------------
# flatten_safe_setting.rocrate()
# -------------------------------------------------------------------------
test_that("flatten_safe_setting.rocrate returns tibble with correct columns", {
  # create an RO-Crate with setting entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#setting-1",
        type = "PropertyValue",
        name = "analysis.threshold",
        value = "0.05"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#software-1",
        type = "SoftwareApplication",
        name = "DataSHIELD",
        version = "6.2.0"
      )
    )

  out <- flatten_safe_setting(rocrate)

  expect_s3_class(out, "tbl_df")
  expect_true(all(c("id", "name", "value", "version") %in% names(out)))
})

test_that("flatten_safe_setting.rocrate extracts PropertyValue fields", {
  # create an RO-Crate with setting entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#setting-1",
        type = "PropertyValue",
        name = "analysis.threshold",
        value = "0.05"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#software-1",
        type = "SoftwareApplication",
        name = "DataSHIELD",
        version = "6.2.0"
      )
    )

  out <- flatten_safe_setting(rocrate)

  expect_true("#setting-1" %in% out$id)
  expect_true("analysis.threshold" %in% out$name)
  expect_true("0.05" %in% out$value)
})

test_that("flatten_safe_setting.rocrate extracts SoftwareApplication version", {
  # create an RO-Crate with setting entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#setting-1",
        type = "PropertyValue",
        name = "analysis.threshold",
        value = "0.05"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#software-1",
        type = "SoftwareApplication",
        name = "DataSHIELD",
        version = "6.2.0"
      )
    )

  out <- flatten_safe_setting(rocrate)

  software_row <- out[out$id == "#software-1", ]

  expect_equal(software_row$name, "DataSHIELD")
  expect_equal(software_row$version, "6.2.0")
})

test_that("flatten_safe_setting.rocrate filters by id correctly", {
  # create an RO-Crate with setting entities
  rocrate <- rocrateR::rocrate_5s() |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#setting-1",
        type = "PropertyValue",
        name = "analysis.threshold",
        value = "0.05"
      )
    ) |>
    rocrateR::add_entity(
      entity = rocrateR::entity(
        id = "#software-1",
        type = "SoftwareApplication",
        name = "DataSHIELD",
        version = "6.2.0"
      )
    )

  out <- flatten_safe_setting(rocrate, id = "#software-1")

  expect_equal(nrow(out), 1)
  expect_equal(out$id, "#software-1")
  expect_equal(out$name, "DataSHIELD")
})

test_that("flatten_safe_setting.rocrate returns empty tibble on error", {
  # Force error inside tryCatch by giving malformed object
  bad_obj <- structure(list(), class = "rocrate")

  out <- flatten_safe_setting(bad_obj)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0)
})

# -------------------------------------------------------------------------
# extract_safe_setting.opal() (unit + integration strategy)
# -------------------------------------------------------------------------
test_that("extract_safe_setting.opal calls safe_setting and returns rocrate", {
  # dummy opal-like object (S3 dispatch target)
  opal_con <- structure(list(), class = "opal")

  # mock safe_setting to avoid real server dependency
  mock_safe_setting <- function(x, rocrate) {
    setting <- rocrateR::entity(
      id = "#mock-setting",
      type = "PropertyValue",
      name = "mock.option",
      value = "TRUE"
    )
    rocrateR::add_entity(rocrate, setting)
  }

  with_mocked_bindings(
    safe_setting = mock_safe_setting,
    {
      out <- extract_safe_setting.opal(opal_con)
    }
  )

  expect_s3_class(out, "rocrate")

  pv <- .get_entity(out, type = "PropertyValue")
  expect_true(length(pv) >= 1)
})

# -------------------------------------------------------------------------
# optional TRUE Opal demo integration (matches your demo server workflow)
# -------------------------------------------------------------------------
test_that("extract_safe_setting.opal works with Opal demo server (if available)", {
  con <- try(opal_demo_con(), silent = TRUE)
  skip_if(inherits(con, "try-error"), "Opal demo server not available")

  rocrate <- extract_safe_setting(con)

  expect_s3_class(rocrate, "rocrate")
})

#' Initialise a Five Safes RO-Crate
#'
#' Creates a new RO-Crate configured for Five Safes auditing.
#'
#' @param x This can be a connection to a 'DataSHIELD' server (e.g., object with
#'     the `opal` class, see [opalr::opal.login()]), an RO-Crate
#'     ([rocrate][rocrateR::rocrate()] class) or a string with the path to an
#'     RO-Crate.
#' @param ... Other optional arguments. See the full documentation,
#'     [`?dsROCrate::init`][init()].
#' @param connection Connection object for the 'DataSHIELD' server where the
#'     values will be extracted from (e.g., OBiBa's Opal). Optional, if `x` is
#'     set to a connection object. If so, then `rocrate` is required.
#' @param rocrate RO-Crate object. Optional, if `x` is either an RO-Crate
#'     object or a path to a valid RO-Crate. If so, then `connection` is
#'     required (default: `rocrateR::rocrate_5s()`).
#' @param path String with path pointing to the root of the RO-Crate. This will
#'     be used to store log files. If not provided, logs will be stored within
#'     the RO-Crate returned by this function.
#' @param profile String with profile name (used for OBiBa's Opal backend).
#' @param project String with the name of the [Safe Project][safe_project()].
#' @param resources Vector of strings with the names of the resources, part of
#'     `project`. Optional, if not provided, all the resources associated to
#'     `project` will be included in the RO-Crate.
#' @param tables Vector of strings with the names of the tables/datasets, part
#'     of `project`. Optional, if not provided, all the tables/datasets
#'     associated to `project` will be included in the RO-Crate.
#' @param user List (or [entity][rocrateR::entity()] object) with details for
#'     the Safe People, it must include `@id` and `name` entries. Alternatively,
#'     this can be a string with the `name` of the current user.
#'
#' @returns Five Safes RO-Crate object.
#'
#' @references
#' Wilkinson, M., Dumontier, M., Aalbersberg, I. et al. (2016) The FAIR Guiding
#' Principles for scientific data management and stewardship. Sci Data 3,
#' 160018. https://doi.org/10.1038/sdata.2016.18
#'
#' @export
init <- function(x, ...) {
  UseMethod("init")
}

#' @rdname init
#' @export
init.opal <- function(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL,
  path = NULL,
  user = NULL
) {
  # x is a valid opal connection object
  validate_opal_con(x)

  # attach input arguments as attributes
  attr(rocrate, "connection") <- x
  attr(rocrate, "path") <- path
  attr(rocrate, "profile") <- profile
  attr(rocrate, "project") <- project
  attr(rocrate, "resources") <- resources
  attr(rocrate, "tables") <- tables
  attr(rocrate, "user") <- user

  return(rocrate)
}

#' @rdname init
#' @export
init.rocrate <- function(
  x,
  ...,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  user = attr(x, "user")
) {
  init(
    connection,
    rocrate = x,
    path = path,
    profile = profile,
    project = project,
    resources = resources,
    tables = tables,
    user = user
  )
}

# # helper functions ----
# `%||%` <- function(a, b) if (!is.null(a) && !is.na(a) && a != "") a else b

test_that("init() dispatches to init.opal() with real Opal connection", {
  # setup
  crate <- rocrateR::rocrate_5s()
  opal_con <- opal_demo_con()

  res <- init(
    opal_con,
    rocrate = crate,
    project = attr(opal_con, "PROJECT"),
    tables = attr(opal_con, "TABLES"),
    user = attr(opal_con, "PEOPLE")
  )

  expect_s3_class(res, "rocrate")
  expect_identical(attr(res, "connection"), opal_con)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("init.opal() attaches metadata using demo Opal server", {
  # setup
  crate <- rocrateR::rocrate_5s()
  opal_con <- opal_demo_con()

  res <- init.opal(
    opal_con,
    rocrate = crate,
    path = tempdir(),
    project = attr(opal_con, "PROJECT"),
    tables = attr(opal_con, "TABLES"),
    user = attr(opal_con, "PEOPLE")
  )

  expect_identical(attr(res, "connection"), opal_con)
  expect_identical(attr(res, "project"), attr(opal_con, "PROJECT"))
  expect_identical(attr(res, "tables"), attr(opal_con, "TABLES"))
  expect_identical(attr(res, "user"), attr(opal_con, "PEOPLE"))
  expect_true(dir.exists(attr(res, "path")))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("init.rocrate() reuses stored attributes with demo Opal server", {
  # setup
  crate <- rocrateR::rocrate_5s()
  opal_con <- opal_demo_con()

  attr(crate, "connection") <- opal_con
  attr(crate, "project") <- attr(opal_con, "PROJECT")
  attr(crate, "tables") <- attr(opal_con, "TABLES")
  attr(crate, "user") <- attr(opal_con, "PEOPLE")
  attr(crate, "path") <- tempdir()

  res <- init(crate)

  expect_identical(attr(res, "connection"), opal_con)
  expect_identical(attr(res, "project"), attr(opal_con, "PROJECT"))
  expect_identical(attr(res, "tables"), attr(opal_con, "TABLES"))
  expect_identical(attr(res, "user"), attr(opal_con, "PEOPLE"))
  expect_identical(attr(res, "path"), attr(crate, "path"))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("init.opal() errors on invalid connection object", {
  # setup
  crate <- rocrateR::rocrate_5s()

  expect_error(
    init.opal(list(not = "opal"), rocrate = crate),
    class = "error"
  )
})

#' Safe Data details
#'
#' Safe Data details for the RO-Crate.
#'
#' Researchers only use de-identified data that is relevant to their study.
#'
#' In compliance with the Digital Economy Act, data is effectively anonymised
#' within TREs (Trusted Research Environments).
#'
#' This means any sensitive information that might lead to an individual being
#' identified, such as names and addresses, is either removed or replaced with
#' a random code. Researchers are not processing personal data when using data
#' prepared in this way and when the other Safes are in place. Find out more
#' about de-identification:
#' <https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-data-de-identification/>
#'
#' @param x This can be a connection to a 'DataSHIELD' server (e.g., object with
#'     the `opal` class, see [opalr::opal.login()]), an RO-Crate
#'     ([rocrate][rocrateR::rocrate()] class) or a string with the path to an
#'     RO-Crate.
#' @param ... Other optional arguments. See the full documentation,
#'     [`?dsROCrate::safe_data`][safe_data()].
#' @param include Vector of strings with types of assets to be included, either
#'     `"resources"`, `"tables"` or both.
#' @param asset_id_suffix String with ID suffix for the tables/datasets
#'     entities in the RO-Crate (default: `"#asset:"`).
#' @param project_id_suffix String with ID suffix for the project entities
#'     in the RO-Crate (default: `"#project:"`).
#' @inheritParams init
#'
#' @returns Updated RO-Crate object with Safe Data information.
#' @export
#'
#' @source
#' \itemize{
#'  \item Research Data Scotland, 2025. "What is the Five Safes framework?".
#'  <https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/>
#' }
safe_data <- function(x, ...) {
  UseMethod("safe_data")
}

#' @export
safe_data.default <- function(x, ...) {
  stop(
    "Unknown class, please try either a file path or",
    " an object with `rocrate` class!"
  )
}

#' @rdname safe_data
#' @export
safe_data.character <- function(
  x,
  ...,
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  user = attr(x, "user")
) {
  # attempt loading the RO-Crate
  rocrate <- rocrateR::load_rocrate(x)

  # call method with given `rocrate` object:
  safe_data(
    rocrate,
    connection = connection,
    profile = profile,
    project = project,
    resources = resources,
    tables = tables,
    asset_id_suffix = asset_id_suffix,
    project_id_suffix = project_id_suffix,
    path = path,
    user = user
  )
}

#' @rdname safe_data
#' @export
safe_data.opal <- function(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL,
  include = c("tables", "resources"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  path = NULL,
  user = NULL
) {
  # declare local bindings
  created <- lastUpdate <- name <- new_dataset_entity <- subject <- NULL

  # x is a valid opal connection object
  validate_opal_con(x)

  # match the assets to be included
  include <- match.arg(include, several.ok = TRUE)

  # enforce that `project` is a single value
  if (is.null(project)) {
    stop("A value for `project` is required!", call. = FALSE)
  } else if (length(project) != 1) {
    stop("`project` must be a single value!", call. = FALSE)
  }

  project_id <- id_hash(project_id_suffix, project)

  # extract assets for the given project
  all_assets <- list()

  # ---- tables ----
  if ("tables" %in% include) {
    tbl_assets <- get_project_assets(x, project, "tables")
    if (!is.null(tbl_assets)) {
      if (!is.null(tables)) {
        tbl_assets <- dplyr::filter(tbl_assets, name %in% tables)
      }
      all_assets$tables <- tbl_assets
    }
  }

  # ---- resources ----
  if ("resources" %in% include) {
    res_assets <- get_project_assets(x, project, "resources")
    if (!is.null(res_assets)) {
      if (!is.null(resources)) {
        res_assets <- dplyr::filter(res_assets, name %in% resources)
      }
      all_assets$resources <- res_assets
    }
  }

  # combine assets
  assets_tbl <- dplyr::bind_rows(all_assets)

  # check that project assets were found
  err_msg <- sprintf("No details were found for `project = '%s'!", project)
  if (is.null(assets_tbl) || nrow(assets_tbl) == 0) {
    stop(err_msg, call. = FALSE)
  }

  # build asset entities ----
  asset_entities <- build_asset_entities(
    assets_tbl,
    project_id = project_id,
    asset_id_suffix = asset_id_suffix
  )

  # link asset entities with project entity
  rocrate <- link_assets_to_project(
    rocrate,
    project_id,
    vapply(asset_entities, `[[`, "", "@id")
  )

  # build ID lookup ----
  id_lookup <- stats::setNames(
    vapply(asset_entities, `[[`, "", "@id"),
    assets_tbl$name
  )

  # lineage ----
  lineage_df <- infer_table_resource_lineage(assets_tbl)

  rocrate <- add_lineage_relations(
    rocrate,
    lineage_df,
    id_lookup
  )

  # permissions ----
  rocrate <- add_asset_permissions_to_crate(
    rocrate,
    x,
    project,
    assets_tbl,
    id_lookup
  )

  # add asset entities to the RO-Crate
  rocrate <- purrr::reduce(
    asset_entities,
    rocrateR::add_entity,
    .init = rocrate,
    overwrite = TRUE
  )

  # attach input arguments as attributes
  attr(rocrate, "connection") <- x
  attr(rocrate, "path") <- path
  attr(rocrate, "profile") <- profile
  attr(rocrate, "project") <- project
  attr(rocrate, "resources") <- resources
  attr(rocrate, "tables") <- tables
  attr(rocrate, "user") <- user

  # return RO-Crate with the new entity
  return(rocrate)
}

#' @rdname safe_data
#' @export
safe_data.rocrate <- function(
  x,
  ...,
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  user = attr(x, "user")
) {
  # check if the connection was given
  if (is.null(connection)) {
    stop("A `connection` object is required!", call. = FALSE)
  }

  # call method with given `connection` object:
  safe_data(
    connection,
    rocrate = x,
    profile = profile,
    project = project,
    resources = resources,
    tables = tables,
    asset_id_suffix = asset_id_suffix,
    project_id_suffix = project_id_suffix,
    path = path,
    user = user
  )
}

#' Add Asset Permissions to RO-Crate
#'
#' Creates a Permission entity describing dataset-level access for a person
#' within a project and links it to the relevant assets.
#'
#' @param rocrate RO-Crate object, see [rocrateR::rocrate].
#' @param project Project identifier.
#' @param person User identifier.
#' @param assets Vector of strings with asset identifiers.
#'
#' @return Updated RO-Crate.
#'
#' @noRd
add_asset_permissions_to_crate <- function(
  rocrate,
  x,
  project,
  assets_tbl,
  id_lookup
) {
  for (i in seq_len(nrow(assets_tbl))) {
    asset <- assets_tbl[i, ]

    asset_id <- id_lookup[[asset$name]]
    asset_type <- asset$asset_type

    # retrieve asset permissions
    perms <- get_asset_permissions(x, project, asset_type, asset$name)

    if (is.null(perms)) {
      next
    }

    # iterate through person permissions for an asset
    for (j in seq_len(nrow(perms))) {
      person <- perms$person[j]
      permission <- perms$permission[j]

      person_id <- id_hash("#person:", person)

      # create permission entities
      perm_ents <- user_asset_perm_entities(
        person = person,
        person_id = person_id,
        asset = asset$name,
        asset_id = asset_id,
        permission = permission,
        asset_type = asset_type
      )

      # add permission entities
      rocrate <- purrr::reduce(
        perm_ents,
        rocrateR::add_entity,
        overwrite = TRUE,
        .init = rocrate
      )
    }
  }

  rocrate
}

#' @noRd
add_lineage_relations <- function(rocrate, lineage_df, id_lookup) {
  if (is.null(lineage_df) || nrow(lineage_df) == 0) {
    return(rocrate)
  }

  for (i in seq_len(nrow(lineage_df))) {
    tbl_id <- id_lookup[[lineage_df$table_name[i]]]
    res_id <- id_lookup[[lineage_df$resource_name[i]]]

    tbl_ent <- rocrate$graph[[tbl_id]]

    tbl_ent$isBasedOn <- list(`@id` = res_id)

    rocrate$graph[[tbl_id]] <- tbl_ent
  }

  rocrate
}

#' Build entities for project assets
#'
#' @param assets_tbl Tibble with project assets (see `get_project_assets()`).
#' @param project_id String with project `@id`.
#' @param asset_id_suffix String with asset `@id` suffix.
#'
#' @returns List with entities for the project assets.
#' @noRd
build_asset_entities <- function(assets_tbl, project_id, asset_id_suffix) {
  purrr::pmap(
    assets_tbl,
    function(
      asset_type,
      project,
      name,
      description,
      created,
      updated,
      url,
      meta
    ) {
      rocrateR::entity(
        id = id_hash(asset_id_suffix, paste0(project, name)),
        type = map_asset_type(asset_type, meta, url),
        name = name,
        description = safe_desc(description, name, asset_type),
        url = url,
        dateCreated = safe_time(created),
        dateModified = safe_time(updated),
        isPartOf = list(`@id` = project_id) #,
        # extra = list(
        #   assetKind = asset_type,
        #   backend = meta
        # )
      )
    }
  )
}

#' Get project's asset permissions
#'
#' @param x Connection to OBiBa's Opal server (see [opalr::opal.login()]).
#' @param project String with project name.
#' @param asset_type String with type of asset, either `tables` or `resources`.
#' @param name String with asset name.
#'
#' @returns Tibble with project asset permissions
#' @noRd
get_asset_permissions <- function(x, project, asset_type, name) {
  if (asset_type == "table") {
    perms <- opalr::opal.table_perm(x, project, name)
  } else {
    perms <- opalr::opal.resource_perm(x, project, name)
  }

  if (is.null(perms) || length(perms$subject) == 0) {
    return(NULL)
  }

  tibble::tibble(
    person = perms$subject,
    permission = perms$permission
  )
}

#' Get project assets.
#'
#' @param x Connection to OBiBa's Opal server (see [opalr::opal.login()]).
#' @param project String with project name.
#' @param type Type of assets to extract, either `tables` or `resources`.
#'
#' @returns Tibble with project assets details.
#' @noRd
get_project_assets <- function(x, project, type = c("tables", "resources")) {
  type <- match.arg(type)

  # verify if project exists
  project_exists(x, project = project)

  if (type == "tables") {
    prj <- opalr::opal.tables(x, project, df = FALSE)

    if (length(prj) == 0) {
      return(NULL)
    }

    tibble::tibble(
      asset_type = "table",
      project = project,
      name = vapply(prj, \(x) x$name %||% "", ""),
      description = vapply(prj, \(x) x$description %||% "", ""),
      created = vapply(prj, \(x) safe_time(x$timestamps$created), character(1)),
      #   prj,
      #   \(x) as.POSIXct(x$timestamps$created, tz = "UTC") %||% NA_real_,
      #   character(1)
      # ),
      updated = vapply(
        prj,
        \(x) safe_time(x$timestamps$lastUpdate),
        character(1)
      ),
      #   prj,
      #   \(x) as.POSIXct(x$timestamps$lastUpdate, tz = "UTC") %||% NA_real_,
      #   character(1)
      # ),
      url = vapply(prj, \(x) x$link %||% NA_character_, ""),
      meta = vector("list", length(prj))
    )
  } else if (type == "resources") {
    res <- opalr::opal.resources(x, project, df = FALSE)

    if (length(res) == 0) {
      return(NULL)
    }

    tibble::tibble(
      asset_type = "resource",
      project = project,
      name = vapply(res, \(x) x$name %||% "", ""),
      description = vapply(res, \(x) x$description %||% "", ""),
      created = vapply(res, \(x) safe_time(x$created), character(1)),
      #   res,
      #   \(x) as.POSIXct(x$created, tz = "UTC") %||% NA_real_,
      #   character(1)
      # ),
      updated = vapply(res, \(x) safe_time(x$updated), character(1)),
      #   res,
      #   \(x) as.POSIXct(x$updated, tz = "UTC") %||% NA_real_,
      #   character(1)
      # ),
      url = vapply(res, function(x) x$resource$url %||% NA_character_, ""),
      meta = parse_resource_params(
        vapply(res, `[[`, "", "parameters")
      )
    )
  }
}

#' Get project details (including tables)
#'
#' @inheritParams get_table_permissions
#'
#' @returns Data frame with project and tables associated
#' @keywords internal
#'
#' @family Opal
get_project_details <- function(x, project) {
  project |>
    lapply(function(p) {
      tryCatch(
        {
          # check if the given project exists
          project_exists(x, project = p)

          # retrieve tables for the given project
          project_tables <- get_project_tables(x, p)
          # check if any tables were found attached to the project
          if (length(project_tables) < 1) {
            return(tibble::tibble(project = p, table = NA))
          }
          tibble::tibble(
            project = p,
            table = project_tables
          )
        },
        error = function(e) {
          return(tibble::tibble(project = p, table = NA))
        }
      )
    }) |>
    dplyr::bind_rows() |>
    dplyr::filter(!is.na(table))
}

#' Get project tables
#'
#' Wrapper for [opalr::opal.project()].
#'
#' @inheritParams get_table_permissions
#'
#' @returns List of project tables
#' @keywords internal
#'
#' @family Opal
get_project_tables <- function(x, project) {
  # verify if project exists
  project_exists(x, project = project)

  # extract table names associated to `project`
  project_tables <- opalr::opal.project(x, project) |>
    getElement("datasource") |>
    getElement("table") |>
    unlist()

  # verify if `project_tables` is missing or NULL, if so, print warning message
  if (all(is.na(project_tables)) || all(is.null(project_tables))) {
    warning(
      "The given `project`, does not have any tables associated!",
      call. = FALSE
    )

    # return empty list, invisibly
    return(invisible(list()))
  }

  # return project tables
  return(project_tables)
}

#' Get table permissions
#'
#' Wrapper for the [opalr::opal.table_perm()] function.
#'
#' @inheritParams validate_opal_con
#' @param project String with project name.
#' @param tables String (or vector of strings) with table names for the given
#'     project.
#'
#' @returns Data frame with permissions for each table in `tables`.
#' @keywords internal
#'
#' @family Opal
get_table_permissions <- function(x, project, tables) {
  seq_along(tables) |>
    lapply(function(j) {
      tryCatch(
        {
          tibble::tibble(
            project = project,
            table = tables[j],
            # get permissions for each dataset inside each project
            opalr::opal.table_perm(x, project, tables[j])
          )
        },
        error = function(e) {
          if (grepl("HTTP 403", e$message)) {
            stop(
              "The provided connection does not have access to retrieve ",
              "table permissions!",
              call. = FALSE
            )
          } else if (grepl("HTTP 404", e$message)) {
            warning(
              "Error when retrieving permissions for ",
              paste0(project, ".", tables[j], "!"),
              call. = FALSE
            )
          } else {
            warning(e$message, call. = FALSE)
          }

          # return empty tibble, only with `project` and `table` details
          return(tibble::tibble(
            project = project,
            table = tables[j]
          ))
        }
      )
    }) |>
    # combine results from the permissions for each table
    dplyr::bind_rows()
}

#' Flatten user permission entities
#'
#' @param x List with entities, generated by [user_perm_entity()].
#'
#' @returns Tibble with properties of the entities.
#' @keywords internal
flatten_user_perm_entity <- function(x) {
  # local bindings
  agent <- object <- asset_id <- person_id <- NULL

  # flatten list of entities
  x_tbl <- x |>
    lapply(tibble::as_tibble, .name_repair = "minimal") |>
    dplyr::bind_rows()
  # check there were any entities found in the previous step
  if (nrow(x_tbl) == 0) {
    return(NULL)
  }
  # edit tibble with entities
  x_tbl |>
    dplyr::rename(
      perm_id = "@id",
      type = "@type",
      person_id = agent,
      asset_id = object
    ) |>
    dplyr::mutate(
      person_id = unlist(person_id),
      asset_id = unlist(asset_id),
    ) |>
    # add permission field, based on `type`
    dplyr::mutate(
      permission = dplyr::case_when(
        type == "ReadAction" ~ "read",
        type == "WriteAction" ~ "write",
        type == "ControlAction" ~ "administrate",
        TRUE ~ NA_character_
      )
    )
}

#' @noRd
infer_table_resource_lineage <- function(assets_tbl) {
  # local binding
  asset_type <- NULL

  tables <- dplyr::filter(assets_tbl, asset_type == "table")
  resources <- dplyr::filter(assets_tbl, asset_type == "resource")

  if (nrow(tables) == 0 || nrow(resources) == 0) {
    return(NULL)
  }

  purrr::map(seq_len(nrow(tables)), function(i) {
    tbl <- tables[i, ]

    matches <- purrr::map(resources$meta, function(m) {
      !is.null(m$table) && identical(m$table, tbl$name)
    }) |>
      purrr::list_c()

    if (!any(matches)) {
      return(NULL)
    }

    tibble::tibble(
      table_name = tbl$name,
      resource_name = resources$name[matches]
    )
  }) |>
    purrr::list_c()
}

#' Verify if connection was created by an administrative user
#'
#' @inheritParams validate_opal_con
#'
#' @returns Boolean flag to indicate whether the given connection was created
#'     by an administrative user.
#' @keywords internal
is_opal_admin_con <- function(x) {
  # validate connection
  validate_opal_con(x)

  # extract the user profile, `uprofile` from the connection object
  uprofile <- getElement(x, "uprofile")

  # extract logged in user's groups
  groups <- getElement(uprofile, "groups") |>
    sapply(unlist)

  # check if the user has admin in their groups
  if ("admin" %in% groups) {
    return(TRUE)
  } else {
    return(FALSE)
  }
}

#' @noRd
link_assets_to_project <- function(rocrate, project_id, asset_ids) {
  proj_entity <- .get_entity(rocrate, id = project_id)[[1]]

  # check if no project entity was found (safe_data called before safe_project)
  if (is.null(proj_entity)) {
    return(rocrate)
  }

  # if project entity was found, check if it has a `hasPart` value
  if (is.null(proj_entity$hasPart)) {
    proj_entity$hasPart <- list()
  }

  rocrate |>
    rocrateR::add_entity_value(
      id = project_id,
      key = "hasPart",
      value = c(
        proj_entity$hasPart,
        lapply(asset_ids, function(id) list(`@id` = id))
      ),
      overwrite = TRUE
    )
}

#' @noRd
map_asset_type <- function(asset_type, meta, url) {
  if (asset_type == "table") {
    return("Dataset")
  }

  if (asset_type == "resource") {
    if (!is.null(meta$driver)) {
      return("Dataset")
    }
    if (grepl("^file:", url)) {
      return("File")
    }
    return("Dataset")
  }

  "DigitalDocument"
}

normalise_permission <- function(p) {
  allowed <- c("view", "view-values", "edit", "edit-values", "administrate")
  if (p %in% allowed) {
    return(p)
  }
  "view"
}

#' Parse project resources parametres
#'
#' @param params_json JSON object with resource params.
#'
#' @returns list with simplified JSON params.
#'
#' @noRd
parse_resource_params <- function(params_json) {
  purrr::map(params_json, function(p) {
    if (is.na(p) || p == "") {
      return(list())
    }
    jsonlite::fromJSON(p, simplifyVector = TRUE)
  })
}

safe_desc <- function(desc, name, type) {
  desc %||% paste(type, name)
}

safe_time <- function(x) {
  if (is.null(x) || is.na(x)) {
    return(character(1))
  }
  tryCatch(
    format(as.POSIXct(x), "%Y-%m-%dT%H:%M:%SZ"),
    error = function(e) character(1)
  )
}

#' Update datasets linked to a project
#'
#' Update datasets linked to a project (`hasPart`)
#'
#' @inheritParams safe_data
#' @param ds_ids Vector with `@id`s of the datasets to be linked to `project`
#'
#' @returns Update RO-Crate with updated project entity
#' @keywords internal
update_project_datasets <- function(rocrate, project, ds_ids) {
  # attempt to retrieve the project entities to link up the IDs to the project
  # this only valid if safe_project is called before safe_data
  suppressWarnings({
    project_ents <- rocrate |>
      .get_entity(type = "Project") |>
      # only keep entity for `project`
      Filter(f = \(x) getElement(x, "name") == project)
  })

  # if any entity was found, then filter to keep those for which their @id
  # starts with `project_id_suffix` as set by `safe_project()`:
  if (length(project_ents) == 1 && !is.null(project)) {
    # extract the `hasPart` section
    has_part <- project_ents |>
      sapply("[[", "hasPart") |>
      unlist()

    # update the `hasPart` section
    rocrate <- rocrate |>
      rocrateR::add_entity_value(
        id = project_ents[[1]]["@id"],
        key = "hasPart",
        value = c(has_part, ds_ids) |>
          unique() |>
          lapply(\(id) list(`@id` = id)),
        overwrite = TRUE
      )
  }

  # return update RO-Crate
  return(rocrate)
}

#' @noRd
user_asset_perm_entities <- function(
  person,
  person_id,
  asset,
  asset_id,
  permission,
  asset_type = c("table", "resource")
) {
  asset_type <- match.arg(asset_type)

  # reuse existing function by aliasing "table" args
  user_perm_entity(
    person = person,
    person_id = person_id,
    asset = asset,
    asset_id = asset_id,
    permission = permission
  )
}

#' Create user/person permission entities
#'
#' @param person String with person name/username.
#' @param person_id String with person `@id`.
#' @param asset String with dataset/table/resource name.
#' @param asset_id String with dataset/table `@id`.
#' @param permission String with permission ('view', 'view-values', 'edit',
#'     'edit-values' OR 'administrate').
#' @param ... Other additional values.
#'
#' @returns List of [rocrateR::entity] objects
#' @keywords internal
user_perm_entity <- function(
  person,
  person_id,
  asset,
  asset_id,
  permission,
  ...
) {
  # set local bindings
  description <- type <- NULL
  action_status <- "PotentialActionStatus"
  agent <- list(list(`@id` = person_id))
  object <- list(list(`@id` = asset_id))

  # create combined @id
  comb_id <- id_hash("#perm:", paste0(person, "-", asset))

  # update combine @id, @type and description based on permission
  if (permission == "view") {
    comb_id <- paste0(comb_id, "-dict-summary-read")
    type <- "ReadAction"
    description <- "User may view table dictionary and summary statistics only; access to individual values is restricted."
  } else if (permission == "view-values") {
    comb_id <- paste0(comb_id, "-read-all")
    type <- "ReadAction"
    description <- "User may view table dictionary and all individual values."
  } else if (permission == "edit") {
    comb_id <- paste0(comb_id, c("-write-dict", "-summary-read"))
    type <- c("WriteAction", "ReadAction")
    description <- c(
      "User may edit the table dictionary but cannot view individual values.",
      "User may view summary statistics only; access to individual values is restricted."
    )
  } else if (permission == "edit-values") {
    comb_id <- paste0(comb_id, c("-write-dict", "-read-all"))
    type <- c("WriteAction", "ReadAction")
    description <- c(
      "User may edit the table dictionary.",
      "User may view table dictionary and all individual values."
    )
  } else if (permission == "administrate") {
    comb_id <- paste0(comb_id, "-admin-table")
    type <- "ControlAction"
    description <- "User has full administrative rights: view/edit dictionary and view/edit individual values."
  } else {
    return(NULL)
  }

  # create entity objects
  permission_entities_tbl <- tibble::tibble(
    id = comb_id,
    type = type,
    agent = agent,
    object = object,
    actionStatus = action_status,
    description = description
  )

  # seq_len(nrow(permission_entities_tbl)) |>
  #   lapply(function(i) {
  #     rocrateR::entity(permission_entities_tbl$`id`[i],
  #                      type = permission_entities_tbl$type[i])
  #   })
  purrr::pmap(permission_entities_tbl, rocrateR::entity)
  # rocrateR::entity(as.list(permission_entities_tbl))
}

#' Validate OBiBa's Opal connection
#'
#' @param x Connection to OBiBa's Opal server (see [opalr::opal.login()]).
#'
#' @returns Nothing, call for its side effect.
#' @keywords internal
validate_opal_con <- function(x) {
  tryCatch(
    {
      status <- xptr::is_null_xptr(x$handle$handle)
      if (status) {
        stop("The given connection is not valid!", call. = FALSE)
      }
    },
    error = function(e) {
      stop("The given connection is not valid!", call. = FALSE)
    }
  )
}

test_that("validate_opal_con works with real Opal connection", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  expect_no_error(validate_opal_con(opal_con))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("is_opal_admin_con detects admin group correctly", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  res <- is_opal_admin_con(opal_con)

  expect_type(res, "logical")
  expect_true(res)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("get_project_tables retrieves real tables from demo server", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  project <- attr(opal_con, "PROJECT")

  tables <- get_project_tables(opal_con, project)

  expect_type(tables, "character")
  expect_true(length(tables) >= 1)
  expect_true(all(nzchar(tables)))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("get_project_details works end-to-end with demo Opal project", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  project <- attr(opal_con, "PROJECT")

  res <- get_project_details(opal_con, project)

  expect_s3_class(res, "data.frame")
  expect_true("project" %in% names(res))
  expect_true("table" %in% names(res))
  expect_true(nrow(res) >= 1)
  expect_true(all(res$project == project))

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("get_table_permissions errors for non-admin connection", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con(admin = FALSE)
  project <- attr(opal_con, "PROJECT")
  tables <- attr(opal_con, "TABLES")

  expect_error(
    get_table_permissions(opal_con, project, tables),
    "The provided connection does not have access to retrieve table permissions!"
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("get_table_permissions throws a warning for an invalid project", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  tables <- attr(opal_con, "TABLES")

  expect_warning(
    get_table_permissions(opal_con, "INVALID PROJECT", tables),
    "404"
  )

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("get_table_permissions retrieves real permissions", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  project <- attr(opal_con, "PROJECT")
  tables <- attr(opal_con, "TABLES")

  res <- get_table_permissions(opal_con, project, tables)

  expect_s3_class(res, "data.frame")
  expect_true(all(c("project", "table") %in% names(res)))
  expect_equal(unique(res$project), project)
  expect_equal(unique(res$table), tables)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("get_project_details handles non-existent project gracefully", {
  opal_con <- opal_demo_con()

  res <- get_project_details(opal_con, "THIS_PROJECT_DOES_NOT_EXIST_123")

  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 0)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("validate_opal_con errors on invalid connection object", {
  bad_con <- list(handle = list(handle = NULL))

  expect_error(
    validate_opal_con(bad_con),
    "connection is not valid"
  )
})

test_that("update_project_datasets updates project entities of an RO-Crate", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()

  # create basic RO-Crate with a project
  rocrate <- rocrateR::rocrate_5s() |>
    dsROCrate::safe_project(
      connection = opal_con,
      project = attr(opal_con, "PROJECT")
    )

  # update project datasets
  expect_no_error(
    rocrate <- rocrate |>
      update_project_datasets(project = attr(opal_con, "PROJECT"), ds_ids = 1:5)
  )

  # extract `hasPart` for the project entity
  expect_no_error(
    has_part <- rocrateR::get_entity(rocrate, type = "Project") |>
      sapply(getElement, name = "hasPart") |>
      sapply(getElement, name = "@id") |>
      unlist()
  )

  expect_equal(length(has_part), 5)
  expect_equal(has_part, 1:5)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("user_perm_entity works for all values of 'permission'", {
  input_tbl <- tibble::tibble(
    person = "dsuser",
    person_id = "dsuser",
    asset = "tab1",
    asset_id = "tab1",
    permission = c(
      "view",
      "view-values",
      "edit",
      "edit-values",
      "administrate"
    )
  )

  expect_no_error(
    output_tbl <- purrr::pmap(input_tbl, user_perm_entity)
  )

  expect_length(output_tbl, 5)
  expect_type(output_tbl, "list")
})

test_that("user_perm_entity returns NULL for unknown 'permission'", {
  input_tbl <- tibble::tibble(
    person = "dsuser",
    person_id = "dsuser",
    asset = "tab1",
    asset_id = "tab1",
    permission = "INVALID"
  )

  expect_no_error(
    output_tbl <- purrr::pmap(input_tbl, user_perm_entity)
  )

  expect_null(output_tbl[[1]])
  expect_length(output_tbl, 1)
  expect_type(output_tbl, "list")
})

#' Extract Safe Data entity(ies)
#'
#' @inheritParams safe_data
#' @param ... Other optional arguments. See the full documentation
#'
#' @returns RO-Crate with Safe Data entity(ies).
#' @rdname extract_safe_data
#' @keywords internal
extract_safe_data <- function(x, ...) {
  UseMethod("extract_safe_data")
}

#' @param rocrate (Optional) RO-Crate object to update with Safe Data details.
#' @rdname extract_safe_data
#' @export
extract_safe_data.opal <- function(x, ..., rocrate = rocrateR::rocrate_5s()) {
  # extract all data sources
  ds <- opalr::opal.datasources(x)

  # extract project names and ignore NAs
  projects <- ds$name[!is.na(ds$name) & !is.null(ds$name)]

  # create RO-Crate with Safe Data details for the projects found on the server
  purrr::reduce(
    projects,
    \(crate, p) {
      safe_data(crate, connection = x, project = p)
    },
    .init = rocrate
  )
}

#' @param id (Optional) Vector with `@id` strings for Safe Data entity(ies)
#'     to be extracted from the given RO-Crate, `x`.
#' @rdname extract_safe_data
#' @export
extract_safe_data.rocrate <- function(
  x,
  ...,
  id = NULL,
  asset_id_suffix = "#asset:",
  rocrate = rocrateR::rocrate_5s()
) {
  # validate RO-Crate
  rocrateR::is_rocrate(x)

  # filter out asset entities associated with the project based on the
  # value for `asset_id_suffix`.
  entities_lst <- x$`@graph` |>
    purrr::keep(\(x) grepl(paste0("^", asset_id_suffix), x$`@id`))

  # if `id` was provided, then filter out only those entities
  if (!is.null(id)) {
    entities_lst <- entities_lst |>
      purrr::keep(\(x) getElement(x, "@id") %in% id)
  }

  # check if any entities were found
  if (length(entities_lst) == 0) {
    stop("No matching entities were found!", call. = FALSE)
  } else {
    message(
      length(entities_lst),
      " asset entit",
      ifelse(length(entities_lst) == 1, "y was", "ies were"),
      " found!"
    )
  }

  # add entities to the RO-Crate
  suppressWarnings({
    purrr::reduce(entities_lst, rocrateR::add_entity, .init = rocrate)
  })
}

#' Flatten object with Safe Data details
#'
#' @param x Object (e.g., RO-Crate) with Safe Data details. This can be
#'     generated with the [extract_safe_data()] function.
#' @param id Vector of strings with the `@id`s for the datasets to be extracted.
#'     If not provided, extract all entities with `@type = 'Dataset'`.
#'
#' @returns Data frame with Safe Data details.
#' @rdname flatten_safe_data
#' @keywords internal
flatten_safe_data <- function(x, ...) {
  UseMethod("flatten_safe_data")
}

#' @rdname flatten_safe_data
#' @export
flatten_safe_data.default <- function(x, ...) {
  return(tibble::tibble())
}

#' @rdname flatten_safe_data
#' @export
flatten_safe_data.rocrate <- function(
  x,
  ...,
  id = NULL,
  asset_id_suffix = "#asset:"
) {
  # local bindings
  asset_id <- person_id <- NULL

  tryCatch(
    {
      # extract asset entities
      entities_tbl <- x$`@graph` |>
        purrr::keep(\(x) grepl(paste0("^", asset_id_suffix), x$`@id`)) |>
        # extract @id and name for each entity
        lapply(function(ent) {
          tibble::tibble(
            asset_id = getElement(ent, "@id"),
            asset = getElement(ent, "name")
          )
        }) |>
        # combine all rows
        dplyr::bind_rows() |>
        # filter out root (./) entity
        dplyr::filter(asset_id != "./")

      # if `id` is provided, then only keep those entities
      if (!is.null(id)) {
        entities_tbl <- entities_tbl |>
          dplyr::filter(asset_id %in% !!id)
      }

      # return dataset entities
      return(entities_tbl)
    },
    error = function(e) {
      tibble::tibble()
    }
  )
}

test_that("extract_safe_data.opal returns a valid RO-Crate with datasets", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  # terminate connection when done with tests
  withr::defer(opalr::opal.logout(opal_con))

  # create local mock binding, to avoid long run times
  local_mocked_bindings(
    opal.datasources = function(...) {
      tibble::tibble(name = attr(opal_con, "PROJECT"))
    },
    .package = "opalr"
  )

  # ignore warning of project not having tables associated
  suppressWarnings(
    roc <- extract_safe_data(opal_con)
  )

  # validate rocrate structure
  expect_no_error(rocrateR::is_rocrate(roc))

  # should contain Dataset entities (if demo server has datasources)
  datasets <- rocrateR::get_entity(roc, type = "Dataset")

  expect_type(datasets, "list")
  expect_true(length(datasets) >= 1)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("extract_safe_data.opal updates an existing RO-Crate", {
  # open connection to OBiBa's Opal demo server
  opal_con <- opal_demo_con()
  # terminate connection when done with tests
  withr::defer(opalr::opal.logout(opal_con))

  base_roc <- rocrateR::rocrate_5s()

  # create local mock binding, to avoid long run times
  local_mocked_bindings(
    opal.datasources = function(...) {
      tibble::tibble(name = attr(opal_con, "PROJECT"))
    },
    .package = "opalr"
  )

  # ignore warning of project not having tables associated
  suppressWarnings(
    roc <- extract_safe_data(opal_con, rocrate = base_roc)
  )

  # validate rocrate structure
  expect_no_error(rocrateR::is_rocrate(roc))

  datasets <- rocrateR::get_entity(roc, type = "Dataset")
  expect_true(length(datasets) >= 1)

  # close connection to OBiBa's Opal demo server
  opalr::opal.logout(opal_con)
})

test_that("extract_safe_data.rocrate copies Dataset entities", {
  # create source crate with Dataset entities
  src <- rocrateR::rocrate_5s()

  src <- rocrateR::add_entity(
    src,
    entity = rocrateR::entity(
      id = "#asset:1",
      type = "Dataset",
      name = "Dataset 1"
    )
  )

  src <- rocrateR::add_entity(
    src,
    entity = rocrateR::entity(
      id = "#asset:2",
      type = "Dataset",
      name = "Dataset 2"
    )
  )

  # ignore warning of project not having tables associated
  suppressWarnings(
    new_roc <- extract_safe_data(src)
  )

  # validate rocrate structure
  expect_no_error(rocrateR::is_rocrate(new_roc))

  ents <- rocrateR::get_entity(new_roc, type = "Dataset")
  expect_equal(length(ents), 3) # 2 + 1 root entity
})

test_that("extract_safe_data.rocrate filters Dataset entities by id", {
  # create source crate with Dataset entities
  src <- rocrateR::rocrate_5s()

  src <- rocrateR::add_entity(
    src,
    entity = rocrateR::entity(
      id = "#asset:keep_me",
      type = "Dataset",
      name = "Keep"
    )
  )

  src <- rocrateR::add_entity(
    src,
    entity = rocrateR::entity(
      id = "#asset:drop_me",
      type = "Dataset",
      name = "Drop"
    )
  )

  # ignore warning of project not having tables associated
  suppressWarnings(
    new_roc <- extract_safe_data(src, id = "#asset:keep_me")
  )

  ents <- rocrateR::get_entity(new_roc, type = "Dataset")
  ids <- vapply(ents, function(e) e[["@id"]], character(1))

  expect_equal(ids, c("./", "#asset:keep_me"))
})

test_that("extract_safe_data.rocrate errors when no Dataset entities exist", {
  # create source crate with Dataset entities
  src <- rocrateR::rocrate_5s()

  expect_error(
    extract_safe_data(src),
    "No matching entities were found"
  )
})

test_that("flatten_safe_data.default returns empty tibble", {
  res <- flatten_safe_data(list())

  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 0)
})

test_that("flatten_safe_data.rocrate extracts id and name correctly", {
  roc <- rocrateR::rocrate_5s()

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "#asset:1",
      type = "Dataset",
      name = "Dataset One"
    )
  )

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "#asset:2",
      type = "Dataset",
      name = "Dataset Two"
    )
  )

  res <- flatten_safe_data(roc)

  expect_s3_class(res, "data.frame")
  expect_true(all(c("asset_id", "asset") %in% names(res)))
  expect_equal(nrow(res), 2)
})

test_that("flatten_safe_data.rocrate filters by id argument", {
  roc <- rocrateR::rocrate_5s()

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "#asset:A",
      type = "Dataset",
      name = "A"
    )
  )

  roc <- rocrateR::add_entity(
    roc,
    entity = rocrateR::entity(
      id = "#asset:B",
      type = "Dataset",
      name = "B"
    )
  )

  res <- flatten_safe_data(roc, id = "#asset:A")

  expect_equal(nrow(res), 1)
  expect_equal(res$asset_id, "#asset:A")
})

test_that("flatten_safe_data.rocrate returns empty tibble on error", {
  bad_obj <- list() # not a rocrate

  res <- flatten_safe_data.rocrate(bad_obj)

  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 0)
})

audit_intent <- function(intent, excluded_args = c("project", "user"), ...) {
  # if `intent` is NOT NULL, audit this object
  intent_audit <- if (!is.null(intent)) {
    audit(intent, ...)
  } else {
    NULL
  }

  # list of additional args
  main_audit_args <- list(...)

  # check if `intent_audit` is not NULL
  if (!is.null(intent_audit)) {
    # extract Project(s) and People from the `intent` audit crate
    safe_project_tbl <- intent_audit |>
      flatten_safe_project()
    safe_people_tbl <- intent_audit |>
      flatten_safe_people()

    main_audit_args <- exclude_args(main_audit_args, excluded = excluded_args)

    main_audit_args <- c(
      main_audit_args,
      project = safe_project_tbl$project,
      user = safe_people_tbl$name
    )
  }

  list(intent_audit = intent_audit, main_audit_args = main_audit_args)
}

exclude_args <- function(..., excluded) {
  # capture additional args
  args <- list(...)
  arg_names <- names(args)
  # exclude args that shouldn't be passed to the next function
  args[!(arg_names %in% excluded)]
}

#' Append entity reference
#'
#' Adds a reference to an existing property while
#' preserving existing values and avoiding duplicates.
#'
#' @param rocrate RO-Crate object.
#' @param id String with `@id` of the entity.
#' @param key String with property/key of the entity.
#' @param ref_id String with reference `@id`.
#'
#' @returns Updated RO-Crate object.
#' @keywords internal
append_entity_ref <- function(rocrate, id, key, ref_id) {
  entity <- rocrateR::get_entity(rocrate, id = id)[[1]]

  current <- entity[[key]]

  existing_ids <- if (!is.null(current)) {
    getElement(current, "@id") %||% purrr::map_chr(current, "@id")
  } else {
    character()
  }

  ids <- unique(c(existing_ids, ref_id))

  value <- purrr::map(ids, ~ list("@id" = .x))

  rocrateR::add_entity_value(
    rocrate,
    id = id,
    key = key,
    value = value,
    overwrite = TRUE
  )
}

#' Wrapper for [rocrateR::get_entity()]
#'
#' This wrapper is used to suppress warning messages like
#' 'No matching entities were found with ...'.
#'
#' @importFrom rocrateR get_entity
#'
#' @returns List with entity of objects
#' @noRd
.get_entity <- function(rocrate, id = NULL, type = NULL) {
  suppressWarnings(rocrateR::get_entity(rocrate, id = id, type = type))
}

#' Open connection to OBiBa's Opal demo server
#'
#' @param admin Boolean flag to indicate if the connection should be with admin
#'     rights (default behaviour).
#'
#' @returns Connection object to OBiBa's Opal demo server with extra attributes.
opal_demo_con <- function(admin = TRUE) {
  # Skip tests if offline or on CRAN
  skip_on_cran()
  skip_if_offline()

  ## Opal server access
  USERNAME <- "administrator"
  USERPASS <- "password"
  SERVER <- "https://opal-demo.obiba.org"
  ## Credentials for `dsuser`
  ### NOTE: this is only used to simulate an analysis and generate logs
  DSUSERPASS <- "P@ssw0rd"

  ## Five safes variables
  PEOPLE <- "dsuser"
  PROJECT <- "CNSIM"
  TABLES <- c("CNSIM1")

  # login to local server with `USERNAME` and `USERPASS`.
  ## as administrator
  if (admin) {
    opal_con <- opalr::opal.login(
      username = USERNAME,
      password = USERPASS,
      url = SERVER
    )
  } else {
    # as non-admin user
    opal_con <- opalr::opal.login(
      username = PEOPLE,
      password = DSUSERPASS,
      url = SERVER
    )
  }

  attr(opal_con, "USERNAME") <- USERNAME
  attr(opal_con, "USERPASS") <- USERPASS
  attr(opal_con, "SERVER") <- SERVER
  attr(opal_con, "DSUSERPASS") <- DSUSERPASS
  attr(opal_con, "PEOPLE") <- PEOPLE
  attr(opal_con, "PROJECT") <- PROJECT
  attr(opal_con, "TABLES") <- TABLES

  return(opal_con)
}

#' Audit Engine
#'
#' Internal function to create audits for various back-ends.
#'
#' @param x This can be a connection to a 'DataSHIELD' server (e.g., object with
#'     the `opal` class, see [opalr::opal.login()]). Alternatively, a governance
#'     archive file, representing the intent of a project and associated
#'     governance details.
#' @param ... Other optional arguments, see full documentation for details.
#' @param project String with project name(s) from which to extra Safe Project
#'     details.
#' @param user String with the user name for which to extract Safe People
#'     details.
#' @param logs_from Lower limit timestamp to filter out the outputs generated
#'     (default: `-Inf`, everything up to `logs_to`)
#' @param logs_to Upper limit timestamp to filter out the outputs generated
#'     (default: `Inf`, everything from `logs_from` onwards).
#' @param path String with path pointing to the root of the RO-Crate. This will
#'     be used to store log files. If not provided, logs will be stored within
#'     the RO-Crate returned by this function.
#'
#' @returns Audit RO-Crate with 5 Safes Components.
#' @keywords internal
audit_engine <- function(x, ...) {
  UseMethod("audit_engine")
}

#' @rdname audit_engine
#' @export
audit_engine.cr8tor <- function(x, ...) {
  # extract individual components from cr8tor bundle
  audit <- list(
    metadata = extract_cr8tor_metadata(x),
    integrity = extract_integrity_cr8tor(x),
    safe_people = extract_safe_people_cr8tor(x),
    safe_projects = extract_safe_projects_cr8tor(x),
    safe_data = extract_safe_data_cr8tor(x),
    safe_settings = extract_safe_settings_cr8tor(x),
    safe_outputs = extract_safe_outputs_cr8tor(x),
    user_projects = extract_user_projects_cr8tor(x),
    user_groups = extract_user_groups_cr8tor(x),
    groups = extract_groups_cr8tor(x),
    permissions = extract_permissions_cr8tor(x)
  )

  as_rocrate_audit(audit)
}

#' @rdname audit_engine
#' @export
audit_engine.opal <- function(
  x,
  ...,
  project = NULL,
  user = NULL,
  logs_from = -Inf,
  logs_to = Inf,
  path = NULL
) {
  # local bindings
  name <- principal <- NULL

  # create RO-Create with the 5 safes profile
  crate <- rocrateR::rocrate_5s()

  # validate Opal connection
  is_opal_admin_con(x)

  # if `project` is missing, then ~extract all project names~ error
  if (is.null(project)) {
    stop("A `project` name is required!", call. = FALSE)
  }

  # extract all data sources to verify `project` contains a valid value.
  ds <- opalr::opal.datasources(x)
  server_prjs <- ds[, "name"]
  idx <- project %in% server_prjs
  if (!all(idx)) {
    stop(
      "The following project",
      ifelse(length(idx) == 1, " is ", "s are "),
      "not valid: \n",
      paste0(" - ", project[!idx], collapse = "\n"),
      call. = FALSE
    )
  }

  # Safe People ----
  # get users' details
  safe_people_tbl <- opalr::oadmin.user_profiles(x, df = FALSE) |>
    dplyr::bind_rows() |>
    dplyr::rename(name = principal) |>
    # exclude system administrators from the report
    dplyr::filter(!(tolower(name) %in% c("admin", "administrator")))

  if (!is.null(user)) {
    safe_people_tbl <- safe_people_tbl |>
      dplyr::filter(tolower(name) %in% user)

    if (nrow(safe_people_tbl) == 0) {
      stop(
        sprintf(
          "No Safe People details were found for the user: %s!",
          paste0("'", user, "'", collapse = ", ")
        ),
        call. = FALSE
      )
    }
  }

  crate <- safe_people_tbl$name |>
    purrr::reduce(
      \(crate, u) {
        safe_people(
          crate,
          connection = x,
          user = u,
          set_author = FALSE,
          set_project = FALSE
        )
      },
      .init = crate
    )

  # Safe Projects ----
  crate <- project |>
    purrr::reduce(
      \(crate, p) {
        safe_project(crate, connection = x, project = p)
      },
      .init = crate
    )

  # Safe Data ----
  crate <- project |>
    purrr::reduce(
      \(crate, p) {
        safe_data(crate, connection = x, project = p)
      },
      .init = crate
    )

  # remove permissions associated with admin users
  non_admin_user_ids <- safe_people_tbl$name |>
    purrr::map_chr(id_hash, prefix = "#person:")
  admin_perm_ents <- crate$`@graph` |>
    purrr::keep(\(x) grepl("^#perm:", getElement(x, "@id"))) |>
    purrr::discard(\(x) getElement(x, "agent")[[1]] %in% non_admin_user_ids)
  crate <- admin_perm_ents |>
    purrr::reduce(rocrateR::remove_entity, .init = crate)

  # Safe Settings ----
  crate <- safe_setting(x, rocrate = crate)

  # Safe Outputs ----
  crate <- safe_people_tbl$name |>
    purrr::reduce(
      \(crate, u) {
        safe_output(
          crate,
          connection = x,
          path = path,
          user = u,
          logs_to = logs_to,
          logs_from = logs_from
        )
      },
      .init = crate
    )
}

test_that("All Rd files have a value section", {
  db <- tools::Rd_db("dsROCrate")

  # drop the package-level Rd (e.g., "dsROCrate-package")
  db <- db[!grepl("-package\\.Rd$", names(db))]

  has_value <- function(rd) {
    any(vapply(rd, function(x) attr(x, "Rd_tag") == "\\value", logical(1)))
  }

  results <- vapply(db, has_value, logical(1))
  missing <- names(db)[!results]

  expect_equal(
    length(missing),
    0L,
    info = paste("Missing \\value in:\n", paste(" -", missing, collapse = "\n"))
  )
})

---
title: "Configuring Audit Permissions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Configuring Audit Permissions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Overview

`{dsROCrate}` requires elevated permissions when auditing backend systems.

Depending on the backend implementation, this may require:

- administrator privileges,
- audit privileges,
- or backend-specific read access to system metadata.

Without these permissions, `{dsROCrate}` may be unable to retrieve
the metadata required for audit and RO-Crate generation workflows.

This vignette describes:

- why elevated permissions are required,
- how permission validation works,
- and how to configure supported backends.


# Why are elevated permissions required?

`dsROCrate` performs metadata and configuration inspection operations
against supported backend systems.

These operations may include access to:

- system configuration metadata,
- authentication configuration,
- datasource metadata,
- project metadata,
- audit-relevant server settings.

Standard user accounts may not have sufficient privileges for these
operations.


# Supported backends

Currently supported backend permission configurations:

- Opal

Future backend-specific instructions may include:

- Armadillo


# Opal

**NOTE:** Opal 5.7+ is required, as that's when the _audit_ role was introduced.

## Required permissions

For Opal backends, the connected user must have at least one of:

- administrator privileges, or
- the `Audit system` permission.

## Configuring an Opal audit user

### Step 1: Log in as an administrator

Log in to the Opal web interface using an administrator account.

### Step 2: Create a user (optional)

If a dedicated audit user does not already exist:

1. Open the **Administration** tab.
2. Under **Data Access**, select **Users and Groups**.
3. Create a new user account for auditing purposes.

### Step 3: Grant audit permissions

1. Return to the **Administration** tab.
2. Under **System**, select **General Settings**.
3. Open the **Permissions** section.
4. Click the **+** button.
5. Select:
   - **Add user permission**
6. Enter the username of the audit user.
7. Select:
   - **Audit system**
8. Click **Submit**.


--------------------------------------------------------------------------------

# Example usage

```{r eval = FALSE}
con <- opalr::opal.login(
  username = "audit_user",
  password = "secret",
  url = "https://opal-demo.obiba.org"
)

dsROCrate::check_permissions(con)
```


# Troubleshooting

## Insufficient permissions

Ensure the connected user has the permissions required for the
target backend.

See the backend-specific sections in this vignette for details.

## Testing Opal permissions directly

```{r eval = FALSE}
dsROCrate::is_opal_admin_con(con)
dsROCrate::is_opal_audit_con(con)
```

---
title: "Deploy a local DataSHIELD server with Opal"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Deploy a local DataSHIELD server with Opal}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  cache = FALSE
)
```


```{r}
#| echo: false
#| label: setup
library(dsROCrate)
```

### What is Opal?

> Opal is [OBiBa](https://www.obiba.org/)’s core database application for epidemiological studies. Participant data, collected by questionnaires, medical instruments, sensors, administrative databases etc. can be integrated and stored in a central data repository under a uniform model. Opal is a web application that can import, process, copy data and has advanced features for cataloguing the data (fully described, annotated and searchable data dictionaries) as recommended by the [Maelstrom Research](https://www.maelstrom-research.org/) group at McGill University, Canada. Opal is typically used by a research centre to analyse the data acquired from assessment centres. Its ultimate purpose is to achieve seamless data-sharing among epidemiological studies. Opal is the reference implementation of the [DataSHIELD](https://datashield.org/) infrastructure. More information on Opal can be found in the [Opal description on OBiBa](https://www.obiba.org/pages/products/opal/).

**Source:** [https://isglobal-brge.github.io/resource_bookdown/opal.html](https://isglobal-brge.github.io/resource_bookdown/opal.html)

### Deploy Docker containers
Here we will spawn a local instance of [DataSHIELD](https://datashield.org) with 
[Docker](https://www.docker.com). We will assume you have installed and
configured Docker on your computer; however, if that's not the case, visit their
[get started with Docker](https://www.docker.com/get-started/) page.

#### Setup
The easiest way to deploy DataSHIELD with docker is by cloning the following
repo: [FederatedMethods/datashield_dev_install](https://github.com/FederatedMethods/datashield_dev_install).
Here, you will find a step by step guide, including a very useful
[`docker-compose.yml`](https://github.com/FederatedMethods/datashield_dev_install/blob/main/docker/docker-compose.yml) 
file, which you can use out of the box.

If you are running Linux or macOS, you can run the following commands:

```sh
git clone https://github.com/FederatedMethods/datashield_dev_install
cd datashield_pcr/docker
docker compose up -d
```

Then you can inspect the Docker GUI, where you should see something like the
following:

<img src="../man/figures/ds_docker_compose.png" alt="Docker conf" height=200px/>


---------

#### Open connection

By default, the [`docker-compose.yml`](https://github.com/FederatedMethods/datashield_dev_install/blob/main/docker/docker-compose.yml) 
file in the repo above defines a demo user, `demo_user`, with the following 
password: `Demo_password1!` (edit the variables `OPAL_DEMO_USER_NAME` and
`OPAL_DEMO_USER_PASSWORD` accordingly). Here, we will open a connection to our
local server using these credentials:

```{r}
# define global variables
USERNAME <- "demo_user"
USERPASS <- "Demo_password1!"
PROJECT <- "DEMO"
TABLES <- c("CNSIM1")
SERVER <- "https://opal-demo.obiba.org"
PROFILE <- "demo"
```

--------

#### Test deployment

1. On a web browser, navigate to `http://localhost:8880` <!-- @URLok --> 
and use the credentials on the `docker-compose.yml` to log in (e.g., `administrator` and `password`).
       
2. Inside RStudio (or your preferred IDE), you can test the connection to the local server with the following command (note that you might need to update the `username` and `password`): 
 
     ```r
     opalr::opal.login( 
         username = "administrator", 
         password = "password", 
         url = "https://localhost:8843/" 
     )
     ```

     You should see something like the following:
     ```bash
     url: https://localhost:8843  
     name: localhost  
     version: 5.2.0  
     username: administrator 
     ```

3. (Optional) Next, you can attempt running some commands with `{dsBaseClient}`:
 
    ```r
     # for setup only 
     DSOpal::Opal() 
     # create new login object 
     builder <- DSI::newDSLoginBuilder() 
     builder$append(server = "study1", 
                    url = "https://localhost:8843/", 
                    user = "administrator", 
                    password = "password", 
                    driver = "OpalDriver") 
     logindata <- builder$build() 
     conns <- DSI::datashield.login(logins = logindata) 
      
     # assign table to symbol, `test` 
     DSI::datashield.assign.table(conns["study1"],  
                                  symbol = "test", 
                                  table = "DEMO.CNSIM1", 
                                  errors.print = TRUE) 
      
     dsBaseClient::ds.ls(datasources = conns["study1"]) 
     dsBaseClient::ds.summary("test")
    ```

---
title: "Getting Started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  cache = FALSE,
  out.width = "100%",
  cache = FALSE,
  screenshot.opts = list(vwidth = 2000, vheight = 600, zoom = 3, selector = "div.html-widget")
)

# save the built-in output hook
hook_output <- knitr::knit_hooks$get("output")
hook_message <- knitr::knit_hooks$get("message")

# define a truncation helper
truncate_lines_tail <- function(x, lines) {
  x <- unlist(strsplit(x, "\n"))
  more <- "..."
  if (length(lines) == 1) {
    if (length(x) > lines) {
      # truncate the output, but add `more`
      # x <- c(head(x, n = lines), more)
      x <- c(more, tail(x, n = lines))
    }
  } else {
    x <- c(more, x[lines], more)
  }
  paste(c(x, ""), collapse = "\n")
}

# set the new hooks
knitr::knit_hooks$set(
  output = function(x, options) {
    lines <- options$out.lines
    if (is.null(lines)) return(hook_output(x, options))
    x <- truncate_lines_tail(x, lines)
    return(hook_output(x, options))
  },
  message = function(x, options) {
    lines <- options$out.lines
    if (is.null(lines)) return(hook_message(x, options))
    x <- truncate_lines_tail(x, lines)
    return(hook_message(x, options))
  }
)

# helper function to get the number of rows of an RO-Crate
rocrate_lines <- function(rocrate) {
  rocrate_txt(rocrate) |>
    length()
}

# helper function to read RO-Crate as text file
rocrate_txt <- function(rocrate) {
  # create tmp file to write a JSON file with the output
  tmp_file <- tempfile(fileext = ".json")
  on.exit(unlink(tmp_file, force = TRUE))
  # write RO-Crate in tmp file
  rocrateR::write_rocrate(rocrate, tmp_file)
  # read RO-Crate as text file
  readLines(tmp_file)
}
```

```{r setup}
library(dsROCrate)

# show all the lines in the RO-Crate
oopt <- options(max_lines = Inf)
```

This tutorial assumes that you have an internet connection and can access 
OBiBa's Opal demo server: https://opal-demo.obiba.org

Alternatively, if you want to test a local deployment, please check out the 
following vignette first:

```{r, eval = FALSE}
vignette("deploy-local-datashield-server-with-opal", package = "dsROCrate")
```

---------------------

## 1. Creating your first RO-Crate

> **&#9432;** Here we are using an administrator account, as this one is 
available by default on the demo server; however, we recommend you create a
dedicated auditing account with the `Audit system` permission level. See the 
vignette:
>
> `vignette('audit-permissions', package = 'dsROCrate')` 
>
> for further details.

### 1.1. Connect to an Opal server

Here we will use OBiBa's Opal demo server: https://opal-demo.obiba.org/ which
can be accessed with the following login credentials:

```{r}
# define global variables
## Opal server access
USERNAME <- "administrator"
USERPASS <- "password"
SERVER <- "https://opal-demo.obiba.org"
## Credentials for `dsuser`
### NOTE: this is only used to simulate an analysis and generate logs
DSUSERPASS <- "P@ssw0rd"
```

Next, define global variables used in generating the RO-Crate, such as project 
name, asset (e.g., tables, resources, etc.) references (within the project) and 
user identifiers.

```{r}
## Five safes variables
PEOPLE <- "dsuser"
PROJECT <- "CNSIM"
TABLES <- c("CNSIM1")
```

#### Open connection

Once the credentials and Five Safes variables are configured, we can start a 
new session on the Opal server with the following command:

```{r}
# login to local server with `USERNAME` and `USERPASS`.
o <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
)

print(o)
```

### 1.2. Create a basic RO-Crate

To create a basic RO-Crate, we will use the 
[`{rocrateR}`](https://github.com/ResearchObject/ro-crate-r) package. This 
package can be installed with the following command:

```{r, eval = FALSE, echo = TRUE}
# install.packages("pak")
pak::pak("rocrateR")

# for development version use
pak::pak("ResearchObject/ro-crate-r@dev")
```

Then, a basic RO-Crate can be created with the following command:

```{r}
basic_rocrate <- rocrateR::rocrate_5s()
```

Note that this RO-Crate uses the [5s-crate](https://trefx.uk/5s-crate/0.4/)
profile.

```{r}
print(basic_rocrate)
```

### 1.3. Add the _Five Safes_ Elements
#### Safe Data
To add details for Safe Data, use the function `dsROCrate::safe_data()`.

```{r, out.lines=-(rocrate_lines(basic_rocrate) - 2)}
basic_rocrate <- o |>
  dsROCrate::safe_data(rocrate = basic_rocrate,
                       project = PROJECT,
                       tables = TABLES)

print(basic_rocrate) # note that the output will be truncated
```

#### Safe Project
To add details for Safe Project, use the function `dsROCrate::safe_project()`.

```{r, out.lines=-(rocrate_lines(basic_rocrate) - 2)}
basic_rocrate <- o |>
  dsROCrate::safe_project(rocrate = basic_rocrate,
                          project = PROJECT)

print(basic_rocrate) # note that the output will be truncated
```

#### Safe People
To add details for Safe People, use the function `dsROCrate::safe_people()`.

```{r safe_people, out.lines=-(rocrate_lines(basic_rocrate) + 1)}
basic_rocrate <- o |>
  dsROCrate::safe_people(rocrate = basic_rocrate, user = PEOPLE)

print(basic_rocrate) # note that the output will be truncated
```

#### Safe Setting
To add details for Safe Setting, use the function `dsROCrate::safe_setting()`.

**&#9432;️ NOTE:** The `dsROCrate::safe_setting` function requires administrator
or auditor privileges, so here, we will have to log in with administrator
credentials (if you used a non-administrator/auditor account previously).

```{r, eval = FALSE}
# close previous connection
opalr::opal.logout(o)

# open new connection as administrator
o <- opalr::opal.login(
  username = "administrator",
  password = "password",
  url = SERVER
)
```

Then, we can proceed as per usual:
```{r, out.lines=-(rocrate_lines(basic_rocrate) - 2)}
basic_rocrate <- o |>
  dsROCrate::safe_setting(rocrate = basic_rocrate)

print(basic_rocrate) # note that the output will be truncated
```

#### Safe Outputs
To add details for Safe Outputs, use the function `dsROCrate::safe_output()`.
Currently, only log files from the operations executed by the user within a 
specific period. Set the period using `logs_from` and `logs_to`. Additionally,
a list of functions executed by the user are extracted in a separate file/entity.

**&#9432; NOTE:** Similar to `dsROCrate::safe_setting`, the `dsROCrate::safe_output` 
function requires of administrator or auditor privileges, so here, we will have to log in with 
administrator/auditor credentials:

```{r, eval = FALSE}
# close previous connection
opalr::opal.logout(o)

# open new connection as administrator
o <- opalr::opal.login(
  username = "administrator",
  password = "password",
  url = SERVER
)
```

---------

##### DataSHIELD operations

**&#9432; NOTE:** Before extracting logs, ensure there is recent activity on 
the server for testing purposes. This can be done using the following commands:

###### Setup
You will need the following packages:

```{r, eval = FALSE}
pak::pak("DSI")
pak::pak("DSOpal")
pak::pak("dsBaseClient")
```

###### Open connection
```{r, eval = TRUE}
# run some test commands with dsBaseClient
## needed to defined the OpalDriver class in the current environment
DSOpal::Opal()
## create new login object, note that here we use the `dsuser`
builder <- DSI::newDSLoginBuilder()
builder$append(server = "study1",
               url = SERVER,
               user = "dsuser",
               password = DSUSERPASS,
               driver = "OpalDriver")
logindata <- builder$build()
conns <- DSI::datashield.login(logins = logindata)
```

###### Simulate some operations
```{r, eval = FALSE}
## assign data
DSI::datashield.assign.table(conns["study1"], 
                             symbol = "dsROCrate_test",
                             table = paste0(PROJECT, ".", TABLES[1]),
                             errors.print = TRUE)

dsBaseClient::ds.ls(datasources = conns["study1"])
#> $study1
#> $study1$environment.searched
#> [1] "R_GlobalEnv"
#> 
#> $study1$objects.found
#> [1] "dsROCrate_test"
dsBaseClient::ds.summary("dsROCrate_test")
#> $study1
#> $study1$class
#> [1] "data.frame"
#> 
#> $study1$`number of rows`
#> [1] 2163
#> 
#> $study1$`number of columns`
#> [1] 11
#> 
#> $study1$`variables held`
#>  [1] "LAB_TSC"            "LAB_TRIG"           "LAB_HDL"           
#>  [4] "LAB_GLUC_ADJUSTED"  "PM_BMI_CONTINUOUS"  "DIS_CVA"           
#>  [7] "MEDI_LPD"           "DIS_DIAB"           "DIS_AMI"           
#> [10] "GENDER"             "PM_BMI_CATEGORICAL"
```

```{r, echo = FALSE, warning = FALSE, message = FALSE, results='hide'}
# check if there are any logs available, if not simulate some operations
dsuser_logs_tbl <- opalr::opal.login(
    username = "administrator",
    password = "password",
    url = SERVER
) |> 
  opalr::dsadmin.log() |>
  dplyr::bind_rows(tibble::tibble(username = NA, ds_action = NA)) |>
  dplyr::filter(!is.na(username), username == "dsuser") |>
  dplyr::filter(ds_action %in% c("ASSIGN", "AGGREGATE", "OPEN"))

if (nrow(dsuser_logs_tbl) < 1) {
  ## assign data
  DSI::datashield.assign.table(conns["study1"], 
                               symbol = "dsROCrate_test",
                               table = paste0(PROJECT, ".", TABLES[1]),
                               errors.print = TRUE)
  dsBaseClient::ds.ls(datasources = conns["study1"])
  dsBaseClient::ds.summary("dsROCrate_test")
}
```

---------

Then, we can proceed as per usual:

```{r safe_outputs_internal, echo=FALSE}
lines_before_safe_outputs <- rocrate_lines(basic_rocrate)
```

```{r safe_outputs}
basic_rocrate <- o |>
  dsROCrate::safe_output(rocrate = basic_rocrate,
                         logs_from = Sys.time() - 8.64E4, # capture the last 24 hours
                         logs_to = Sys.time())
```

```{r, out.lines=-(lines_before_safe_outputs + 6)}
print(basic_rocrate) # note that the output will be truncated
```

### 1.4. Close connection

```{r}
opalr::opal.logout(o)
```

### 1.5. Bag/Save RO-Crate

The resulting RO-Crate can be stored into an RO-Crate bag/archive with the
function `rocrateR::bag_rocrate`:

```{r}
# create temp directory (only for demo purposes)
tmp_path_bag <- file.path(tempdir(), "dsROCrate-getting-started")
dir.create(tmp_path_bag, showWarnings = FALSE)

# create RO-Crate bag
path_to_rocrate_bag <- basic_rocrate |>
  rocrateR::bag_rocrate(path = tmp_path_bag, overwrite = TRUE)
```

We can explore the contents with the following commands:

```{r}
# extract files in temporary directory
path_to_rocrate_bag |>
  # extract contents inside /tmp_path_bag/ROC
  rocrateR::unbag_rocrate(output = file.path(tmp_path_bag, "ROC"), quiet = TRUE) |>
  # create tree with the files
  fs::dir_tree()
```

### 1.6. Clean working directory

```{r}
unlink(tmp_path_bag, recursive = TRUE, force = TRUE)
```

<br />

```{r, echo = FALSE}
options(oopt)
oopt <- options(max_lines = 40)
```

## 2. Auditing RO-Crates and servers

### 2.1. Audit People

##### List accessible tables within a project for an user
```{r, warning=FALSE}
safe_people_crate_v1 <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
) |>
  dsROCrate::audit(user = "dsuser", project = "CNSIM")

print(safe_people_crate_v1)
```

###### Markdown report

A markdown report can be created with an overview and details for an RO-Crate,
using the `dsROCrate::report`:

**Only generate .Rmd file**

```{r safe_people_crate_audit_v1}
safe_people_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_people_crate_contents <- safe_people_crate_v1 |>
  dsROCrate::report(filepath = safe_people_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_people_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_people_crate_contents$overview_data |>
  knitr::kable()
```

**Render and display report (HTML)**

```{r, eval = FALSE}
safe_people_crate_v1 |>
  dsROCrate::report(filepath = safe_people_crate_v1_rmd,
                            title = "DataSHIELD Safe People - Audit Report",
                            render = TRUE, 
                            overwrite = TRUE)
```

### 2.2. Audit Project

##### List users and dataset/table level permissions within a project
```{r, warning=FALSE}
safe_project_crate_v1 <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
) |>
  dsROCrate::audit(project = "CNSIM")

print(safe_project_crate_v1)
```

###### Markdown report

A markdown report can be created with an overview and details for an RO-Crate,
using the `dsROCrate::report`:

**Only generate .Rmd file**

```{r safe_project_crate_audit_v1}
safe_project_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_project_crate_contents <- safe_project_crate_v1 |>
  dsROCrate::report(filepath = safe_project_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_project_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_project_crate_contents$overview_data |>
  knitr::kable()
```

**Render and display report (HTML)**

```{r, eval = FALSE}
safe_project_crate_v1 |>
  dsROCrate::report(filepath = safe_project_crate_v1_rmd, 
                            title = "DataSHIELD Safe Project - Audit Report",
                            render = TRUE, 
                            overwrite = TRUE)
```

<br />

### 2.3. Audit Study

##### List users and dataset/table level permissions within a study (i.e., multiple servers)
```{r, warning=FALSE}
study_crate_v1 <- 
  list(
    "opal_test" = opalr::opal.login(
      username = USERNAME,
      password = USERPASS,
      url = "https://opal-demo.obiba.org"
    ),
    "opal_demo" = opalr::opal.login(
      username = USERNAME,
      password = USERPASS,
      url = "https://opal-demo.obiba.org"
    )
  ) |>
  dsROCrate::audit(project = "CNSIM")

print(study_crate_v1)
```

###### Markdown report

A markdown report can be created with an overview and details for an RO-Crate,
using the `dsROCrate::report`:

**Only generate .Rmd file**

```{r study_crate_audit_v1}
study_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_project_crate_contents <- study_crate_v1 |>
  dsROCrate::report(filepath = study_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_project_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_project_crate_contents$overview_data |>
  knitr::kable()
```

**Render and display report (HTML)**

```{r, eval = FALSE}
study_crate_v1 |>
  dsROCrate::report(filepath = study_crate_v1_rmd, 
                            title = "DataSHIELD Study audit",
                            render = TRUE, 
                            overwrite = TRUE)
```

<br />

```{r, echo = FALSE, message = FALSE, error = FALSE}
unlink(safe_people_crate_v1_rmd, TRUE, TRUE)
unlink(safe_project_crate_v1_rmd, TRUE, TRUE)
unlink(study_crate_v1_rmd, TRUE, TRUE)
# reverse options
options(oopt)
```

---
title: "Getting Started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  cache = FALSE,
  out.width = "100%",
  cache = FALSE,
  screenshot.opts = list(vwidth = 2000, vheight = 600, zoom = 3, selector = "div.html-widget")
)

# save the built-in output hook
hook_output <- knitr::knit_hooks$get("output")
hook_message <- knitr::knit_hooks$get("message")

# define a truncation helper
truncate_lines_tail <- function(x, lines) {
  x <- unlist(strsplit(x, "\n"))
  more <- "..."
  if (length(lines) == 1) {
    if (length(x) > lines) {
      # truncate the output, but add `more`
      # x <- c(head(x, n = lines), more)
      x <- c(more, tail(x, n = lines))
    }
  } else {
    x <- c(more, x[lines], more)
  }
  paste(c(x, ""), collapse = "\n")
}

# set the new hooks
knitr::knit_hooks$set(
  output = function(x, options) {
    lines <- options$out.lines
    if (is.null(lines)) return(hook_output(x, options))
    x <- truncate_lines_tail(x, lines)
    return(hook_output(x, options))
  },
  message = function(x, options) {
    lines <- options$out.lines
    if (is.null(lines)) return(hook_message(x, options))
    x <- truncate_lines_tail(x, lines)
    return(hook_message(x, options))
  }
)

# helper function to get the number of rows of an RO-Crate
rocrate_lines <- function(rocrate) {
  rocrate_txt(rocrate) |>
    length()
}

# helper function to read RO-Crate as text file
rocrate_txt <- function(rocrate) {
  # create tmp file to write a JSON file with the output
  tmp_file <- tempfile(fileext = ".json")
  on.exit(unlink(tmp_file, force = TRUE))
  # write RO-Crate in tmp file
  rocrateR::write_rocrate(rocrate, tmp_file)
  # read RO-Crate as text file
  readLines(tmp_file)
}
```

```{r setup}
library(dsROCrate)

# show all the lines in the RO-Crate
oopt <- options(max_lines = Inf)
on.exit(options(oopt), add = TRUE)
```

This tutorial assumes that you have an internet connection and can access 
OBiBa's Opal demo server: https://opal-demo.obiba.org

Alternatively, if you want to test a local deployment, please check out the 
following vignette first:

```{r, eval = FALSE}
vignette("deploy-local-datashield-server-with-opal", package = "dsROCrate")
```

---------------------

## 1. Creating your first RO-Crate

### 1.1. Connect to an Opal server

Here we will use OBiBa's Opal demo server: https://opal-demo.obiba.org/ which
can be accessed with the following login credentials:

```{r}
# define global variables
## Opal server access
USERNAME <- "administrator"
USERPASS <- "password"
SERVER <- "https://opal-demo.obiba.org"
## Credentials for `dsuser`
### NOTE: this is only used to simulate an analysis and generate logs
DSUSERPASS <- "P@ssw0rd"
```

Next, define global variables used in generating the RO-Crate, such as project 
name, asset (e.g., tables, resources, etc.) references (within the project) and 
user identifiers.

```{r}
## Five safes variables
PEOPLE <- "dsuser"
PROJECT <- "CNSIM"
TABLES <- c("CNSIM1")
```

#### Open connection

Once the credentials and Five Safes variables are configured, we can start a 
new session on the Opal server with the following command:

```{r}
# login to local server with `USERNAME` and `USERPASS`.
o <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
)

print(o)
```

### 1.2. Create a basic RO-Crate

To create a basic RO-Crate, we will use the 
[`{rocrateR}`](https://github.com/ResearchObject/ro-crate-r) package. This 
package can be installed with the following command:

```{r, eval = FALSE, echo = TRUE}
# install.packages("pak")
pak::pak("rocrateR")

# for development version use
pak::pak("ResearchObject/ro-crate-r@dev")
```

Then, a basic RO-Crate can be created with the following command:

```{r}
basic_rocrate <- rocrateR::rocrate_5s()
```

Note that this RO-Crate uses the [5s-crate](https://trefx.uk/5s-crate/0.4/)
profile.

```{r}
print(basic_rocrate)
```

### 1.3. Add the _Five Safes_ Elements
#### Safe Data
To add details for Safe Data, use the function `dsROCrate::safe_data()`.

```{r, out.lines=-(rocrate_lines(basic_rocrate) - 2)}
basic_rocrate <- o |>
  dsROCrate::safe_data(rocrate = basic_rocrate,
                       project = PROJECT,
                       tables = TABLES)

print(basic_rocrate) # note that the output will be truncated
```

#### Safe Project
To add details for Safe Project, use the function `dsROCrate::safe_project()`.

```{r, out.lines=-(rocrate_lines(basic_rocrate) - 2)}
basic_rocrate <- o |>
  dsROCrate::safe_project(rocrate = basic_rocrate,
                          project = PROJECT)

print(basic_rocrate) # note that the output will be truncated
```

#### Safe People
To add details for Safe People, use the function `dsROCrate::safe_people()`.

```{r safe_people, out.lines=-(rocrate_lines(basic_rocrate) + 1)}
basic_rocrate <- o |>
  dsROCrate::safe_people(rocrate = basic_rocrate, user = PEOPLE)

print(basic_rocrate) # note that the output will be truncated
```

#### Safe Setting
To add details for Safe Setting, use the function `dsROCrate::safe_setting()`.

**⚠️NOTE:** The `dsROCrate::safe_setting` function requires administrator
privileges, so here, we will have to log in with administrator credentials (if
you used a non-administrator account previously).

```{r, eval = FALSE}
# close previous connection
opalr::opal.logout(o)

# open new connection as administrator
o <- opalr::opal.login(
  username = "administrator",
  password = "password",
  url = SERVER
)
```

Then, we can proceed as per usual:
```{r, out.lines=-(rocrate_lines(basic_rocrate) - 2)}
basic_rocrate <- o |>
  dsROCrate::safe_setting(rocrate = basic_rocrate)

print(basic_rocrate) # note that the output will be truncated
```

#### Safe Outputs
To add details for Safe Outputs, use the function `dsROCrate::safe_output()`.
Currently, only log files from the operations executed by the user within a 
specific period. Set the period using `logs_from` and `logs_to`. Additionally,
a list of functions executed by the user are extracted in a separate file/entity.

**⚠️NOTE:** Similar to `dsROCrate::safe_setting`, the `dsROCrate::safe_output` 
function requires of administrator rights, so here, we will have to log in with 
administrator credentials:

```{r, eval = FALSE}
# close previous connection
opalr::opal.logout(o)

# open new connection as administrator
o <- opalr::opal.login(
  username = "administrator",
  password = "password",
  url = SERVER
)
```

---------

##### DataSHIELD operations

**⚠️NOTE:** Before extracting logs, ensure there is recent activity on the 
server for testing purposes. This can be done using the following commands:

###### Setup
You will need the following packages:

```{r, eval = FALSE}
pak::pak("DSI")
pak::pak("DSOpal")
pak::pak("dsBaseClient")
```

###### Open connection
```{r, eval = TRUE}
# run some test commands with dsBaseClient
## needed to defined the OpalDriver class in the current environment
DSOpal::Opal()
## create new login object, note that here we use the `dsuser`
builder <- DSI::newDSLoginBuilder()
builder$append(server = "study1",
               url = SERVER,
               user = "dsuser",
               password = DSUSERPASS,
               driver = "OpalDriver")
logindata <- builder$build()
conns <- DSI::datashield.login(logins = logindata)
```

###### Simulate some operations
```{r, eval = TRUE}
## assign data
DSI::datashield.assign.table(conns["study1"], 
                             symbol = "dsROCrate_test",
                             table = paste0(PROJECT, ".", TABLES[1]),
                             errors.print = TRUE)

dsBaseClient::ds.ls(datasources = conns["study1"])
dsBaseClient::ds.summary("dsROCrate_test")
```

```{r, echo = FALSE, warning = FALSE, message = FALSE}
# check if there are any logs available, if not simulate some operations
```

---------

Then, we can proceed as per usual:

```{r safe_outputs_internal, echo=FALSE}
lines_before_safe_outputs <- rocrate_lines(basic_rocrate)
```

```{r safe_outputs}
basic_rocrate <- o |>
  dsROCrate::safe_output(rocrate = basic_rocrate,
                         logs_from = Sys.time() - 60, # capture the last minute
                         logs_to = Sys.time())
```

```{r, out.lines=-(lines_before_safe_outputs + 1)}
print(basic_rocrate) # note that the output will be truncated
```

### 1.4. Close connection

```{r}
opalr::opal.logout(o)
```

### 1.5. Bag/Save RO-Crate

The resulting RO-Crate can be stored into an RO-Crate bag/archive with the
function `rocrateR::bag_rocrate`:

```{r}
# create temp directory
tmp_path_bag <- file.path(tempdir(), "dsROCrate-getting-started")
dir.create(tmp_path_bag, showWarnings = FALSE)

# create RO-Crate bag
path_to_rocrate_bag <- basic_rocrate |>
  rocrateR::bag_rocrate(path = tmp_path_bag, overwrite = TRUE)
```

We can explore the contents with the following commands:

```{r}
# extract files in temporary directory
path_to_rocrate_bag |>
  # extract contents inside /tmp_path_bag/ROC
  rocrateR::unbag_rocrate(output = file.path(tmp_path_bag, "ROC"), quiet = TRUE) |>
  # create tree with the files
  fs::dir_tree()
```

### 1.6. Clean working directory

```{r}
unlink(tmp_path_bag, recursive = TRUE, force = TRUE)
```


<br />

## 2. Auditing RO-Crates and servers

### 2.1. Audit People

##### List accessible tables within a project for an user
```{r, warning=FALSE}
safe_people_crate_v1 <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
) |>
  dsROCrate::audit(user = "dsuser", project = "CNSIM")

print(safe_people_crate_v1)
```

###### Markdown report

A markdown report can be created with an overview and details for an RO-Crate,
using the `dsROCrate::report`:

**Only generate .Rmd file**

```{r safe_people_crate_audit_v1}
safe_people_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_people_crate_contents <- safe_people_crate_v1 |>
  dsROCrate::report(filepath = safe_people_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_people_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_people_crate_contents$overview_data |>
  knitr::kable()
```

**Render and display report (HTML)**

```{r, eval = FALSE}
safe_people_crate_v1 |>
  dsROCrate::report(filepath = safe_people_crate_v1_rmd,
                            title = "DataSHIELD Safe People - Audit Report",
                            render = TRUE, 
                            overwrite = TRUE)
```

### 2.2. Audit Project

##### List users and dataset/table level permissions within a project
```{r, warning=FALSE}
safe_project_crate_v1 <- opalr::opal.login(
  username = USERNAME,
  password = USERPASS,
  url = SERVER
) |>
  dsROCrate::audit(project = "CNSIM")

print(safe_project_crate_v1)
```

###### Markdown report

A markdown report can be created with an overview and details for an RO-Crate,
using the `dsROCrate::report`:

**Only generate .Rmd file**

```{r safe_project_crate_audit_v1}
safe_project_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_project_crate_contents <- safe_project_crate_v1 |>
  dsROCrate::report(filepath = safe_project_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_project_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_project_crate_contents$overview_data |>
  knitr::kable()
```

**Render and display report (HTML)**

```{r, eval = FALSE}
safe_project_crate_v1 |>
  dsROCrate::report(filepath = safe_project_crate_v1_rmd, 
                            title = "DataSHIELD Safe Project - Audit Report",
                            render = TRUE, 
                            overwrite = TRUE)
```

<br />

### 2.3. Audit Study

##### List users and dataset/table level permissions within a study (i.e., multiple servers)
```{r, warning=FALSE}
study_crate_v1 <- 
  list(
    "opal_test" = opalr::opal.login(
      username = USERNAME,
      password = USERPASS,
      url = "https://opal-test.obiba.org"
    ),
    "opal_demo" = opalr::opal.login(
      username = USERNAME,
      password = USERPASS,
      url = "https://opal-demo.obiba.org"
    )
  ) |>
  dsROCrate::audit(project = "CNSIM")

print(study_crate_v1)
```

###### Markdown report

A markdown report can be created with an overview and details for an RO-Crate,
using the `dsROCrate::report`:

**Only generate .Rmd file**

```{r study_crate_audit_v1}
study_crate_v1_rmd <- tempfile(fileext = ".Rmd") # temporary file

safe_project_crate_contents <- study_crate_v1 |>
  dsROCrate::report(filepath = study_crate_v1_rmd, render = FALSE)

# display Overview diagram
safe_project_crate_contents$overview_diagram

# display Overview data (Safe People, Safe Projects and Safe Data)
safe_project_crate_contents$overview_data |>
  knitr::kable()
```

**Render and display report (HTML)**

```{r, eval = FALSE}
study_crate_v1 |>
  dsROCrate::report(filepath = study_crate_v1_rmd, 
                            title = "DataSHIELD Study audit",
                            render = TRUE, 
                            overwrite = TRUE)
```

<br />

```{r, echo = FALSE, message = FALSE, error = FALSE}
unlink(safe_people_crate_v1_rmd, TRUE, TRUE)
unlink(safe_project_crate_v1_rmd, TRUE, TRUE)
unlink(study_crate_v1_rmd, TRUE, TRUE)
# reverse options
options(oopt)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-safe_data.R
\name{flatten_safe_data}
\alias{flatten_safe_data}
\alias{flatten_safe_data.default}
\alias{flatten_safe_data.rocrate}
\title{Flatten object with Safe Data details}
\usage{
flatten_safe_data(x, ...)

\method{flatten_safe_data}{default}(x, ...)

\method{flatten_safe_data}{rocrate}(x, ..., id = NULL, asset_id_suffix = "#asset:")
}
\arguments{
\item{x}{Object (e.g., RO-Crate) with Safe Data details. This can be
generated with the \code{\link[=extract_safe_data]{extract_safe_data()}} function.}

\item{id}{Vector of strings with the \verb{@id}s for the datasets to be extracted.
If not provided, extract all entities with \verb{@type = 'Dataset'}.}
}
\value{
Data frame with Safe Data details.
}
\description{
Flatten object with Safe Data details
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-safe_setting.R
\name{extract_safe_setting}
\alias{extract_safe_setting}
\alias{extract_safe_setting.opal}
\alias{extract_safe_setting.rocrate}
\title{Extract Safe Setting entity(ies)}
\usage{
extract_safe_setting(x, ...)

\method{extract_safe_setting}{opal}(x, ..., rocrate = rocrateR::rocrate_5s())

\method{extract_safe_setting}{rocrate}(x, ..., id = NULL, rocrate = rocrateR::rocrate_5s())
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}), an RO-Crate
(\link[rocrateR:rocrate]{rocrate} class) or a string with the path to an
RO-Crate.}

\item{...}{Other optional arguments. See the full documentation}

\item{rocrate}{(Optional) RO-Crate object to update with Safe Setting details.}

\item{id}{(Optional) Vector with \verb{@id} strings for Safe Setting entity(ies)
to be extracted from the given RO-Crate, \code{x}.}
}
\value{
RO-Crate with Safe Setting entity(ies).
}
\description{
Extract Safe Setting entity(ies)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-opal.R
\name{get_table_permissions}
\alias{get_table_permissions}
\title{Get table permissions}
\usage{
get_table_permissions(x, project, tables)
}
\arguments{
\item{x}{Connection to OBiBa's Opal server (see \code{\link[opalr:opal.login]{opalr::opal.login()}}).}

\item{project}{String with project name.}

\item{tables}{String (or vector of strings) with table names for the given
project.}
}
\value{
Data frame with permissions for each table in \code{tables}.
}
\description{
Wrapper for the \code{\link[opalr:opal.table_perm]{opalr::opal.table_perm()}} function.
}
\seealso{
Other Opal: 
\code{\link{get_project_details}()},
\code{\link{get_project_tables}()},
\code{\link{parse_user_profiles}()},
\code{\link{project_exists}()}
}
\concept{Opal}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/safe_project.R
\name{safe_project}
\alias{safe_project}
\alias{safe_project,armadillo-method}
\alias{safe_project.character}
\alias{safe_project.opal}
\alias{safe_project.rocrate}
\alias{safe_project.ArmadilloCredentials}
\title{Safe Project details}
\source{
\itemize{
\item Research Data Scotland, 2025. "What is the Five Safes framework?".
\url{https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/}
}
}
\usage{
\S4method{safe_project}{armadillo}(
  x,
  ...,
  profile = "default",
  project = NULL,
  rocrate = rocrateR::rocrate_5s(),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  path = NULL,
  resources = NULL,
  tables = NULL,
  user = NULL
)

\method{safe_project}{character}(
  x,
  ...,
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  user = attr(x, "user")
)

\method{safe_project}{opal}(
  x,
  ...,
  profile = "default",
  project = NULL,
  rocrate = rocrateR::rocrate_5s(),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  path = NULL,
  resources = NULL,
  tables = NULL,
  user = NULL
)

\method{safe_project}{rocrate}(
  x,
  ...,
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  user = attr(x, "user")
)

\method{safe_project}{ArmadilloCredentials}(
  x,
  ...,
  profile = "default",
  project = NULL,
  rocrate = rocrateR::rocrate_5s(),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  path = NULL,
  resources = NULL,
  tables = NULL,
  user = NULL
)
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}), an RO-Crate
(\link[rocrateR:rocrate]{rocrate} class) or a string with the path to an
RO-Crate.}

\item{...}{Other optional arguments. See the full documentation,
\code{\link[=safe_project]{?dsROCrate::safe_project}}.}

\item{profile}{String with profile name (used for OBiBa's Opal backend).}

\item{project}{String with the name of the \link[=safe_project]{Safe Project}.}

\item{asset_id_suffix}{String with ID suffix for the tables/datasets
entities in the RO-Crate (default: \code{"#asset:"}).}

\item{project_id_suffix}{String with ID suffix for the project entities
in the RO-Crate (default: \code{"#project:"}).}

\item{connection}{Connection object for the 'DataSHIELD' server where the
values will be extracted from (e.g., OBiBa's Opal). Optional, if \code{x} is
set to a connection object. If so, then \code{rocrate} is required.}

\item{path}{String with path pointing to the root of the RO-Crate. This will
be used to store log files. If not provided, logs will be stored within
the RO-Crate returned by this function.}

\item{resources}{Vector of strings with the names of the resources, part of
\code{project}. Optional, if not provided, all the resources associated to
\code{project} will be included in the RO-Crate.}

\item{tables}{Vector of strings with the names of the tables/datasets, part
of \code{project}. Optional, if not provided, all the tables/datasets
associated to \code{project} will be included in the RO-Crate.}

\item{user}{List (or \link[rocrateR:entity]{entity} object) with details for
the Safe People, it must include \verb{@id} and \code{name} entries. Alternatively,
this can be a string with the \code{name} of the current user.}

\item{rocrate}{RO-Crate object. Optional, if \code{x} is either an RO-Crate
object or a path to a valid RO-Crate. If so, then \code{connection} is
required (default: \code{rocrateR::rocrate_5s()}).}
}
\value{
Updated RO-Crate object with Safe Project information.
}
\description{
Safe Project details for the RO-Crate.
}
\details{
Data must be used ethically, for research that delivers clear public benefit.

As part of their application, researchers are asked to provide an overview
of their project, including how the data will be used and what outputs will
be achieved. This allows data providers to make an informed decision about
whether they are comfortable preparing data for the researcher to use for
ethical purposes serving a public good.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-safe_output.R
\name{extract_safe_output}
\alias{extract_safe_output}
\alias{extract_safe_output.opal}
\alias{extract_safe_output.rocrate}
\title{Extract Safe Output entity(ies)}
\usage{
extract_safe_output(x, ...)

\method{extract_safe_output}{opal}(
  x,
  ...,
  path = NULL,
  user = NULL,
  logs_to = Sys.time(),
  logs_from = logs_to - 24 * 60^2,
  rocrate = rocrateR::rocrate_5s()
)

\method{extract_safe_output}{rocrate}(
  x,
  ...,
  id = NULL,
  user = NULL,
  rocrate = rocrateR::rocrate_5s()
)
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}), an RO-Crate
(\link[rocrateR:rocrate]{rocrate} class) or a string with the path to an
RO-Crate.}

\item{...}{Other optional arguments. See the full documentation}

\item{path}{String with path pointing to the root of the RO-Crate. This will
be used to store log files. If not provided, logs will be stored within
the RO-Crate returned by this function.}

\item{user}{List (or \link[rocrateR:entity]{entity} object) with details for
the Safe People, it must include \verb{@id} and \code{name} entries. Alternatively,
this can be a string with the \code{name} of the current user.}

\item{logs_to}{Upper limit timestamp to filter out the outputs generated
(default: \code{Sys.time()}, current system time).}

\item{logs_from}{Lower limit timestamp to filter out the outputs generated
(default: \code{Sys.time() - 24 * 60 ^ 2}, last 24 hours).}

\item{rocrate}{(Optional) RO-Crate object to update with Safe Output details.}

\item{id}{(Optional) Vector with \verb{@id} strings for Safe Output entity(ies)
to be extracted from the given RO-Crate, \code{x}.}
}
\value{
RO-Crate with Safe Output entity(ies).
}
\description{
Extract Safe Output entity(ies)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-safe_data.R
\name{extract_safe_data}
\alias{extract_safe_data}
\alias{extract_safe_data.opal}
\alias{extract_safe_data.rocrate}
\title{Extract Safe Data entity(ies)}
\usage{
extract_safe_data(x, ...)

\method{extract_safe_data}{opal}(x, ..., rocrate = rocrateR::rocrate_5s())

\method{extract_safe_data}{rocrate}(
  x,
  ...,
  id = NULL,
  asset_id_suffix = "#asset:",
  rocrate = rocrateR::rocrate_5s()
)
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}), an RO-Crate
(\link[rocrateR:rocrate]{rocrate} class) or a string with the path to an
RO-Crate.}

\item{...}{Other optional arguments. See the full documentation}

\item{rocrate}{(Optional) RO-Crate object to update with Safe Data details.}

\item{id}{(Optional) Vector with \verb{@id} strings for Safe Data entity(ies)
to be extracted from the given RO-Crate, \code{x}.}

\item{asset_id_suffix}{String with ID suffix for the tables/datasets
entities in the RO-Crate (default: \code{"#asset:"}).}
}
\value{
RO-Crate with Safe Data entity(ies).
}
\description{
Extract Safe Data entity(ies)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dsROCrate-package.R
\docType{package}
\name{dsROCrate-package}
\alias{dsROCrate}
\alias{dsROCrate-package}
\title{dsROCrate: 'DataSHIELD' RO-Crate Governance Functions}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Tools for wrapping 'DataSHIELD' analyses into RO-Crate (Research Object Crate) objects. Provides functions to create structured metadata for federated data analysis projects, enabling governance tracking of data access, project membership, analysis execution and output validation across distributed data sources.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/FederatedMethods/dsROCrate}
  \item Report bugs at \url{https://github.com/FederatedMethods/dsROCrate/issues}
}

}
\author{
\strong{Maintainer}: Roberto Villegas-Diaz \email{r.villegas-diaz@outlook.com} (\href{https://orcid.org/0000-0001-5036-8661}{ORCID})

Authors:
\itemize{
  \item Becca Wilson (\href{https://orcid.org/0000-0003-2294-593X}{ORCID})
  \item Olly Butters (\href{https://orcid.org/0000-0003-0354-8461}{ORCID})
  \item Stuart Wheater (\href{https://orcid.org/0009-0003-2419-1964}{ORCID})
}

Other contributors:
\itemize{
  \item University of Liverpool [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-tibble.R
\name{refill_vec}
\alias{refill_vec}
\title{Fill vector}
\usage{
refill_vec(x, val = "")
}
\arguments{
\item{x}{Input vector.}

\item{val}{Default value.}
}
\value{
Filled vector.
}
\description{
Fill vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-opal.R
\name{user_perm_entity}
\alias{user_perm_entity}
\title{Create user/person permission entities}
\usage{
user_perm_entity(person, person_id, asset, asset_id, permission, ...)
}
\arguments{
\item{person}{String with person name/username.}

\item{person_id}{String with person \verb{@id}.}

\item{asset}{String with dataset/table/resource name.}

\item{asset_id}{String with dataset/table \verb{@id}.}

\item{permission}{String with permission ('view', 'view-values', 'edit',
'edit-values' OR 'administrate').}

\item{...}{Other additional values.}
}
\value{
List of \link[rocrateR:entity]{rocrateR::entity} objects
}
\description{
Create user/person permission entities
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-opal.R
\name{flatten_user_perm_entity}
\alias{flatten_user_perm_entity}
\title{Flatten user permission entities}
\usage{
flatten_user_perm_entity(x)
}
\arguments{
\item{x}{List with entities, generated by \code{\link[=user_perm_entity]{user_perm_entity()}}.}
}
\value{
Tibble with properties of the entities.
}
\description{
Flatten user permission entities
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-connection.R
\name{parse_user_profiles}
\alias{parse_user_profiles}
\alias{parse_user_profiles,armadillo-method}
\alias{parse_user_profiles.opal}
\alias{parse_user_profiles.ArmadilloCredentials}
\title{Parse user profiles}
\usage{
\S4method{parse_user_profiles}{armadillo}(x, ..., user)

\method{parse_user_profiles}{opal}(x, ..., user)

\method{parse_user_profiles}{ArmadilloCredentials}(x, ..., user)
}
\arguments{
\item{x}{Connection object to backend for DataSHIELD server (e.g., Opal).}

\item{...}{Optional arguments, unused.}
}
\value{
Data frame with given \code{user}'s profile details, as captured on the
server pointed by \code{x}.
}
\description{
Parse user profiles
}
\seealso{
Other Armadillo: 
\code{\link{project_exists}()}

Other Opal: 
\code{\link{get_project_details}()},
\code{\link{get_project_tables}()},
\code{\link{get_table_permissions}()},
\code{\link{project_exists}()}
}
\concept{Armadillo}
\concept{Opal}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-opal.R
\name{get_project_tables}
\alias{get_project_tables}
\title{Get project tables}
\usage{
get_project_tables(x, project)
}
\arguments{
\item{x}{Connection to OBiBa's Opal server (see \code{\link[opalr:opal.login]{opalr::opal.login()}}).}

\item{project}{String with project name.}
}
\value{
List of project tables
}
\description{
Wrapper for \code{\link[opalr:opal.project]{opalr::opal.project()}}.
}
\seealso{
Other Opal: 
\code{\link{get_project_details}()},
\code{\link{get_table_permissions}()},
\code{\link{parse_user_profiles}()},
\code{\link{project_exists}()}
}
\concept{Opal}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-opal.R
\name{is_opal_admin_con}
\alias{is_opal_admin_con}
\title{Verify if connection was created by an administrative user}
\usage{
is_opal_admin_con(x)
}
\arguments{
\item{x}{Connection to OBiBa's Opal server (see \code{\link[opalr:opal.login]{opalr::opal.login()}}).}
}
\value{
Boolean flag to indicate whether the given connection was created
by an administrative user.
}
\description{
Verify if connection was created by an administrative user
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/safe_data.R
\name{safe_data}
\alias{safe_data}
\alias{safe_data.character}
\alias{safe_data.opal}
\alias{safe_data.rocrate}
\title{Safe Data details}
\source{
\itemize{
\item Research Data Scotland, 2025. "What is the Five Safes framework?".
\url{https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/}
}
}
\usage{
safe_data(x, ...)

\method{safe_data}{character}(
  x,
  ...,
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  user = attr(x, "user")
)

\method{safe_data}{opal}(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL,
  include = c("tables", "resources"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  path = NULL,
  user = NULL
)

\method{safe_data}{rocrate}(
  x,
  ...,
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  asset_id_suffix = "#asset:",
  project_id_suffix = "#project:",
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  user = attr(x, "user")
)
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}), an RO-Crate
(\link[rocrateR:rocrate]{rocrate} class) or a string with the path to an
RO-Crate.}

\item{...}{Other optional arguments. See the full documentation,
\code{\link[=safe_data]{?dsROCrate::safe_data}}.}

\item{profile}{String with profile name (used for OBiBa's Opal backend).}

\item{project}{String with the name of the \link[=safe_project]{Safe Project}.}

\item{resources}{Vector of strings with the names of the resources, part of
\code{project}. Optional, if not provided, all the resources associated to
\code{project} will be included in the RO-Crate.}

\item{tables}{Vector of strings with the names of the tables/datasets, part
of \code{project}. Optional, if not provided, all the tables/datasets
associated to \code{project} will be included in the RO-Crate.}

\item{asset_id_suffix}{String with ID suffix for the tables/datasets
entities in the RO-Crate (default: \code{"#asset:"}).}

\item{project_id_suffix}{String with ID suffix for the project entities
in the RO-Crate (default: \code{"#project:"}).}

\item{connection}{Connection object for the 'DataSHIELD' server where the
values will be extracted from (e.g., OBiBa's Opal). Optional, if \code{x} is
set to a connection object. If so, then \code{rocrate} is required.}

\item{path}{String with path pointing to the root of the RO-Crate. This will
be used to store log files. If not provided, logs will be stored within
the RO-Crate returned by this function.}

\item{user}{List (or \link[rocrateR:entity]{entity} object) with details for
the Safe People, it must include \verb{@id} and \code{name} entries. Alternatively,
this can be a string with the \code{name} of the current user.}

\item{rocrate}{RO-Crate object. Optional, if \code{x} is either an RO-Crate
object or a path to a valid RO-Crate. If so, then \code{connection} is
required (default: \code{rocrateR::rocrate_5s()}).}

\item{include}{Vector of strings with types of assets to be included, either
\code{"resources"}, \code{"tables"} or both.}
}
\value{
Updated RO-Crate object with Safe Data information.
}
\description{
Safe Data details for the RO-Crate.
}
\details{
Researchers only use de-identified data that is relevant to their study.

In compliance with the Digital Economy Act, data is effectively anonymised
within TREs (Trusted Research Environments).

This means any sensitive information that might lead to an individual being
identified, such as names and addresses, is either removed or replaced with
a random code. Researchers are not processing personal data when using data
prepared in this way and when the other Safes are in place. Find out more
about de-identification:
\url{https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-data-de-identification/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/safe_setting.R
\name{safe_setting}
\alias{safe_setting}
\alias{safe_setting.character}
\alias{safe_setting.cr8tor}
\alias{safe_setting.opal}
\alias{safe_setting.rocrate}
\title{Safe Setting details}
\source{
\itemize{
\item Research Data Scotland, 2025. "What is the Five Safes framework?".
\url{https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/}
}
}
\usage{
safe_setting(x, ...)

\method{safe_setting}{character}(
  x,
  ...,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  tables = attr(x, "tables"),
  resources = attr(x, "resources"),
  user = attr(x, "user")
)

\method{safe_setting}{cr8tor}(x, ..., rocrate = rocrateR::rocrate_5s())

\method{safe_setting}{opal}(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  path = NULL,
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL,
  user = NULL
)

\method{safe_setting}{rocrate}(
  x,
  ...,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  user = attr(x, "user")
)
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}), an RO-Crate
(\link[rocrateR:rocrate]{rocrate} class) or a string with the path to an
RO-Crate.}

\item{...}{Other optional arguments. See the full documentation,
\code{\link[=safe_setting]{?dsROCrate::safe_setting}}.}

\item{connection}{Connection object for the 'DataSHIELD' server where the
values will be extracted from (e.g., OBiBa's Opal). Optional, if \code{x} is
set to a connection object. If so, then \code{rocrate} is required.}

\item{path}{String with path pointing to the root of the RO-Crate. This will
be used to store log files. If not provided, logs will be stored within
the RO-Crate returned by this function.}

\item{profile}{String with profile name (used for OBiBa's Opal backend).}

\item{project}{String with the name of the \link[=safe_project]{Safe Project}.}

\item{tables}{Vector of strings with the names of the tables/datasets, part
of \code{project}. Optional, if not provided, all the tables/datasets
associated to \code{project} will be included in the RO-Crate.}

\item{resources}{Vector of strings with the names of the resources, part of
\code{project}. Optional, if not provided, all the resources associated to
\code{project} will be included in the RO-Crate.}

\item{user}{List (or \link[rocrateR:entity]{entity} object) with details for
the Safe People, it must include \verb{@id} and \code{name} entries. Alternatively,
this can be a string with the \code{name} of the current user.}

\item{rocrate}{RO-Crate object. Optional, if \code{x} is either an RO-Crate
object or a path to a valid RO-Crate. If so, then \code{connection} is
required (default: \code{rocrateR::rocrate_5s()}).}
}
\value{
Updated RO-Crate object with Safe Settings information.
}
\description{
Safe Setting details for the RO-Crate.
}
\details{
The organisational and technical settings used to access data are designed
to minimise the risk of accidental disclosure of data.

These settings also prevent the deliberate disclosure of data to others.

Physical settings for data access can include locations like
\href{https://safepodnetwork.ac.uk/}{SafePods} – secured rooms that use controlled
door access, CCTV and secure technology to ensure that sensitive data cannot
be mishandled or removed from the Safe Setting. Researchers can analyse the
data in these secure rooms, but do not have access to the internet, external
devices (such as printers), or any other way of removing protected data from
the space.

Digital Safe Settings provide secure access to data from a remote location.
To be approved for remote data access, researchers will need to prove that
their organisation meets physical and IT security standards.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-opal.R
\name{update_project_datasets}
\alias{update_project_datasets}
\title{Update datasets linked to a project}
\usage{
update_project_datasets(rocrate, project, ds_ids)
}
\arguments{
\item{rocrate}{RO-Crate object. Optional, if \code{x} is either an RO-Crate
object or a path to a valid RO-Crate. If so, then \code{connection} is
required (default: \code{rocrateR::rocrate_5s()}).}

\item{project}{String with the name of the \link[=safe_project]{Safe Project}.}

\item{ds_ids}{Vector with \verb{@id}s of the datasets to be linked to \code{project}}
}
\value{
Update RO-Crate with updated project entity
}
\description{
Update datasets linked to a project (\code{hasPart})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit.R
\name{audit}
\alias{audit}
\alias{audit.ArmadilloCredentials}
\alias{audit.character}
\alias{audit.cr8tor}
\alias{audit.list}
\alias{audit.opal}
\alias{audit.rocrate}
\title{Create an audit RO-Crate}
\usage{
audit(x, ...)

\method{audit}{ArmadilloCredentials}(x, ..., intent = NULL)

\method{audit}{character}(x, ..., intent = NULL)

\method{audit}{cr8tor}(x, ..., intent = NULL)

\method{audit}{list}(x, ..., intent = NULL)

\method{audit}{opal}(
  x,
  ...,
  intent = NULL,
  project = NULL,
  user = NULL,
  logs_from = -Inf,
  logs_to = Inf,
  path = NULL
)

\method{audit}{rocrate}(x, ..., intent = NULL)
}
\arguments{
\item{x}{Object to be audited. This can be
\itemize{
\item a \emph{connection} to a DataSHIELD server (e.g., OBiBa's Opal).
\item a \emph{path} pointing to an RO-Crate OR a \code{cr8tor} archive /
.        governance bundle.
\item an \emph{RO-Crate} object.
}
Alternatively, a list of any of the above.}

\item{...}{Additional arguments.}

\item{intent}{Additional object with governance bundle/specification of the
intent of a project. It takes the same types as \code{x}.}

\item{project}{String with project name(s) from which to extra Safe Project
details.}

\item{user}{String with the user name for which to extract Safe People
details.}

\item{logs_from}{Lower limit timestamp to filter out the outputs generated
(default: \code{-Inf}, everything up to \code{logs_to})}

\item{logs_to}{Upper limit timestamp to filter out the outputs generated
(default: \code{Inf}, everything from \code{logs_from} onwards).}

\item{path}{String with path pointing to the root of the RO-Crate. This will
be used to store log files. If not provided, logs will be stored within
the RO-Crate returned by this function.}
}
\value{
RO-Crate with audit details.
}
\description{
Create an audit RO-Crate following the 5 Safes Principles.
}
\details{
This function handles various audit types, which will be dispatched based on
the input object. If the input object is

\itemize{
\item a \emph{connection} to a DataSHIELD server (e.g., OBiBa's Opal):
.    generates an RO-Crate object with deployment details, including outputs.
\item a \emph{path} pointing to
\itemize{
\item \strong{a \code{cr8tor} archive / governance bundle}: generates an
RO-Crate object with pre-deployment governance details.
\item \strong{an RO-Crate object}: generates an RO-Crate object with
clearly defined 5 Safes elements.
}
\item an \emph{RO-Crate} object: generates an RO-Crate object with clearly
defined 5 Safes elements.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-opal.R
\name{validate_opal_con}
\alias{validate_opal_con}
\title{Validate OBiBa's Opal connection}
\usage{
validate_opal_con(x)
}
\arguments{
\item{x}{Connection to OBiBa's Opal server (see \code{\link[opalr:opal.login]{opalr::opal.login()}}).}
}
\value{
Nothing, call for its side effect.
}
\description{
Validate OBiBa's Opal connection
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-date.R
\name{is_valid_posixct}
\alias{is_valid_posixct}
\title{Validate POSIXct string}
\usage{
is_valid_posixct(x, tz = "UTC")
}
\arguments{
\item{x}{POSIXct string.}

\item{tz}{String with time zone,}
}
\value{
Boolean value to indicate if the given string is a valid POSIXct
string.
}
\description{
Validate POSIXct string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-rocrateR.R
\name{append_entity_ref}
\alias{append_entity_ref}
\title{Append entity reference}
\usage{
append_entity_ref(rocrate, id, key, ref_id)
}
\arguments{
\item{rocrate}{RO-Crate object.}

\item{id}{String with \verb{@id} of the entity.}

\item{key}{String with property/key of the entity.}

\item{ref_id}{String with reference \verb{@id}.}
}
\value{
Updated RO-Crate object.
}
\description{
Adds a reference to an existing property while
preserving existing values and avoiding duplicates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-armadillo.R
\name{armadillo_login}
\alias{armadillo_login}
\title{Login to a MOLGENIS' Armadillo server}
\usage{
armadillo_login(server)
}
\arguments{
\item{server}{the URL of the Armadillo server}
}
\value{
MOLGENIS' Armadillo connection object.
}
\description{
Login to a MOLGENIS' Armadillo server. Wrapper for the function
\code{\link[DSMolgenisArmadillo:armadillo.get_credentials]{DSMolgenisArmadillo::armadillo.get_credentials()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dsROCrate.R
\name{init}
\alias{init}
\alias{init.opal}
\alias{init.rocrate}
\title{Initialise a Five Safes RO-Crate}
\usage{
init(x, ...)

\method{init}{opal}(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL,
  path = NULL,
  user = NULL
)

\method{init}{rocrate}(
  x,
  ...,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables"),
  user = attr(x, "user")
)
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}), an RO-Crate
(\link[rocrateR:rocrate]{rocrate} class) or a string with the path to an
RO-Crate.}

\item{...}{Other optional arguments. See the full documentation,
\code{\link[=init]{?dsROCrate::init}}.}

\item{rocrate}{RO-Crate object. Optional, if \code{x} is either an RO-Crate
object or a path to a valid RO-Crate. If so, then \code{connection} is
required (default: \code{rocrateR::rocrate_5s()}).}

\item{profile}{String with profile name (used for OBiBa's Opal backend).}

\item{project}{String with the name of the \link[=safe_project]{Safe Project}.}

\item{resources}{Vector of strings with the names of the resources, part of
\code{project}. Optional, if not provided, all the resources associated to
\code{project} will be included in the RO-Crate.}

\item{tables}{Vector of strings with the names of the tables/datasets, part
of \code{project}. Optional, if not provided, all the tables/datasets
associated to \code{project} will be included in the RO-Crate.}

\item{path}{String with path pointing to the root of the RO-Crate. This will
be used to store log files. If not provided, logs will be stored within
the RO-Crate returned by this function.}

\item{user}{List (or \link[rocrateR:entity]{entity} object) with details for
the Safe People, it must include \verb{@id} and \code{name} entries. Alternatively,
this can be a string with the \code{name} of the current user.}

\item{connection}{Connection object for the 'DataSHIELD' server where the
values will be extracted from (e.g., OBiBa's Opal). Optional, if \code{x} is
set to a connection object. If so, then \code{rocrate} is required.}
}
\value{
Five Safes RO-Crate object.
}
\description{
Creates a new RO-Crate configured for Five Safes auditing.
}
\references{
Wilkinson, M., Dumontier, M., Aalbersberg, I. et al. (2016) The FAIR Guiding
Principles for scientific data management and stewardship. Sci Data 3,
160018. https://doi.org/10.1038/sdata.2016.18
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/safe_output.R
\name{safe_output}
\alias{safe_output}
\alias{safe_output.character}
\alias{safe_output.opal}
\alias{safe_output.rocrate}
\title{Safe Output details}
\source{
\itemize{
\item Research Data Scotland, 2025. "What is the Five Safes framework?".
\url{https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/}
}
}
\usage{
safe_output(x, ...)

\method{safe_output}{character}(
  x,
  ...,
  path = attr(x, "path"),
  user = attr(x, "user"),
  logs_to = Sys.time(),
  logs_from = logs_to - 24 * 60^2,
  connection = attr(x, "connection"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables")
)

\method{safe_output}{opal}(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  path = NULL,
  user = NULL,
  logs_to = Sys.time(),
  logs_from = logs_to - 24 * 60^2,
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL
)

\method{safe_output}{rocrate}(
  x,
  ...,
  path = attr(x, "path"),
  user = attr(x, "user"),
  logs_to = Sys.time(),
  logs_from = logs_to - 24 * 60^2,
  connection = attr(x, "connection"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables")
)
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}), an RO-Crate
(\link[rocrateR:rocrate]{rocrate} class) or a string with the path to an
RO-Crate.}

\item{...}{Other optional arguments. See the full documentation,
\code{\link[=safe_output]{?dsROCrate::safe_output}}.}

\item{path}{String with path pointing to the root of the RO-Crate. This will
be used to store log files. If not provided, logs will be stored within
the RO-Crate returned by this function.}

\item{user}{List (or \link[rocrateR:entity]{entity} object) with details for
the Safe People, it must include \verb{@id} and \code{name} entries. Alternatively,
this can be a string with the \code{name} of the current user.}

\item{logs_to}{Upper limit timestamp to filter out the outputs generated
(default: \code{Sys.time()}, current system time).}

\item{logs_from}{Lower limit timestamp to filter out the outputs generated
(default: \code{Sys.time() - 24 * 60 ^ 2}, last 24 hours).}

\item{connection}{Connection object for the 'DataSHIELD' server where the
values will be extracted from (e.g., OBiBa's Opal). Optional, if \code{x} is
set to a connection object. If so, then \code{rocrate} is required.}

\item{profile}{String with profile name (used for OBiBa's Opal backend).}

\item{project}{String with the name of the \link[=safe_project]{Safe Project}.}

\item{resources}{Vector of strings with the names of the resources, part of
\code{project}. Optional, if not provided, all the resources associated to
\code{project} will be included in the RO-Crate.}

\item{tables}{Vector of strings with the names of the tables/datasets, part
of \code{project}. Optional, if not provided, all the tables/datasets
associated to \code{project} will be included in the RO-Crate.}

\item{rocrate}{RO-Crate object. Optional, if \code{x} is either an RO-Crate
object or a path to a valid RO-Crate. If so, then \code{connection} is
required (default: \code{rocrateR::rocrate_5s()}).}
}
\value{
Updated RO-Crate object with Safe Outputs information.
}
\description{
Safe Output details for the RO-Crate.
}
\details{
All research outputs are checked to ensure individuals cannot be identified
even in the public domain.

This means the other four Safes no longer apply.

Before outputs are released from the TRE, they are checked to make sure it
is reasonably unlikely that any individuals can be identified on publication.
Compliance with the Digital Economy Act also requires that the TRE applies
methods and standards for output checking that are accredited by the UK
Statistics Authority.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-cr8tor.R
\name{load_cr8tor_bundle}
\alias{load_cr8tor_bundle}
\title{Load cr8tor governance bundle / project archive}
\usage{
load_cr8tor_bundle(x, ...)
}
\arguments{
\item{x}{Path to cr8tor ZIP archive.}

\item{...}{Unused.}
}
\value{
Object of class \code{cr8tor}.
}
\description{
A cr8tor project archive contains:
\itemize{
\item bagit/      → RO-Crate metadata layer.
\item resources/  → deployment & governance YAML specs.
\item config.toml → platform configuration.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-safe_setting.R
\name{flatten_safe_setting}
\alias{flatten_safe_setting}
\alias{flatten_safe_setting.default}
\alias{flatten_safe_setting.rocrate}
\title{Flatten object with Safe Setting details}
\usage{
flatten_safe_setting(x, ...)

\method{flatten_safe_setting}{default}(x, ...)

\method{flatten_safe_setting}{rocrate}(x, ..., id = NULL)
}
\arguments{
\item{x}{Object (e.g., RO-Crate) with Safe Setting details. This can be
generated with the \code{\link[=extract_safe_setting]{extract_safe_setting()}} function.}

\item{id}{Vector of strings with the \verb{@id}s for the settings to be extracted.
If not provided, extract all entities with \verb{@type = 'PropertyValue'} or
\verb{@type = 'SoftwareApplication'}.}
}
\value{
Data frame with Safe Settings.
}
\description{
Flatten object with Safe Setting details
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-tibble.R
\name{unfill_vec}
\alias{unfill_vec}
\title{Unfill vector}
\source{
https://github.com/tidyverse/tidyr/issues/250#issuecomment-344984802
}
\usage{
unfill_vec(x, val = "")
}
\arguments{
\item{x}{Input vector.}

\item{val}{Default value.}
}
\value{
Unfilled vector.
}
\description{
Unfill vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit_engine.R
\name{audit_engine}
\alias{audit_engine}
\alias{audit_engine.cr8tor}
\alias{audit_engine.opal}
\title{Audit Engine}
\usage{
audit_engine(x, ...)

\method{audit_engine}{cr8tor}(x, ...)

\method{audit_engine}{opal}(
  x,
  ...,
  project = NULL,
  user = NULL,
  logs_from = -Inf,
  logs_to = Inf,
  path = NULL
)
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}). Alternatively, a governance
archive file, representing the intent of a project and associated
governance details.}

\item{...}{Other optional arguments, see full documentation for details.}

\item{project}{String with project name(s) from which to extra Safe Project
details.}

\item{user}{String with the user name for which to extract Safe People
details.}

\item{logs_from}{Lower limit timestamp to filter out the outputs generated
(default: \code{-Inf}, everything up to \code{logs_to})}

\item{logs_to}{Upper limit timestamp to filter out the outputs generated
(default: \code{Inf}, everything from \code{logs_from} onwards).}

\item{path}{String with path pointing to the root of the RO-Crate. This will
be used to store log files. If not provided, logs will be stored within
the RO-Crate returned by this function.}
}
\value{
Audit RO-Crate with 5 Safes Components.
}
\description{
Internal function to create audits for various back-ends.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-safe_project.R
\name{flatten_safe_project}
\alias{flatten_safe_project}
\alias{flatten_safe_project.default}
\alias{flatten_safe_project.rocrate}
\title{Flatten object with Safe Project details}
\usage{
flatten_safe_project(x, ...)

\method{flatten_safe_project}{default}(x, ...)

\method{flatten_safe_project}{rocrate}(x, ..., y = x)
}
\arguments{
\item{x}{Object (e.g., RO-Crate) with Safe Project details. This
can be generated with the \code{\link[=extract_safe_project]{extract_safe_project()}} function.}

\item{...}{Other optional arguments (not in used).}

\item{y}{Object (e.g., RO-Crate) with Safe Data details. This can be
generated with the \code{\link[=extract_safe_data]{extract_safe_data()}} function. If not provided, it
uses the \code{x} by default.}
}
\value{
Data frame with safe project details.
}
\description{
Flatten object with Safe Project details
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report.R
\name{report}
\alias{report}
\alias{report.character}
\alias{report.default}
\alias{report.list}
\alias{report.rocrate}
\title{Create an RO-Crate report}
\usage{
report(x, ...)

\method{report}{character}(
  x,
  ...,
  title = "DataSHIELD Report",
  filepath = tempfile(fileext = ".md"),
  render = TRUE,
  doc_format = "html",
  overwrite = FALSE,
  include_user_perm = TRUE,
  diag_title = "DataSHIELD server",
  diag_width = NULL,
  diag_height = NULL,
  max_line_length = 200
)

\method{report}{default}(x, ...)

\method{report}{list}(
  x,
  ...,
  study_name,
  title = "DataSHIELD Report",
  filepath = tempfile(fileext = ".md"),
  render = TRUE,
  doc_format = "html",
  overwrite = FALSE,
  include_user_perm = TRUE,
  diag_title = "DataSHIELD server",
  diag_width = NULL,
  diag_height = NULL,
  max_line_length = 200
)

\method{report}{rocrate}(
  x,
  ...,
  title = "DataSHIELD Report",
  filepath = tempfile(fileext = ".md"),
  render = TRUE,
  doc_format = "html",
  overwrite = FALSE,
  include_user_perm = TRUE,
  diag_title = "DataSHIELD server",
  diag_width = NULL,
  diag_height = NULL,
  max_line_length = 200
)
}
\arguments{
\item{x}{This can be an RO-Crate (\link[rocrateR:rocrate]{rocrate} class) or a
string with the path to an RO-Crate.}

\item{...}{Other optional arguments. See the full documentation,
\code{\link[=report]{?dsROCrate::report}}.}

\item{title}{String with title for the report (default: 'DataSHIELD Report').}

\item{filepath}{String with file path for Markdown report with the summary
of the given object, \code{x}.}

\item{render}{Boolean flag to indicate whether to render the markdown report.}

\item{doc_format}{String with file format for the markdown report.}

\item{overwrite}{Boolean flag to indicate whether to overwrite a previous
version of markdown report.}

\item{include_user_perm}{Boolean flag to indicate whether to include user
permissions in the report overview's diagram.}

\item{diag_title}{String with title for the 'root' of the diagram (default:
'DataSHIELD server').}

\item{diag_width}{Numeric value with width (in inches) for the report
overview's diagram (default: \code{NULL}, estimated based on number of nodes).}

\item{diag_height}{Numeric value with height (in inches) for the report
overview's diagram (default: \code{NULL}, estimated based on number of nodes).}

\item{max_line_length}{Integer with the maximum number of characters per line
in the RO-Crate to be printed in the report.}

\item{study_name}{String with the study name.}
}
\value{
RO-Crate report as markdown (.md) file and/or HTML.
}
\description{
Create an RO-Crate report
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/safe_people.R
\name{safe_people}
\alias{safe_people}
\alias{safe_people.character}
\alias{safe_people.opal}
\alias{safe_people.rocrate}
\title{Safe People details}
\source{
\itemize{
\item Research Data Scotland, 2025. "What is the Five Safes framework?".
\url{https://www.researchdata.scot/engage-and-learn/data-explainers/what-is-the-five-safes-framework/}
}
}
\usage{
safe_people(x, ...)

\method{safe_people}{character}(
  x,
  ...,
  user = attr(x, "user"),
  user_id_suffix = "#person:",
  set_author = TRUE,
  set_project = TRUE,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables")
)

\method{safe_people}{opal}(
  x,
  ...,
  rocrate = rocrateR::rocrate_5s(),
  user = NULL,
  user_id_suffix = "#person:",
  set_author = TRUE,
  set_project = TRUE,
  path = NULL,
  profile = "default",
  project = NULL,
  resources = NULL,
  tables = NULL
)

\method{safe_people}{rocrate}(
  x,
  ...,
  user = attr(x, "user"),
  user_id_suffix = "#person:",
  set_author = TRUE,
  set_project = TRUE,
  connection = attr(x, "connection"),
  path = attr(x, "path"),
  profile = attr(x, "profile"),
  project = attr(x, "project"),
  resources = attr(x, "resources"),
  tables = attr(x, "tables")
)
}
\arguments{
\item{x}{This can be a connection to a 'DataSHIELD' server (e.g., object with
the \code{opal} class, see \code{\link[opalr:opal.login]{opalr::opal.login()}}), an RO-Crate
(\link[rocrateR:rocrate]{rocrate} class) or a string with the path to an
RO-Crate.}

\item{...}{Other optional arguments. See the full documentation,
\code{\link[=safe_people]{?dsROCrate::safe_people}}.}

\item{user}{List (or \link[rocrateR:entity]{entity} object) with details for
the Safe People, it must include \verb{@id} and \code{name} entries. Alternatively,
this can be a string with the \code{name} of the current user.}

\item{user_id_suffix}{String with ID suffix for the tables/datasets
entities in the RO-Crate (default: \code{"#dataset:"}).}

\item{set_author}{Boolean flag to indicate if the current user should be
set as the author of the RO-Crate.}

\item{set_project}{Boolean flag to indicate if any \code{Project} entities found
in \code{x} should be linked to the Safe People entity.}

\item{connection}{Connection object for the 'DataSHIELD' server where the
values will be extracted from (e.g., OBiBa's Opal). Optional, if \code{x} is
set to a connection object. If so, then \code{rocrate} is required.}

\item{path}{String with path pointing to the root of the RO-Crate. This will
be used to store log files. If not provided, logs will be stored within
the RO-Crate returned by this function.}

\item{profile}{String with profile name (used for OBiBa's Opal backend).}

\item{project}{String with the name of the \link[=safe_project]{Safe Project}.}

\item{resources}{Vector of strings with the names of the resources, part of
\code{project}. Optional, if not provided, all the resources associated to
\code{project} will be included in the RO-Crate.}

\item{tables}{Vector of strings with the names of the tables/datasets, part
of \code{project}. Optional, if not provided, all the tables/datasets
associated to \code{project} will be included in the RO-Crate.}

\item{rocrate}{RO-Crate object. Optional, if \code{x} is either an RO-Crate
object or a path to a valid RO-Crate. If so, then \code{connection} is
required (default: \code{rocrateR::rocrate_5s()}).}
}
\value{
Updated RO-Crate object with Safe People information.
}
\description{
Safe People details for the RO-Crate.
}
\details{
Researchers must be accredited and trained before accessing the data that
has been prepared for them.

The access service provider may require the researcher to sign a statement
that they know and understand the regulations of the TRE.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-rocrate.R
\name{load_content}
\alias{load_content}
\title{Load \code{content} from external files}
\usage{
load_content(rocrate, roc_path)
}
\arguments{
\item{rocrate}{Object with the \link[rocrateR:rocrate]{rocrate} class.}

\item{roc_path}{String with path to the root of the RO-Crate.}
}
\value{
Update \code{rocrate} object.
}
\description{
Load \code{content} from external files
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-connection.R
\name{project_exists}
\alias{project_exists}
\alias{project_exists,armadillo-method}
\alias{project_exists.opal}
\alias{project_exists.ArmadilloCredentials}
\title{Verify if project exists}
\usage{
\S4method{project_exists}{armadillo}(
  x,
  ...,
  project
)

\method{project_exists}{opal}(x, ..., project)

\method{project_exists}{ArmadilloCredentials}(x, ..., project)
}
\arguments{
\item{x}{Connection object to backend for DataSHIELD server (e.g., Opal).}

\item{...}{Optional arguments, unused.}

\item{project}{String with project name to be verified.}
}
\value{
Nothing, call for its side effect. Stop execution of script if
\code{project} does not exist in the given server.
}
\description{
Wrapper for the \code{\link[opalr:opal.project_exists]{opalr::opal.project_exists()}} and
\code{\link[MolgenisArmadillo:armadillo.list_projects]{MolgenisArmadillo::armadillo.list_projects()}} functions.
}
\seealso{
Other Armadillo: 
\code{\link{parse_user_profiles}()}

Other Opal: 
\code{\link{get_project_details}()},
\code{\link{get_project_tables}()},
\code{\link{get_table_permissions}()},
\code{\link{parse_user_profiles}()}

Other Armadillo: 
\code{\link{parse_user_profiles}()}
}
\concept{Armadillo}
\concept{Opal}
\keyword{internal}
